#ifndef _UDM_COMMON_H
#define _UDM_COMMON_H

#include <stdio.h> /* for FILE etc. */

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#define uint4 unsigned int
#define uint8 unsigned long long

/* Some constants */
#define UDM_LANGPERDOC				16	/* FIXME */
#define UDM_MAXPARSER				256	/* FIXME */
#define UDM_USER_AGENT				"UdmSearch"
#define UDM_MAXWORDINFO				1024
#define UDM_MAXWORDPERQUERY			32

/* Some sizes and others definitions */
#define UDM_URLSIZE				127
#define UDM_MAXDOCSIZE				1024*1024  /* 1Mb */
#define UDM_MAXWORDSIZE				32
#define UDM_MAXDISCWORDSIZE			64
#define UDM_DEFAULT_REINDEX_TIME		7*24*60*60
#define UDM_DEFAULT_MAX_HOPS			256
#define UDM_READ_TIMEOUT			30
#define UDM_DOC_TIMEOUT				90
#define UDM_MAXNETERRORS			16
#define UDM_MAXTITLESIZE			128
#define UDM_MAXKEYWORDSIZE			255
#define UDM_MAXDESCSIZE				100
#define UDM_DEFAULT_NET_ERROR_DELAY_TIME	86400
#define UDM_MAXFROMSIZE				255
#define UDM_MAXSUBJSIZE				1024
#define UDM_MAXGROUPSIZE			255
#define UDM_MAXREFSIZE				1024

#ifdef NEWS_EXT
#define UDM_MAXTEXTSIZE				256000
#else
#define UDM_MAXTEXTSIZE				255
#endif


/* Locking consts */
#define UDM_LOCK		1
#define UDM_UNLOCK		2
#define UDM_LOCK_USER		0
#define UDM_LOCK_WIN		15
#define UDM_LOCK_ROBOTS		16
#define UDM_LOCK_QUERY		17
#define UDM_LOCK_TARGET		18
#define UDM_LOCK_RESOLVE	19
#define UDM_LOCK_CACHE		20
#define UDM_LOCK_CACHE_STORE	21
#define UDM_LOCK_WORDS		22
#define UDM_LOCK_CACHE_BUSY	23
#define UDM_LOCK_MAX		24


/* storage types */
#define UDM_DBMODE_SINGLE	0
#define UDM_DBMODE_MULTI	1
#define UDM_DBMODE_SINGLE_CRC	2
#define UDM_DBMODE_WORD2URL	3
#define UDM_DBMODE_MULTI_CRC	4
#define UDM_DBMODE_CACHE	5
#define UDM_DBMODE_BLOB		6


/* database open modes */
#define UDM_OPEN_MODE_READ	0
#define UDM_OPEN_MODE_WRITE	1


/* search modes */
#define UDM_ORD_RATE		0
#define UDM_ORD_DATE		1
#define UDM_MODE_ALL		0
#define UDM_MODE_ANY		1
#define UDM_MODE_BOOL		2
#define UDM_MODE_PHRASE		3

/* word match type */
#define UDM_MATCH_FULL		0
#define UDM_MATCH_BEGIN		1
#define UDM_MATCH_SUBSTR	2
#define UDM_MATCH_END		3
#define UDM_MATCH_REGEX		4
#define UDM_MATCH_WILD		5
#define UDM_MATCH_ICASE		1

/* search cache */
#define UDM_CACHE_ENABLED	1
#define UDM_CACHE_DISABLED	0

/* Indexer return codes */
#define IND_UNKNOWN		0
#define IND_OK			1
#define IND_ERROR		2
#define IND_NO_TARGET		3
#define IND_TERMINATED		4
#define IND_FLUSH		5


/* Flags for indexing */
#define UDM_FLAG_REINDEX	1
#define UDM_FLAG_EXP_FIRST	2
#define UDM_FLAG_ADD_SERV	4
#define UDM_FLAG_MARK		8
#define UDM_FLAG_SPELL		16
#define UDM_FLAG_INSERT		32
#define UDM_FLAG_SORT_HOPS	64
#define UDM_FLAG_INIT		128
#define UDM_FLAG_SKIP_LOCKING	256
#define UDM_FLAG_LOAD_LANGMAP	512

/* Tracking modes */
#define UDM_TRACK_DISABLED	0
#define UDM_TRACK_QUERIES	1
#define UDM_TRACK_WORDS		2


/* URLFile actions */
#define UDM_URL_FILE_REINDEX	1
#define UDM_URL_FILE_CLEAR	2
#define UDM_URL_FILE_INSERT	3
#define UDM_URL_FILE_PARSE	4

/* Ispell mode binary flags */
#define UDM_ISPELL_MODE_DB	1
#define UDM_ISPELL_USE_PREFIXES	2
#define UDM_ISPELL_MODE_SERVER  4

/* Action type: HTTP methods */
#define UDM_NOTFOUND	0
#define UDM_GET		1
#define UDM_DISALLOW	2
#define UDM_HEAD	3
#define UDM_HREFONLY	4
#define UDM_CHECKMP3	5

/************** Language and charset guesser *************/


#define UDM_LM_MAXGRAM		3
#define UDM_LM_HASHMASK		0x0FFF

typedef struct {
	int count;
	char str[UDM_LM_MAXGRAM+3];
} UDM_LANGITEM;

typedef struct {
	float  expectation;			/* Average value   */
	float  dispersion;			/* Math dispersion */
	char * lang;				/* Map Language    */
	char * charset;				/* Map charset     */
	UDM_LANGITEM  memb[UDM_LM_HASHMASK+1];	/* Items list      */
} UDM_LANGMAP;

typedef struct {
	size_t nmaps;
	UDM_LANGMAP * maps;
} UDM_LANGMAPLIST;

/*****************************************************/

/* StopList unit */
typedef struct udm_stopword_struct {
	char *word;
	char *lang;
} UDM_STOPWORD;

/* Main search structure */
typedef struct{
	int url_id;
	uint4 coord;
} UDM_URL_CRD;

/* Word list unit */
typedef struct {
	int coord;
	char *word;
} UDM_WORD;

typedef struct {
	size_t mwords;		/* Number of memory allocated for words     */
	size_t nwords;		/* Real number of words in list             */
	size_t swords;		/* Number of words in sorted list           */
	UDM_WORD *Word;		/* Word list  itself                        */
} UDM_WORDLIST;


#define UDM_WRDCOORD(p,w)	( (((unsigned int)(p))<<16)+(((unsigned int)(w))<<8) )
#define UDM_WRDSEC(c)		( (((unsigned int)(c))>>8)&0xFF )
#define UDM_WRDPOS(c)		( ((unsigned int)(c))>>16 )
#define UDM_WRDNUM(c)		( ((unsigned int)(c))&0xFF )
#define UDM_WRDMASK(c)		( 1 << (((unsigned int)(c))&0xFF) )


/********************** BLOB mode ******************************/

#define UDM_FLASH_BEST	0
#define UDM_FLASH_ALL	1

#define UDM_WRD_MASK		0x3F	/* 64 CRC tables: must be <256 */
#define UDM_WRD_TABLEN		0xFFF	/* 4K hash for each one        */

#define UDM_WRD_TABNUM(c)	((c>>24)&UDM_WRD_MASK)
#define UDM_WRD_TABPOS(c)	((c)&UDM_WRD_TABLEN)

#define UDM_WRD_MAXLENNO	16	/* 16 parts by length */
#define UDM_WRD_LENSIZE(l)	((l-1)/4+1)*4
#define UDM_WRD_LENNO(l)	((l-1)/4)

typedef struct {
	size_t mwords;		/* Number of memory allocated for words     */
	size_t nwords;		/* Real number of words in list             */
	char * word;
	char hasfile;
	UDM_URL_CRD *Coords;	/* Coords list  itself                      */
} UDM_WRDLIST;

/***************************************************************/

/* Cross-word list unit */
typedef struct {
	short pos;
	short weight;
	char *word;
	char *url;
	int  referree_id;
} UDM_CROSSWORD;

typedef struct {
	size_t ncrosswords;
	size_t mcrosswords;
	UDM_CROSSWORD * CrossWord;
} UDM_CROSSLIST;

/* Language quesser structure for one language   */
/* We could use the type above instead of course */
typedef struct udm_lang_struct {
	int  count;
	char lang[3];
} UDM_LANG;


/* Structure to store server parameters */
typedef struct {
	/* General items */
	int  rec_id;	/* to store order of appearence */
	int  match_type;
	void * regexp; 	/* for realm */
	char *url;
	char *alias;    /* For primary aliases */
	char *charset;
	char *category;
	char *tag;
	char *lang;
	int  period;

	/* Auth */
	char *basic_auth;
	char *user;
	char *passwd;

	/* Proxy stuff */
	char *proxy;
	char *proxy_basic_auth;
	int  proxy_port;

        /* HTDB */
	char *htdb_list;
	char *htdb_doc;

	/* Mirror stuff */
	int  use_mirror;
	char *mirror_root;
	char *mirror_headers;

	/* Network stuff */
	int  net_errors;
	int  max_net_errors;
	int  net_error_delay_time;
	int  read_timeout;
	int  doc_timeout;

	/* Spider & indexer stuff */
	int  index;
	int  follow;
	int  maxhops;
	int  deletebad;
	int  use_robots;
	int  use_clones;
	int  delete_no_server;

	/* Words sections */
	int  cross_section;
	int  body_section;
	int  title_section;	
	int  desc_section;
	int  kwrd_section;
	int  proto_section;
	int  host_section;
	int  path_section;
	int  file_section;
	
	/* Words factors */
	int  correct_factor;
	int  incorrect_factor;
	int  number_factor;
	int  alnum_factor;

	/* MP3 extensions */
	int check_mp3_tag;
	int check_only_mp3_tag;
} UDM_SERVER;

/* All links are stored in the cache of this structure */
/* before actual INSERT into database                  */
typedef struct {
	char * url;
	char * tag;
	char * category;
	int referrer;
	int hops;
	int stored;	
} UDM_HREF;

typedef struct {
	size_t mhrefs;
	size_t nhrefs;
	size_t shrefs;
	size_t dhrefs;
	UDM_HREF * Href;
} UDM_HREFLIST;


/* HTTP Headers */
typedef struct {
	char * name;
	char * value;
} UDM_HTTPHEADER;


typedef struct {
	size_t nheaders;
	UDM_HTTPHEADER * Headers;
} UDM_HTTPHEADERLIST;


/* Some components in this structure correspondent */
/* to fields in the "url" table                   */

#define UDM_CATSIZE	8
#define UDM_TAGSIZE	4

typedef struct {
	/* Static fields */
	int	stored; /* For AddHref() */
	int	url_id;
	int	origin_id;
	int	status;
	int	size;
	int	referrer;	
	int	hops;	
	int	crc32;
	time_t	last_mod_time;
	time_t  next_index_time;
	int	site_id;
	int	parent_id;
	char	*category;
	char	*tag;
	char	*lang;

	/* Variable fields */
	char	*url;
	char	*content_type;
	char	*charset;
	char	*subject;
	char	*from;
	char	*newsgroups;
	char	*references;
	char	*message_id;

	char	*title;
	char	*keywords;
	char	*description;
	char 	*text;

	char	*textbuf;		/* Doc text representation */

	char	*headers;
	char	*content;
	char	*request;

	int	method;
	int	use_proxy;
	int	is_mp3;

	UDM_LANGMAP LangMap;		/* Guesser data */
	UDM_HREFLIST HrefList;		/* Link list    */
	UDM_WORDLIST   WordList;	/* Word list    */
	UDM_CROSSLIST  CrossList;	/* Crosswords   */
	UDM_HTTPHEADERLIST HeaderList;	/* HEaders	*/
	int wordpos;			/* for phrases	*/

	/* For search result */
	int	rating;
	int	order;
	int	sdnum;

	int    content_encoding;  /* Content encoding type */
} UDM_DOCUMENT;

typedef struct udm_conn_struct {
        int status;
        int connected;
        int err;
        int retry;
        int conn_fd;
        int port;
        int timeout;
        char *hostname;
        struct sockaddr_in sin;
        int buf_len;
        int buf_len_total;
        int len;
        char *buf;
        struct udm_conn_struct *connp;
        struct udm_indexer_struct *indexer;
} UDM_CONN;

typedef struct {
	int match_type;
	int case_sense;
	char *pattern;
	char *reg;
} UDM_MATCH;

typedef struct {
	int beg;
	int end;
} UDM_MATCH_PART;

/* Aliases item structure */
typedef struct {
	UDM_MATCH	match;
	char		*replace;
} UDM_ALIAS;

typedef struct {
	size_t naliases;
	size_t maliases;
	UDM_ALIAS * Alias;
} UDM_ALIASLIST;

/* External Parsers */
typedef struct udm_parser_struct{
        char *from_mime;
	char *to_mime;
	char *cmd;
} UDM_PARSER;

/* Resolve stuff */
typedef struct udm_host_addr_struct {
	char *hostname;
	struct in_addr addr;
	int net_errors;
	time_t last_used;
}UDM_HOST_ADDR;

/* Unicode regex lite BEGIN */

#define UDM_UNIREG_SUB	1
#define UDM_UNIREG_BEG	2
#define UDM_UNIREG_END	3
#define UDM_UNIREG_INC	4
#define UDM_UNIREG_EXC	5

typedef struct{
	int type;
	int * str;
} UDM_UNIREG_TOK;

typedef struct {
	size_t ntokens;
	UDM_UNIREG_TOK * Tokens;
} UDM_UNIREG_EXP;

/* Unicode regex lite END */


/* Ispell BEGIN */


typedef struct spell_struct {
	int * word;
	char flag[11];
	char lang[33];
} UDM_SPELL;


typedef struct aff_struct {
	char flag;
	char type;
	char lang[33];
	int mask[41];
        int find[16];
	int repl[16];
	UDM_UNIREG_EXP reg;
	size_t replen;
        size_t findlen;
        char compile;
} UDM_AFFIX;

/* Ispell END */


typedef struct{
	int cmd; /* 'allow' or 'disallow' */
	char *path;
} UDM_ROBOT_RULE;

typedef struct{
	char *hostinfo;
	size_t nrules;
	UDM_ROBOT_RULE *rules;
} UDM_ROBOT;

typedef struct{
	size_t nrobots;
	UDM_ROBOT *robots;
} UDM_ROBOTS;

#define MAX_SEARCH_LIMIT 8

typedef struct{
	int	type;
	char	file_name[1024];
	uint4	hi;
	uint4	lo;
} UDM_SEARCH_LIMIT;

/* VARLISTs */

#define UDM_VARSRC_QSTRING	1
#define UDM_VARSRC_GLOBAL	2

typedef struct {
	int scope;
	char * name;
	char * val;
} UDM_VAR;

typedef struct {
	int nvars;
	UDM_VAR * var;
} UDM_VARLIST;


typedef struct Tree_struct {
        int Left[256], Right[256];
        char lang[3];
} Tree_struct;

/* Search daemon types */
typedef struct {
	int fd;
	char * addr;
} UDM_SDCLIENT;

typedef struct {
	size_t	nclients;
	UDM_SDCLIENT * Clients;
} UDM_SDCLLIST;

typedef struct {
	size_t	order;
	size_t	count;
	int	crcword;
	char	*word;
	int	*uword;
	size_t	len;
} UDM_WIDEWORD;

typedef struct {
	size_t nuniq;
	size_t nwords;
	UDM_WIDEWORD * Words;
} UDM_WIDEWORDLIST;


typedef struct {
	UDM_WIDEWORD p;
	UDM_WIDEWORD s;
} UDM_SYNONYM;

typedef struct {
	size_t nsynonyms;
	size_t msynonyms;
	UDM_SYNONYM * Synonyms;
} UDM_SYNONYMLIST;

typedef struct {
	int type;
	int section;
	char * field;
	size_t fieldlen;
	char * name;
} UDM_SECTION;

typedef struct {
	size_t nsec;
	UDM_SECTION * Sections;
} UDM_SECTIONLIST;


typedef struct {
	size_t first;
	size_t last;
	size_t total_found;
	size_t num_rows;
	size_t work_time;
	char wordinfo[UDM_MAXWORDINFO];
	UDM_DOCUMENT * Doc;
} UDM_RESULT;


typedef struct {
	size_t num_rows;
	size_t num_indexed;
	UDM_DOCUMENT * Doc;
} UDM_TARGETS;


/* Config file */
typedef struct udm_config_struct {
	char vardir[1024];
	int errcode;
	char errstr[2048];
	char * local_charset;
	char * browser_charset;
	int force1251;
	int  min_word_len;
	int  max_word_len;
	size_t max_doc_size;
	char user_agent[128];
	char extra_headers[1024];
	char * url_file_name;
	size_t nservers;
	size_t mservers;
	UDM_TARGETS Targets;
	UDM_LANGMAPLIST LangMapList;
	UDM_SERVER * Server;
	UDM_SERVER * csrv;
	UDM_ALIASLIST AliasList;
	UDM_ALIASLIST ReverseAliasList;
	char * alias_prog;
	UDM_HREFLIST HrefList;
	UDM_SECTIONLIST SectionList;
	int have_targets;
	void * Filter;
	size_t nfilters;
	size_t mfilters;
	size_t nmimes;
	size_t mmimes;
	void * Mime;
	size_t nparsers;
	UDM_PARSER *parsers;
	size_t nhost_addr;
	size_t mhost_addr;
	UDM_HOST_ADDR * host_addr;
	size_t naffixes;
	size_t maffixes;
	UDM_AFFIX * Affix;
	size_t nspell;
	size_t mspell;
	UDM_SPELL * Spell;
	Tree_struct SpellTree[UDM_LANGPERDOC];
	Tree_struct PrefixTree[UDM_LANGPERDOC];
	Tree_struct SuffixTree[UDM_LANGPERDOC];
	size_t nstoplist;
	UDM_STOPWORD * stoplist;
	UDM_ROBOTS Robots;
	UDM_SYNONYMLIST SynList;
	char * DBAddr;
	int    DBMode;
	int    DBUseLock;

	char catstr[4096];
	char langstr[4096];
	char tagstr[4096];
	char statusstr[4096];
	char urlstr[4096];
	char timestr[4096];

	UDM_SEARCH_LIMIT	limits[8];
	int			nlimits;

	char catlimit[4096];
	char taglimit[4096];
	int  sitelimit;

	char srv_tables[512];

	/* for logger */
	int logFacility;
	int logLevel;
	int is_log_open;
	FILE *logFD;

	/* For bult-in database */
	char * ul;
	char * ttag;
	int n_urls;
	UDM_DOCUMENT * URLs;

	void (*ThreadInfo)(int handle,const char * state,const char * str);
	void (*StatInfo)(int handle, int code, int expired, int total, const char * str);
	void (*RefInfo)(int code,const char *url, const char *ref);
	void (*LockProc)(int command, int type);

	/* Cache mode logs file descriptor */
	int logd_fd;
	char * logd_addr;
	
	/* URL server */
	int urld_fd;
	char * urld_addr;
		
	int use_crc32_url_id;	/* Whether to generate url_id using crc32 */
	int use_remote_cont_type;
	int nLang;
	int use_crossword;
	int use_newsext;
	char *spellhost;
	UDM_WRDLIST WrdList[UDM_WRD_MASK+1][UDM_WRD_TABLEN+1];
	size_t ndoc;
	int flashmode;
	UDM_VARLIST *vars;	/* cgi params 		*/
	UDM_SDCLLIST sdcl;	/* Searchd addresses	*/

        char *stored_addr;      /* StoreServer hostname */
} UDM_ENV;


typedef struct udm_url {
	char schema[UDM_URLSIZE];
	char specific[UDM_URLSIZE];
	char hostinfo[UDM_URLSIZE];
	char auth[UDM_URLSIZE];
	char hostname[UDM_URLSIZE];
	char path[UDM_URLSIZE];
	char filename[UDM_URLSIZE];
	char anchor[UDM_URLSIZE];
	int  port;
	int  default_port;
} UDM_URL;

typedef struct udm_tag {
	char *tag;
	char *href;
	char *src;
	char *content;
	char *value;
	char *selected;
	char *name;
	char *type;
	char *lang;
} UDM_TAG;


typedef struct udm_category_struct {
	int   rec_id;
	char  path[128];
	char  link[128];
	char  name[128];
} UDM_CATEGORY;


/* Boolean search constants and types */
#define UDM_MAXSTACK 100
#define UDM_STACK_LEFT  0
#define UDM_STACK_RIGHT 1
#define UDM_STACK_BOT   2
#define UDM_STACK_OR    3
#define UDM_STACK_AND   4
#define UDM_STACK_NOT   5
#define UDM_STACK_WORD  200

typedef struct {
	int ncstack;
	int cstack[UDM_MAXSTACK];
	int nastack;
	int astack[UDM_MAXSTACK];
} UDM_BOOLSTACK;

typedef struct {
	int cmd;
	int arg;
} UDM_STACK_ITEM;




/* Indexer */
typedef struct udm_indexer_struct{
	int handle;		/* Indexer handle for  threaded version */
	int DBError;		/* Database error code			*/
	int action;		/* Callback function may request action	*/
	char *buf;		/* Buffer for document			*/
	void * db;              /* Database handle			*/
	UDM_SDCLLIST sdcl;	/* Searchd clients list			*/
	UDM_CONN *connp;
	UDM_ENV * Conf;
	char wordinfo[UDM_MAXWORDINFO];
	int  search_mode;
	int  word_match;
	int  sort_order;
	int  page_number;
	int  page_size;
	int  track_mode;
	UDM_SPELL t_Spell;
	int read_timeout;
	int doc_timeout;
	int cache_mode;

	size_t total_found;
	size_t loaded_nwords;
	size_t withoffset;
	int spellang;

	/* Search stuff */
	UDM_WIDEWORDLIST WWList;

	/* Bool stuff */
	size_t nitems;
	UDM_STACK_ITEM items[UDM_MAXSTACK];

	int wf[256];
	char *glang; /* language selection from CGI form */
} UDM_AGENT;

typedef struct {
	char * url;
	int  status;
} UDM_URLSTATE;

typedef int (*qsort_cmp)(const void*, const void*);

typedef struct {
	uint4	hi,lo;
	uint4	url_id;
} UDM_UINT8_URLID;

typedef struct {
	uint4	hi,lo;
	off_t	pos;
	size_t	len;
} UDM_UINT8_POS_LEN;

typedef struct {
	uint4	val;
	uint4	url_id;
} UDM_UINT4_URLID;

typedef struct {
	uint4	val;
	off_t	pos;
	size_t	len;
} UDM_UINT4_POS_LEN;


#define UDM_SEARCHD_CMD_ERROR	1
#define UDM_SEARCHD_CMD_MESSAGE	2
#define UDM_SEARCHD_CMD_WORDS	3
#define UDM_SEARCHD_CMD_GOODBYE	4
#define UDM_SEARCHD_CMD_DOCINFO	5
#define UDM_SEARCHD_CMD_CLONES	6

typedef struct {
	size_t cmd;
	size_t len;
} UDM_SEARCHD_PACKET_HEADER;


#define UDM_SECTION_BODY	1
#define UDM_SECTION_TITLE	2
#define UDM_SECTION_DESC	3
#define UDM_SECTION_KWRD	4
#define UDM_SECTION_FILE	5
#define UDM_SECTION_PATH	6
#define UDM_SECTION_HOST	7
#define UDM_SECTION_PROTO	8
#define UDM_SECTION_CROSSWORDS	9

#define UDM_SECTION_HEADER	257
#define UDM_SECTION_META	258


#define UDM_MAXTAGVAL	64

typedef struct {
	int type;
	int script;
	int style;
	int title;
	int body;
	int follow;
	int index;
	int comment;
	char * lasthref;
	size_t ntoks;
	struct {
		const char *name;
		const char *val;
		size_t     nlen;
		size_t     vlen;
	} toks[UDM_MAXTAGVAL+1];
} UDM_HTMLTOK;


#endif
