#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <regex.h>

#include "udm_common.h"
#include "udm_spell.h"
#include "udm_db.h"
#include "udm_proto.h"
#include "udm_url.h"
#include "udm_parser.h"
#include "udm_conf.h"
#include "udm_indexer.h"
#include "udm_log.h"
#include "udm_hrefs.h"
#include "udm_robots.h"
#include "udm_utils.h"
#include "udm_xmalloc.h"
#include "udm_host.h"
#include "udm_server.h"
#include "udm_alias.h"
#include "udm_filter.h"
#include "udm_search_tl.h"
#include "udm_env.h"
#include "udm_mimetype.h"
#include "udm_stopwords.h"
#include "udm_guesser.h"
#include "udm_section.h"
#include "udm_unicode.h"
#include "udm_synonym.h"

/****************************  Load Configuration **********************/

static int UdmFollowType(char * follow){
	if(!strcasecmp(follow,"no"))return(UDM_FOLLOW_NO);
	if(!strcasecmp(follow,"page"))return(UDM_FOLLOW_NO);
	if(!strcasecmp(follow,"yes"))return(UDM_FOLLOW_PATH);
	if(!strcasecmp(follow,"path"))return(UDM_FOLLOW_PATH);
	if(!strcasecmp(follow,"site"))return(UDM_FOLLOW_SITE);
	if(!strcasecmp(follow,"world"))return(UDM_FOLLOW_WORLD);
	return(UDM_FOLLOW_UNKNOWN);
}



#define IF_NUM_PARAM(src,name,dst,def) \
if(!UDM_STRNCASECMP(src,name)){if(sscanf(src+strlen(name),"%d",&(dst))!=1)dst=def;}

#define IF_TIME_PARAM(src,name,dst,def) \
if (!UDM_STRNCASECMP(src,name)){ \
	if ((dst=Udm_dp2time_t(src+strlen(name)))==-1) \
		dst=def; \
}

#define IF_BOOL_PARAM(src,name,dst,def) \
if(!UDM_STRNCASECMP(src,name)){ \
char xxx[1024]; \
if(sscanf(src+strlen(name),"%s",xxx)==1){ \
	if(!UDM_STRNCASECMP(xxx,"yes"))dst=1; \
	else	dst=0; \
}else \
	dst=def; \
}


__INDLIB__  UDM_ENV * UdmLoadConfig(UDM_ENV * Conf,const char *conf_name,int config_level,int load_flags){
	int i=0; 		/* Line number */
	FILE *config;		/* File struct */
	char str[1024]="";	/* Unsafe copy - will be used in strtok	*/
	char str1[1024]="";	/* To concatenate lines			*/
	char ostr[1024]="";	/* Safe copy of original  config string	*/

	Conf->errcode=0;
	Conf->errstr[0]=0;

	if(config_level==0){ /* Do some initialization */
		Conf->force1251=0;
		Conf->nspell=0;
		Conf->csrv=(UDM_SERVER*)UdmXmalloc(sizeof(UDM_SERVER));
		UdmInitServer(Conf->csrv);
		snprintf(Conf->user_agent,sizeof(Conf->user_agent)-1,"%s/%s",UDM_USER_AGENT,VERSION);
		Conf->extra_headers[0]=0;
		Conf->max_doc_size=UDM_MAXDOCSIZE;
		UdmSetLogFacility(Conf, "");

		/* I have removed it to allow adding of URLs  */
		/* Passed in -i with -u command line args     */
		/* But it seems that UnloadConf() is required */
		/* REMOVED      UdmFreeHrefs(); */
	}

	/* Open config file */
	if(!(config=fopen(conf_name,"r"))){
		sprintf(Conf->errstr,"Error: can't open config file '%s': %s",conf_name, strerror(errno));
		Conf->errcode=1;
		return(Conf);
	}

	/*  Read lines and parse */
	while(fgets(str1,sizeof(str1),config)){
		char *end;
		i++;
		if(!str1[0])continue;
		end=str1+strlen(str1)-1;
		while((end>=str1)&&(*end=='\r'||*end=='\n'||*end==' ')){
			*end=0;if(end>str1)end--;
		}
		if(!str1[0])continue;
		if(str1[0]=='#')continue;

		if(*end=='\\'){
			*end=0;strcat(str,str1);
			continue;
		}
		strcat(str,str1);
		strcpy(str1,"");
		strcpy(ostr,str);

		if(!UDM_STRNCASECMP(str,"DBAddr")){
			UDM_FREE(Conf->DBAddr);
			Conf->DBAddr=strdup(UdmTrim(str+6," \t\r\n"));
		}else
		if(!UDM_STRNCASECMP(str,"VarDir")){
			sprintf(Conf->vardir,"%s%s",UdmTrim(str+6," \r\n\t"),UDMSLASHSTR);
		}else
		if(!UDM_STRNCASECMP(str,"DataDir")){
			sprintf(Conf->vardir,"%s%s",UdmTrim(str+6," \r\n\t"),UDMSLASHSTR);
		}else
		if(!UDM_STRNCASECMP(str,"DBMode")){
			UdmEnvSetDBMode(Conf,UdmTrim(str+6," \r\n\t"));
		}else
		if(!UDM_STRNCASECMP(str,"LogdAddr")){
			UDM_FREE(Conf->logd_addr);
			Conf->logd_addr=strdup(UdmTrim(str+8," \r\n\t"));
		}else
		if(!UDM_STRNCASECMP(str,"URLDAddr")){
			UDM_FREE(Conf->urld_addr);
			Conf->urld_addr=strdup(UdmTrim(str+8," \r\n\t"));
		}else
		if(!UDM_STRNCASECMP(str,"StoredAddr")){
			UDM_FREE(Conf->stored_addr);
			Conf->stored_addr=strdup(UdmTrim(str+8," \r\n\t"));
		}else
		if(!UDM_STRNCASECMP(str,"LocalCharset")){
			UDM_FREE(Conf->local_charset);
			Conf->local_charset=strdup(UdmTrim(str+13," \r\n\t"));
			if(!UdmGetCharSet(Conf->local_charset)){
				sprintf(Conf->errstr,"Error in config file '%s' line %d: charset '%s' is not supported\n",conf_name,i,Conf->local_charset);
				Conf->errcode=1;
				return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"UseRemoteContentType")){
					if(!UDM_STRNCASECMP(UdmTrim(str+20,"= \r\t\n"),"no")) 
					     Conf->use_remote_cont_type=0;
					else Conf->use_remote_cont_type=1;
		}else
		if(!UDM_STRNCASECMP(str,"Disallow")){
			if(UdmAddFilterStr(Conf,str,UDM_DISALLOW))
				return(Conf);
		}else	
		if(!UDM_STRNCASECMP(str,"Allow")){
			if(UdmAddFilterStr(Conf,str,UDM_GET))
				return(Conf);
		}else
		if(!UDM_STRNCASECMP(str,"CheckOnly")){
			if(UdmAddFilterStr(Conf,str,UDM_HEAD))
				return(Conf);
		}else
		if(!UDM_STRNCASECMP(str,"HrefOnly")){
			if(UdmAddFilterStr(Conf,str,UDM_HREFONLY))
				return(Conf);
		}else
		if(!UDM_STRNCASECMP(str,"AddType")){
			char *mime_type=NULL,*lt,*ttok;
			int flags=0;

			ttok=UdmGetStrToken(str+7,&lt);
			while(ttok){
				if(!strcasecmp(ttok,"regex"))
					flags|=UDM_MIME_REGEX;
				else
				if(!strcasecmp(ttok,"regexp"))
					flags|=UDM_MIME_REGEX;
				else
				if(!strcasecmp(ttok,"string"))
					flags&=~UDM_MIME_REGEX;
				else
				if(!strcasecmp(ttok,"case"))
					flags|=UDM_MIME_CS;
				else
				if(!strcasecmp(ttok,"nocase"))
					flags&=~UDM_MIME_CS;
				else
				if(!mime_type)
					mime_type=ttok;
				else{
					if(UdmAddType(Conf,mime_type,ttok,flags))
						return(Conf);
				}
				ttok=UdmGetStrToken(NULL,&lt);
			}
		}else
		if(!UDM_STRNCASECMP(str,"CharSet")){
			UDM_FREE(Conf->csrv->charset);
			Conf->csrv->charset=strdup(UdmTrim(str+7," \r\n\t"));
			if(!UdmGetCharSet(Conf->csrv->charset)){
				sprintf(Conf->errstr,"Error in config file '%s' line %d: charset '%s' is not supported\n",conf_name,i,Conf->csrv->charset);
				Conf->errcode=1;
				return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"ProxyAuthBasic")){
			UDM_FREE(Conf->csrv->proxy_basic_auth);
			if(sscanf(str+14,"%s",str1)){
				Conf->csrv->proxy_basic_auth=(char*) UdmXmalloc(BASE64_LEN(strlen(str1)));
				udm_base64_encode(str1,Conf->csrv->proxy_basic_auth,strlen(str1));
			}
		}else
		if(!UDM_STRNCASECMP(str,"Proxy")){
			UDM_FREE(Conf->csrv->proxy);
			if(sscanf(str+5,"%s",str1)&&strcmp(str1,""))
				Conf->csrv->proxy=strdup(str1);
		}else
		if(!UDM_STRNCASECMP(str,"Category")){
			UDM_FREE(Conf->csrv->category);
			if(sscanf(str+9,"%s",str1))
			Conf->csrv->category=strdup(str1);
		}else
		if(!UDM_STRNCASECMP(str,"Tag")){
			UDM_FREE(Conf->csrv->tag);
			if(sscanf(str+3,"%s",str1))
			Conf->csrv->tag=strdup(str1);
		}else
		if(!UDM_STRNCASECMP(str,"HTTPHeader")){
			char * arg;
			arg=UdmTrim(str+11," \r\n\t");
			if(sscanf(arg,"%[^\n\r]s",str1)&&strcmp(str1,"")){
				if(!UDM_STRNCMP(str1,"User-Agent: ")){
					strncpy(Conf->user_agent,str1+12,sizeof(Conf->user_agent)-1);
				}else{
					strcat(Conf->extra_headers,str1);
					strcat(Conf->extra_headers,"\r\n");
				}
			}
		}else
#ifdef HAVE_SYSLOG_H
		if(!UDM_STRNCASECMP(str,"SyslogFacility")){
			if(sscanf(str+14,"%s",str1)&&strcmp(str1,""))
				UdmSetLogFacility(Conf, str1);
		}else
#endif
		if(!UDM_STRNCASECMP(str,"AuthBasic")){
			UDM_FREE(Conf->csrv->basic_auth);
			UDM_FREE(Conf->csrv->user);
			UDM_FREE(Conf->csrv->passwd);
			if(sscanf(str+9,"%s",str1)){
				char * ch;
				Conf->csrv->basic_auth=(char*) UdmXmalloc(BASE64_LEN(strlen(str1)));
				udm_base64_encode(str1,Conf->csrv->basic_auth,strlen(str1));
				if ((ch=strchr(str1, ':'))){
					Conf->csrv->user=(char*)UdmXmalloc((size_t)(ch-str1+1));
					snprintf(Conf->csrv->user, (size_t)(ch-str1+1), "%s", str1);
					Conf->csrv->passwd=(char*)UdmXmalloc(strlen(ch)+1);
					sprintf(Conf->csrv->passwd, "%s", ch+1);
				}
			}
		}else
		if(!UDM_STRNCASECMP(str,"HTDBList")){
			UDM_FREE(Conf->csrv->htdb_list);
			Conf->csrv->htdb_list=strdup(UdmTrim(str+8," \t\r\n"));
		}else
		if(!UDM_STRNCASECMP(str,"HTDBDoc")){
			UDM_FREE(Conf->csrv->htdb_doc);
			Conf->csrv->htdb_doc=strdup(UdmTrim(str+8," \t\r\n"));
		}else
		if(!UDM_STRNCASECMP(str,"Mime")){
			char *mime_arg[4],*lt,*mtok;
			int argn=0;

			mtok=UdmGetStrToken(str+4,&lt);
			for(;(mtok)&&(argn<4);argn++){
				mime_arg[argn]=mtok;
				mtok=UdmGetStrToken(NULL,&lt);	
			}
			if(argn!=3){
				sprintf(Conf->errstr,"Error in config file '%s' line %d too %s arguments for Mime command\n",conf_name,i,(argn<3?"few":"many"));
				Conf->errcode=1;
				return(Conf);
			}else{
				UdmAddParser(Conf,mime_arg[0],mime_arg[1],mime_arg[2]);
			}
		}else
		if(!UDM_STRNCASECMP(str,"AliasProg")){
			char * aprog,* lt;
			if((aprog=UdmGetStrToken(str+9,&lt)))
				Conf->alias_prog=strdup(aprog);
		}else
		if(!UDM_STRNCASECMP(str,"Alias")){
			char * ttok, * lt;
			UDM_ALIAS Alias;
			
			bzero(&Alias,sizeof(Alias));
			Alias.match.match_type=UDM_MATCH_BEGIN;
			
			ttok=UdmGetStrToken(str+5,&lt);
			while(ttok){
				if(!strcasecmp(ttok,"regex"))
					Alias.match.match_type=UDM_MATCH_REGEX;
				else
				if(!strcasecmp(ttok,"regexp"))
					Alias.match.match_type=UDM_MATCH_REGEX;
				else
				if(!strcasecmp(ttok,"case"))
					Alias.match.case_sense=0;
				else
				if(!strcasecmp(ttok,"nocase"))
					Alias.match.case_sense=UDM_MATCH_ICASE;
				else
				if(!Alias.match.pattern){
					Alias.match.pattern=ttok;
				}else{
					char err[120]="";
					Alias.replace=ttok;
					if(UdmAliasListAdd(&Conf->AliasList,&Alias,err,sizeof(err))){
						snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"Error in config file '%s' line %d: %s: %s",conf_name,i,err,ostr);
						Conf->errcode=1;
						return(Conf);
					}
				}
				ttok=UdmGetStrToken(NULL,&lt);
			}
			if(!Alias.replace){
				snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"Error in config file '%s' line %d: too few arguments: %s",conf_name,i,ostr);
				Conf->errcode=1;
				return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"ReverseAlias")){
			char * ttok, * lt;
			UDM_ALIAS Alias;
			
			bzero(&Alias,sizeof(Alias));
			Alias.match.match_type=UDM_MATCH_BEGIN;
			
			ttok=UdmGetStrToken(str+12,&lt);
			while(ttok){
				if(!strcasecmp(ttok,"regex"))
					Alias.match.match_type=UDM_MATCH_REGEX;
				else
				if(!strcasecmp(ttok,"regexp"))
					Alias.match.match_type=UDM_MATCH_REGEX;
				else
				if(!strcasecmp(ttok,"case"))
					Alias.match.case_sense=0;
				else
				if(!strcasecmp(ttok,"nocase"))
					Alias.match.case_sense=UDM_MATCH_ICASE;
				else
				if(!Alias.match.pattern){
					Alias.match.pattern=ttok;
				}else{
					char err[120]="";
					Alias.replace=ttok;
					if(UdmAliasListAdd(&Conf->ReverseAliasList,&Alias,err,sizeof(err))){
						snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"Error in config file '%s' line %d: %s: %s",conf_name,i,err,ostr);
						Conf->errcode=1;
						return(Conf);
					}
				}
				ttok=UdmGetStrToken(NULL,&lt);
			}
			if(!Alias.replace){
				snprintf(Conf->errstr,sizeof(Conf->errstr)-1,"Error in config file '%s' line %d: too few arguments: %s",conf_name,i,ostr);
				Conf->errcode=1;
				return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"LangMapFile")){
			if(load_flags&UDM_FLAG_LOAD_LANGMAP){
				int res;
				char * arg;
				char filename[1024];

				arg=UdmTrim(str+11," \r\n\t");
				if(arg[0]=='/'){
					sprintf(filename,"%s",arg);
				}else{
					sprintf(filename,"%s/%s",UDM_CONF_DIR,arg);
				}
				if((res=UdmLoadLangMapFile(Conf,filename))){
					return(Conf);
				}
			}
		}else
		if(!UDM_STRNCASECMP(str,"DefaultLang")){
			if(sscanf(str+11,"%s",str1)&&strcmp(str1,""))
				Conf->csrv->lang=strdup(str1);
		}else
		if(!UDM_STRNCASECMP(str,"ServerTable")){
			int len;
			char * tables;
			
			tables=UdmTrim(str+11," \r\n\t");
			len=strlen(Conf->srv_tables);
			if((len+strlen(tables)+3)>sizeof(Conf->srv_tables)){
				sprintf(Conf->errstr,"Error in config file '%s' line %d: too many ServerTable commands",conf_name,i);
				Conf->errcode=1;
				return(Conf);
			}
			sprintf(Conf->srv_tables+len," %s",tables);
		}else
		if(!UDM_STRNCASECMP(str,"Server")||!UDM_STRNCASECMP(str,"Realm")||!UDM_STRNCASECMP(str,"Subnet")){
			int match_type, follow;
			char * cmd, * lt;

			if(!UDM_STRNCASECMP(str,"Server")){
				match_type=UDM_SERVER_SUBSTR|UDM_SERVER_MATCH;
			}else
			if(!UDM_STRNCASECMP(str,"Subnet")){
				match_type=UDM_SERVER_SUBNET|UDM_SERVER_MATCH;
			}else{
				match_type=UDM_SERVER_STRING|UDM_SERVER_MATCH;
			}

			/* Store follow value */
			follow=Conf->csrv->follow;
			
			cmd=UdmGetToken(str," \t\r\n",&lt);
			
			while((cmd=UdmGetToken(NULL," \t\r\n",&lt))){
				if(!strcasecmp(cmd,"no"))Conf->csrv->follow=UDM_FOLLOW_NO;
				else
				if(!strcasecmp(cmd,"page"))Conf->csrv->follow=UDM_FOLLOW_NO;
				else
				if(!strcasecmp(cmd,"yes"))Conf->csrv->follow=UDM_FOLLOW_PATH;
				else
				if(!strcasecmp(cmd,"path"))Conf->csrv->follow=UDM_FOLLOW_PATH;
				else
				if(!strcasecmp(cmd,"site"))Conf->csrv->follow=UDM_FOLLOW_SITE;
				else
				if(!strcasecmp(cmd,"world"))Conf->csrv->follow=UDM_FOLLOW_WORLD;
				else
				if(!strcasecmp(cmd,"nocase"))match_type&=~UDM_SERVER_CS;
				else
				if(!strcasecmp(cmd,"case"))match_type|=UDM_SERVER_CS;
				else
				if(!strcasecmp(cmd,"match"))match_type|=UDM_SERVER_MATCH;
				else
				if(!strcasecmp(cmd,"nomatch"))match_type&=~UDM_SERVER_MATCH;
				else
				if(!strcasecmp(cmd,"string"))UDM_SET_SRV_TYPE(match_type,UDM_SERVER_STRING);
				else
				if(!strcasecmp(cmd,"regex"))UDM_SET_SRV_TYPE(match_type,UDM_SERVER_REGEX);
				else{
					if(!Conf->csrv->url)
						Conf->csrv->url=strdup(cmd);
					else
					if(!Conf->csrv->alias)
						Conf->csrv->alias=strdup(cmd);
					else{
						Conf->errcode=1;
						sprintf(Conf->errstr,"'%s' line %d: too many argiments: '%s'",conf_name,i,cmd);
						return(Conf);
					}
				}
			}
			if(UdmAddServer(Conf,Conf->csrv,load_flags,match_type)){
				char * s_err;
				Conf->errcode=1;
				s_err=strdup(Conf->errstr);
				sprintf(Conf->errstr,"'%s' line %d: %s",conf_name,i,s_err);
				free(s_err);
				UDM_FREE(Conf->csrv->url);
				return(Conf);
			}
			UDM_FREE(Conf->csrv->url);
			UDM_FREE(Conf->csrv->alias);
			/* Restore follow value */
			Conf->csrv->follow=follow;
			
		}else
		if(!UDM_STRNCASECMP(str,"Follow")){
			if(sscanf(str+6,"%s",str1)!=1)
				continue;
			Conf->csrv->follow=UdmFollowType(str1);
			if(Conf->csrv->follow==UDM_FOLLOW_UNKNOWN){
				Conf->errcode=1;
				sprintf(Conf->errstr,"'%s' line %d: Unknown follow type: %s",conf_name,i,str1);
				return(Conf);
			}
		}
		else	IF_BOOL_PARAM(str,"Robots",Conf->csrv->use_robots,1)
		else	IF_BOOL_PARAM(str,"DeleteBad",Conf->csrv->deletebad,0)
		else	IF_BOOL_PARAM(str,"DeleteNoServer",Conf->csrv->delete_no_server,1)
		else	IF_BOOL_PARAM(str,"ForceIISCharset1251",Conf->force1251,0)
		else	IF_BOOL_PARAM(str,"UseCRC32UrlId",Conf->use_crc32_url_id,0)
		else	IF_BOOL_PARAM(str,"DetectClones",Conf->csrv->use_clones,1)
		else	IF_BOOL_PARAM(str,"CheckMp3Tag",Conf->csrv->check_mp3_tag,0)
		else	IF_BOOL_PARAM(str,"IndexMP3TagOnly",Conf->csrv->check_only_mp3_tag,0)
		else	IF_BOOL_PARAM(str,"Index",Conf->csrv->index,1)
		else	IF_BOOL_PARAM(str,"NewsExtensions",Conf->use_newsext,0)
		else	IF_BOOL_PARAM(str,"CrossWords",Conf->use_crossword,0)
		
		else	IF_NUM_PARAM(str,"MaxNetErrors",Conf->csrv->max_net_errors,UDM_MAXNETERRORS)
		else	IF_TIME_PARAM(str,"NetErrorDelayTime",Conf->csrv->net_error_delay_time,UDM_DEFAULT_NET_ERROR_DELAY_TIME)
		else	IF_TIME_PARAM(str,"ReadTimeOut",Conf->csrv->read_timeout,UDM_READ_TIMEOUT)
		else	IF_TIME_PARAM(str,"DocTimeOut",Conf->csrv->doc_timeout,UDM_DOC_TIMEOUT)
		else	IF_TIME_PARAM(str,"Period",Conf->csrv->period,UDM_DEFAULT_REINDEX_TIME)
		else	IF_NUM_PARAM(str,"MaxHops",Conf->csrv->maxhops,0)
		
		else	IF_NUM_PARAM(str,"IspellCorrectFactor",Conf->csrv->correct_factor,1)
		else	IF_NUM_PARAM(str,"IspellIncorrectFactor",Conf->csrv->incorrect_factor,1)
		else	IF_NUM_PARAM(str,"NumberFactor",Conf->csrv->number_factor,1)
		else	IF_NUM_PARAM(str,"AlnumFactor",Conf->csrv->alnum_factor,1)
		else	IF_NUM_PARAM(str,"MinWordLength",Conf->min_word_len,1)
		else	IF_NUM_PARAM(str,"MaxWordLength",Conf->max_word_len,32)
		else	IF_NUM_PARAM(str,"MaxDocSize",Conf->max_doc_size,UDM_MAXDOCSIZE)
		else  /* Should be after URLWeight etc */
		if(!UDM_STRNCASECMP(str,"URL")){
			if(load_flags&UDM_FLAG_ADD_SERV){
				if(sscanf(str+3,"%s",str1)&&strcmp(str1,"")){
					char al[UDM_URLSIZE*2];
					UDM_SERVER * Srv;
					
					if((Srv=UdmFindServer(Conf,str1,al))){
						UDM_DOCUMENT Href;
						
						Href.url=str1;
						Href.stored=0;
						Href.referrer=0;
						Href.tag=Srv->tag;
						Href.category=Srv->category;
						UdmAddHref(&Conf->HrefList,&Href);
					}
				}
			}
		}else
		if(!UDM_STRNCASECMP(str,"Affix")){
			char lang[33];
			char charset[41];
			if(load_flags&&UDM_FLAG_SPELL){
				if(3==sscanf(str+5,"%32s%40s%s",lang,charset,str1)){
					if(*str1=='/')strcpy(str,str1);
					else	sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
					if(UdmImportAffixes(Conf,lang,charset,str)){
						Conf->errcode=1;
						sprintf(Conf->errstr,"Can't load affix :%s",str);
						return(Conf);
					}
				}
				else {
					Conf->errcode=1;
					sprintf(Conf->errstr,"Error in config file '%s' line %d:%s",conf_name,i,str);
					return(Conf);
				}
			}
		}else
		if(!UDM_STRNCASECMP(str,"Spell")){
			char lang[33];
			char charset[41];
			if(load_flags&&UDM_FLAG_SPELL){
				if(3==sscanf(str+5,"%32s%40s%s",lang,charset,str1)){
					if(*str1=='/')strcpy(str,str1);
					else	sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
					/* -----  0 or 1 ?  */
					if(UdmImportDictionary(Conf,lang,charset,str,0,"")){
					        Conf->errcode=1;
					        sprintf(Conf->errstr,"Can't load dictionary :%s",str);
						return(Conf);
					}
				} else {
					Conf->errcode=1;
					sprintf(Conf->errstr,"Error in config file '%s' line %d:%s",conf_name,i,str);
					return(Conf);
				}
			}
		}else
		if(!UDM_STRNCASECMP(str,"Synonym")){
			char * fname;
			char aname[256]="";
			
			if(load_flags&&UDM_FLAG_SPELL){
				fname=UdmTrim(str+7," \t\r\n");
				if(*fname=='/')strcpy(aname,fname);
				else	sprintf(aname,"%s/%s",UDM_CONF_DIR,fname);
				if(UdmSynonymListLoad(Conf,aname)){
					Conf->errcode=1;
					return(Conf);
				}
			}
		}else
		if(!UDM_STRNCASECMP(str,"Include")){
			if((sscanf(str+7,"%s",str1))&&(config_level<5)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
				UdmLoadConfig(Conf,str,config_level+1,load_flags);
				if(Conf->errcode)return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"MirrorRoot")){
			if((sscanf(str+10,"%s",str1))){
				if(*str1=='/')
					strcpy(str,str1);
				else
					sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
			} else {
				sprintf(str,"%s/mirrors",UDM_CONF_DIR);
			}
			Conf->csrv->mirror_root = strdup(str);
			Conf->csrv->use_mirror=(Conf->csrv->use_mirror>0)?Conf->csrv->use_mirror:0;
		}else
		if(!UDM_STRNCASECMP(str,"MirrorHeadersRoot")){
			if((sscanf(str+17,"%s",str1))){
				if(*str1=='/')
					strcpy(str,str1);
				else
					sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
			} else {
				sprintf(str,"%s/headers",UDM_CONF_DIR);
			}
			Conf->csrv->mirror_headers = strdup(str);
			Conf->csrv->use_mirror=(Conf->csrv->use_mirror>0)?Conf->csrv->use_mirror:0;
		}else
		IF_TIME_PARAM(str,"MirrorPeriod",Conf->csrv->use_mirror,UDM_MIRROR_NO)
		else
		if(!UDM_STRNCASECMP(str,"StopwordFile")){
			if(sscanf(str+12,"%s",str1)){
				if(*str1=='/')strcpy(str,str1);
				else    sprintf(str,"%s/%s",UDM_CONF_DIR,str1);
				if(UdmFileLoadStopList(Conf,str))
					return(Conf);
			} else {
				Conf->errcode=1;
				sprintf(Conf->errstr,"Error in config file '%s' line %d:%s",conf_name,i,str);
				return(Conf);
			}
		}else
		if(!UDM_STRNCASECMP(str,"Section")){
			char * name, * sec, *lt;
			name=UdmGetStrToken(str+7,&lt);
			if(name)sec=UdmGetStrToken(NULL,&lt);
			if(name&&sec){
				int nsec=atoi(sec);
				
				if(!strcasecmp(name,"body")){
					Conf->csrv->body_section=nsec;
				}else
				if(!strcasecmp(name,"title")){
					Conf->csrv->title_section=nsec;
				}else
				if(!strcasecmp(name,"description")){
					Conf->csrv->desc_section=nsec;
				}else
				if(!strcasecmp(name,"keywords")){
					Conf->csrv->kwrd_section=nsec;
				}else
				if(!strcasecmp(name,"crosswords")){
					Conf->csrv->cross_section=nsec;
				}else
				if(!strcasecmp(name,"url:proto")){
					Conf->csrv->proto_section=nsec;
				}else
				if(!strcasecmp(name,"url:host")){
					Conf->csrv->host_section=nsec;
				}else
				if(!strcasecmp(name,"url:path")){
					Conf->csrv->path_section=nsec;
				}else
				if(!strcasecmp(name,"url:file")){
					Conf->csrv->file_section=nsec;
				}else{
					if(!UDM_STRNCASECMP(name,"meta:")){
						UDM_SECTION S;
						bzero(&S,sizeof(S));
						S.name=name+5;
						S.section=nsec;
						S.type=UDM_SECTION_META;
						UdmSectionListAdd(&Conf->SectionList,&S);
					}else
					if(!UDM_STRNCASECMP(name,"header:")){
						UDM_SECTION S;
						bzero(&S,sizeof(S));
						S.name=name+7;
						S.section=nsec;
						S.type=UDM_SECTION_HEADER;
						UdmSectionListAdd(&Conf->SectionList,&S);
					}else{
						snprintf(Conf->errstr,sizeof(Conf->errstr),"Unknown section '%s' in config file '%s' line %d",name,conf_name,i);
						Conf->errcode=1;
						return(Conf);
					}
				}
			}else{
				snprintf(Conf->errstr,sizeof(Conf->errstr),"Too few arguments in config file '%s' line %d",conf_name,i);
				Conf->errcode=1;
				return(Conf);
			}
		}else{
			Conf->errcode=1;
			sprintf(Conf->errstr,"Error in config file '%s' line %d: %s",conf_name,i,str);
			return(Conf);
		}
		str[0]=0;
	}
	fclose(config);
	/* On level0 : Free some variables, prepare others, etc */
	if(config_level==0){

		/* Add one virtual server if we want FollowOutside */
		/* or DeleteNoServer no	*/
		UDM_FREE(Conf->csrv->url);
		if((Conf->csrv->follow==UDM_FOLLOW_WORLD)||(!Conf->csrv->delete_no_server)){
			UdmAddServer(Conf,Conf->csrv,load_flags,UDM_SERVER_SUBSTR|UDM_SERVER_MATCH);
		}

		/* Removed due to new Realm command             */
		/* Servers/Realms now found in appearence order */
		/* UdmSortServers(Conf);                        */

		if (Conf->csrv->regexp) regfree(Conf->csrv->regexp);
		UDM_FREE(Conf->csrv->url);
		UDM_FREE(Conf->csrv->alias);
		UDM_FREE(Conf->csrv->charset);
		UDM_FREE(Conf->csrv->category);
		UDM_FREE(Conf->csrv->tag);
		UDM_FREE(Conf->csrv->proxy);
		UDM_FREE(Conf->csrv->basic_auth);
		UDM_FREE(Conf->csrv->proxy_basic_auth);
		UDM_FREE(Conf->csrv->htdb_list);
		UDM_FREE(Conf->csrv->htdb_doc);
		UDM_FREE(Conf->csrv->user);
		UDM_FREE(Conf->csrv->passwd);
		UDM_FREE(Conf->csrv->mirror_root);
		UDM_FREE(Conf->csrv->mirror_headers);
		UDM_FREE(Conf->csrv);

		/* Sort Aliases for binary search            */
		/* This is incorrect - aliases must          */
		/* be found in the order of their appearance */
		/* UdmSortAliases(Conf); */

		/* Sort ispell dictionay if it has been loaded */
		
		if(Conf->nspell) {
			UdmSortDictionary(Conf);
			UdmSortAffixes(Conf);
		}
		
		/* Sort synonyms */
		UdmSynonymListSort(&Conf->SynList);
	}
	return(0);
}
