#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "udm_common.h"
#include "udm_spell.h"
#include "udm_db.h"
#include "udm_hrefs.h"
#include "udm_utils.h"
#include "udm_xmalloc.h"
#include "udm_sgml.h"

/* TUNE */
/* Max URLs in cache: 4K URLs will use about 200K of RAM         */
/* This should be a configurable parameter but we'll use 4K now  */

#define MAXHSIZE	1023*4
#define HSIZE		256	/* Length of buffer increment  */
#define RESORT_HREFS	256	/* Max length of unsorted part */


/* Function to sort URLs in alphabetic order */
static int cmphrefs(const void * v1, const void * v2){
	return(strcmp(((const UDM_HREF*)v1)->url,((const UDM_HREF*)v2)->url));
}


__INDLIB__ int UdmAddHref(UDM_HREFLIST * HrefList,UDM_DOCUMENT * Href){
	int l,r,c,res;
	size_t i,len;
	char ehref[UDM_URLSIZE];
	
	/* Don't add empty or too long link */
	len=strlen(Href->url);
	if((len<1)||(len>UDM_URLSIZE-1))return(0);

	strcpy(ehref,Href->url);
	UdmTrim(ehref," \t\r\n");
	UdmStrRemoveChars(ehref,"\t\r\n");
	UdmUnescapeSgmlStr(ehref);

	/* Find current URL in sorted part of list */
	l=0;r=HrefList->shrefs-1;
	while(l<=r){
		c=(l+r)/2;
		if(!(res=strcmp(HrefList->Href[c].url,ehref))){
			HrefList->Href[c].stored|=Href->stored;
			return(0);
		}
		if(res<0)
			l=c+1;
		else
			r=c-1;
	}
	/* Find in unsorted part */
	for(i=HrefList->shrefs;i<HrefList->nhrefs;i++){
		if(!strcmp(HrefList->Href[i].url,ehref)){
			HrefList->Href[i].stored|=Href->stored;
			return(0);
		}
	}
	if(HrefList->nhrefs>=HrefList->mhrefs){
		if(HrefList->mhrefs){
			HrefList->mhrefs+=HSIZE;
			HrefList->Href=(UDM_HREF *)UdmXrealloc(HrefList->Href,HrefList->mhrefs*sizeof(UDM_HREF));
		}else{
			HrefList->mhrefs=HSIZE;
			HrefList->Href=(UDM_HREF *)UdmXmalloc(HrefList->mhrefs*sizeof(UDM_HREF));
		}
	}
	HrefList->Href[HrefList->nhrefs].url=strdup(ehref);
	HrefList->Href[HrefList->nhrefs].referrer=Href->referrer;
	HrefList->Href[HrefList->nhrefs].hops=Href->hops;
	HrefList->Href[HrefList->nhrefs].stored=Href->stored;
	HrefList->Href[HrefList->nhrefs].tag=Href->tag?strdup(Href->tag):NULL;
	HrefList->Href[HrefList->nhrefs].category=Href->category?strdup(Href->category):NULL;
	HrefList->nhrefs++;

	/* Sort unsorted part */
	if((HrefList->nhrefs-HrefList->shrefs)>RESORT_HREFS){
		qsort(HrefList->Href,HrefList->nhrefs,sizeof(UDM_HREF),cmphrefs);
		/* Remember count of sorted URLs  */
		HrefList->shrefs=HrefList->nhrefs;
		/* Count of stored URLs became 0  */
		HrefList->dhrefs=0;
	}
	return(1);
}
extern __INDLIB__ void UdmFreeHrefList(UDM_HREFLIST * HrefList){
	size_t i;
	
	for(i=0;i<HrefList->nhrefs;i++){
		UDM_FREE(HrefList->Href[i].url);
		UDM_FREE(HrefList->Href[i].tag);
		UDM_FREE(HrefList->Href[i].category);
	}
	HrefList->nhrefs=0; 
	HrefList->mhrefs=0;
	HrefList->shrefs=0;
	HrefList->dhrefs=0;
	UDM_FREE(HrefList->Href);
}
extern __INDLIB__ int UdmStoreHrefs(UDM_AGENT * Indexer){
	size_t i;
	int added=0;

	for(i=Indexer->Conf->HrefList.dhrefs;i<Indexer->Conf->HrefList.nhrefs;i++){
		if(!(Indexer->Conf->HrefList.Href[i].stored)){
			if(strlen(Indexer->Conf->HrefList.Href[i].url)<=UDM_URLSIZE){
				UDM_DOCUMENT Doc;
				
				Doc.url=Indexer->Conf->HrefList.Href[i].url;
				Doc.referrer=Indexer->Conf->HrefList.Href[i].referrer;
				Doc.hops=Indexer->Conf->HrefList.Href[i].hops;
				Doc.tag=Indexer->Conf->HrefList.Href[i].tag;
				Doc.category=Indexer->Conf->HrefList.Href[i].category;
				if(IND_OK!=UdmURLAction(Indexer,&Doc,UDM_URL_ACTION_ADD)){
					return(added);
				}
			}
			Indexer->Conf->HrefList.Href[i].stored=1;
			added++;
		}
	}
	/* Remember last stored URL num */
	/* Note that it will became 0   */
	/* after next sort in AddUrl    */
	Indexer->Conf->HrefList.dhrefs=Indexer->Conf->HrefList.nhrefs;

	/* We should not free URL list with onw database */
	/* to avoid double indexing of the same document */
	/* So, do it if compiled with SQL only           */

#ifndef HAVE_FILES
	if(Indexer->Conf->HrefList.nhrefs>MAXHSIZE)
		UdmFreeHrefList(&Indexer->Conf->HrefList);
#endif
	return(added);
}

__INDLIB__ UDM_HREFLIST * UdmInitHrefList(UDM_HREFLIST * Hrefs){
	bzero(Hrefs,sizeof(*Hrefs));
	return(Hrefs);
}
