#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "udm_common.h"
#include "udm_robots.h"
#include "udm_utils.h"
#include "udm_mutex.h"

UDM_ROBOT* UdmGetRobot(UDM_ROBOTS *Robots,char *hostinfo){
	size_t i; 

	/* FIXMI: <hash> or <btree> or <bsearch with partial sorting> or <?> */
	for(i=0;i<Robots->nrobots;i++){
		if(!strcmp(hostinfo,Robots->robots[i].hostinfo)){
			return(&Robots->robots[i]);
		}
	}
	return(NULL);
}

static UDM_ROBOT* DeleteRobotRules(UDM_ROBOTS *Robots,char *hostinfo){
	UDM_ROBOT *robot;
	size_t i;
	
	if((robot=UdmGetRobot(Robots, hostinfo))!=NULL){
		for(i=0;i<robot->nrules;i++){
			UDM_FREE(robot->rules[i].path);
		}
		robot->nrules=0;
		UDM_FREE(robot->rules);
		return(robot);
	}
	return(NULL);
}

static UDM_ROBOT* AddEmptyRobot(UDM_ROBOTS *Robots,char *hostinfo){
	Robots->nrobots++;
	Robots->robots=(UDM_ROBOT*)realloc(Robots->robots, Robots->nrobots*sizeof(UDM_ROBOT));
	if(Robots->robots==NULL) return(NULL);

	bzero(&Robots->robots[Robots->nrobots-1],sizeof(UDM_ROBOT));
	Robots->robots[Robots->nrobots-1].hostinfo=strdup(hostinfo);
	
	return(&Robots->robots[Robots->nrobots-1]);
}

static int AddRobotRule(UDM_ROBOT *robot,int cmd,char *path){
	robot->nrules++;
	robot->rules=(UDM_ROBOT_RULE*)realloc(robot->rules,robot->nrules*sizeof(UDM_ROBOT_RULE));
	if(robot->rules==NULL) return(1);
	
	robot->rules[robot->nrules-1].cmd=cmd;
	robot->rules[robot->nrules-1].path=strdup(path);
	
	return(0);
}

int UdmFreeRobots(UDM_ROBOTS *Robots){
	size_t i,j; 
	
	if(!Robots->nrobots){
		return(0);
	}
	for(i=0;i<Robots->nrobots;i++){
		for(j=0;j<Robots->robots[i].nrules;j++){
			UDM_FREE(Robots->robots[i].rules[j].path);
		}
		UDM_FREE(Robots->robots[i].hostinfo);
		UDM_FREE(Robots->robots[i].rules);
	}
	UDM_FREE(Robots->robots);
	Robots->nrobots=0;
	return(0);
}

UDM_ROBOT_RULE* UdmFindRobotRule(UDM_ROBOTS *Robots,UDM_URL *URL){
	UDM_ROBOT *robot;
	size_t j;
	
	if((robot=UdmGetRobot(Robots, URL->hostinfo))!=NULL){
		for(j=0;j<robot->nrules;j++){
			/* FIXME: compare full URL */
			if(!strncmp(URL->path,robot->rules[j].path,strlen(robot->rules[j].path))){
				if(robot->rules[j].cmd==UDM_ROBOT_ALLOW_CMD) return(NULL);
				else return(&robot->rules[j]);
			}
		}
	}
	return(NULL);
}

int UdmParseRobot(UDM_ROBOTS *Robots,char *content,char *hostinfo){
	UDM_ROBOT *robot;
	int myrule=0;
	char *s,*e,*lt;

	robot=DeleteRobotRules(Robots,hostinfo);
	if(robot==NULL) robot=AddEmptyRobot(Robots,hostinfo);
	if(robot==NULL) return(IND_ERROR);
	
	if(content==NULL) return(IND_OK);

	s=UdmGetToken(content,"\r\n",&lt);
	while(s){
		if(*s=='#'){
		}else
		if(!(UDM_STRNCASECMP(s,"User-Agent"))){
			if(strstr(s,UDM_USER_AGENT))myrule=1;
			else
			if(strstr(s,"*"))myrule=1;
			else
				myrule=0;
		}else
		if((!(UDM_STRNCASECMP(s,"Disallow")))&&(myrule)){
			if((e=strchr(s+9,'#')))*e=0;
			e=s+9;UDM_SKIP(e," \t");s=e;
			UDM_SKIPN(e," \t");*e=0;
			if(*s){
				if(AddRobotRule(robot,UDM_ROBOT_DISALLOW_CMD,s))
					return(IND_ERROR);
			}
		}else
		if((!(UDM_STRNCASECMP(s,"Allow")))&&(myrule)){
			if((e=strchr(s+6,'#')))*e=0;
			e=s+6;UDM_SKIP(e," \t");s=e;
			UDM_SKIPN(e," \t");*e=0;
			if(*s){
				if(AddRobotRule(robot,UDM_ROBOT_ALLOW_CMD,s))
					return(IND_ERROR);
			}
		}
		s=UdmGetToken(NULL,"\n\r",&lt);
	}
	return(IND_OK);
}
