#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_proto.h"
#include "udm_services.h"
#include "udm_agent.h"
#include "udm_db.h"
#include "udm_doc.h"
#include "udm_sdp.h"
#include "udm_xmalloc.h"
#include "udm_searchtool.h"
#include "udm_vars.h"

/*
#define DEBUG_SDP
*/

ssize_t UdmSearchdSendData(UDM_SDCLIENT *cl,const void * data,size_t len){
	ssize_t nsent=0;
	while(nsent<len){
		ssize_t n;
		n=send(cl->fd,((const char*)data)+nsent,len,0);
		if(!n)return(nsent);
		nsent+=n;
	}
	return nsent;
}

ssize_t UdmSearchdRecvData(UDM_SDCLIENT *cl,void * data,size_t len){
	ssize_t nrecv=0;
	while(nrecv<len){
		ssize_t n;
		n=recv(cl->fd,((char*)data)+nrecv,len,0);
		if(!n)return(nrecv);
		nrecv+=n;
	}
	return nrecv;
}

ssize_t UdmSearchdSendPacket(UDM_SDCLIENT *cl,const UDM_SEARCHD_PACKET_HEADER * hdr,const void * data){
	ssize_t nsent;
	
	nsent=UdmSearchdSendData(cl,hdr,sizeof(*hdr));
	if(nsent!=sizeof(*hdr))
		return(nsent);
	if(data){
		nsent+=UdmSearchdSendData(cl,data,hdr->len);
	}
	return nsent;
}


#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long) -1)
#endif

static int open_host(char *hostname,int port, int timeout)
{
	int net;
	struct hostent *host;
	struct sockaddr_in sa_in;

	bzero((char*)&sa_in,sizeof(sa_in));

	if (port){
		sa_in.sin_port= htons((u_short)port);
	}else{
		return(UDM_NET_ERROR);
	}

	if ((sa_in.sin_addr.s_addr=inet_addr(hostname)) != INADDR_NONE){
		sa_in.sin_family=AF_INET;
	}else{
		host=gethostbyname(hostname);
		if (host){
			sa_in.sin_family=host->h_addrtype;
			memcpy(&sa_in.sin_addr, host->h_addr, (size_t)host->h_length);
		}else{
			return(UDM_NET_CANT_RESOLVE);
		}
	}
	net=socket(AF_INET, SOCK_STREAM, 0);

	if(connect(net, (struct sockaddr *)&sa_in, sizeof (sa_in)))
		return(UDM_NET_CANT_CONNECT);

	return(net);
}

int UdmSearchdConnect(UDM_AGENT * query,UDM_SDCLIENT *cl){
	int res=IND_OK;
	int port=UDM_SEARCHD_PORT;
	char host[128];
	char * cport;
	
	snprintf(host,sizeof(host)-1,"%s",cl->addr?cl->addr:"localhost");
	if((cport=strchr(host,':'))){
		*cport='\0';
		port=atoi(cport+1);
	}
	cl->fd=open_host(host,port,0);
	if(cl->fd<=0){
		cl->fd=0;
		sprintf(UdmAgentErrorMsg(query),"Can't connect to searchd at '%s:%d'",host,port);
		UdmSetDBErrorCode(query->db,1);
		res=IND_ERROR;
	}
	return res;
}

static void UdmSearchdJustClose(UDM_AGENT * query,UDM_SDCLIENT * cl){
	closesocket(cl->fd);
	cl->fd=0;
}

void UdmSearchdClose(UDM_AGENT * query,UDM_SDCLIENT * cl){
	UDM_SEARCHD_PACKET_HEADER hdr;
	ssize_t nsent;
	
	if(cl->fd>0){
		/* Send goodbye*/
		hdr.cmd=UDM_SEARCHD_CMD_GOODBYE;
		hdr.len=0;
		nsent=UdmSearchdSendPacket(cl,&hdr,NULL);
		closesocket(cl->fd);
		cl->fd=0;
	}
}

static int UdmResAddDocInfoSearchd(UDM_AGENT * query,UDM_SDCLIENT * cl,UDM_RESULT * Res,size_t clnum){
	UDM_SEARCHD_PACKET_HEADER hdr;
	char * msg=NULL;
	size_t i,num=0,curnum=0;
	int * url_id, done=0;
	ssize_t nsent,nrecv;
	char * dinfo=NULL;

	/* Count my own documents, skip others */
	for(i=0;i<Res->num_rows;i++){
		if(Res->Doc[i].sdnum==clnum)
			num++;
	}
	if(!num)return(IND_OK);
	
	if(cl->fd<=0){
		int res;
		if(IND_OK!=(res=UdmSearchdConnect(query,cl))){
			return res;
		}
	}

	url_id=(int*)malloc(num*sizeof(*url_id));
	for(i=0;i<Res->num_rows;i++){
		if(Res->Doc[i].sdnum==clnum){
			url_id[curnum]=Res->Doc[i].url_id;
			curnum++;
		}
	}
	hdr.cmd=UDM_SEARCHD_CMD_DOCINFO;
	hdr.len=num*sizeof(*url_id);

	nsent=UdmSearchdSendPacket(cl,&hdr,url_id);

	while(!done){
		char * tok, * lt;
		nrecv=UdmSearchdRecvData(cl,&hdr,sizeof(hdr));

		if(nrecv!=sizeof(hdr)){
			UdmSearchdJustClose(query,cl);
			sprintf(UdmAgentErrorMsg(query),"Received incomplete header from searchd (%d bytes)",(int)nrecv);
			UdmSetDBErrorCode(query->db,1);
			return(IND_ERROR);
			break;
		}else{
#ifdef DEBUG_SDP
			printf("Received header cmd=%d len=%d\n",hdr.cmd,hdr.len);
#endif
		}
		switch(hdr.cmd){
			case UDM_SEARCHD_CMD_ERROR:
				msg=(char*)malloc(hdr.len+1); /* Add checking */
				nrecv=UdmSearchdRecvData(cl,msg,hdr.len);
				msg[nrecv]='\0';
				sprintf(UdmAgentErrorMsg(query),"Searchd error: '%s'",msg);
				UdmSetDBErrorCode(query->db,1);
				free(msg);
				done=1;
				break;
			case UDM_SEARCHD_CMD_MESSAGE:
				msg=(char*)malloc(hdr.len+1); /* Add checking */
				nrecv=UdmSearchdRecvData(cl,msg,hdr.len);
				msg[nrecv]='\0';
#ifdef DEBUG_SDP
				printf("Message from searchd: '%s'\n",msg);
#endif
				free(msg);
				break;
			case UDM_SEARCHD_CMD_DOCINFO:
				dinfo=(char*)malloc(hdr.len+1);
				nrecv=UdmSearchdRecvData(cl,dinfo,hdr.len);
				dinfo[hdr.len]='\0';
#ifdef DEBUG_SDP
				printf("Received DOCINFO size=%d buf=%s\n",hdr.len,dinfo);
#endif				
				tok=UdmGetToken(dinfo,"\r\n",&lt);
				
				while(tok){
					int id;
					if(!strncasecmp(tok,"DI=",3)){
						id=atoi(tok+3);
						
						for(i=0;i<Res->num_rows;i++){
							if(Res->Doc[i].url_id==id){
								Res->Doc[i].textbuf=strdup(tok);
								UdmDocFromTextBuf(&Res->Doc[i]);
								UDM_FREE(Res->Doc[i].textbuf);
								break;
							}
						}
					}
					tok=UdmGetToken(NULL,"\r\n",&lt);
				}
				free(dinfo);
				done=1;
				break;
			default:
				sprintf(UdmAgentErrorMsg(query),"Unknown searchd response: cmd=%d len=%d",hdr.cmd,hdr.len);
				UdmSetDBErrorCode(query->db,1);
				done=1;
				break;
		}
	}
	return IND_OK;
}

static int UdmSearchdSendWordRequest(UDM_AGENT * query,UDM_SDCLIENT * cl,const char * q){
	UDM_SEARCHD_PACKET_HEADER hdr;
	int res;
	ssize_t nsent;
	
	if(cl->fd<=0){
		if(IND_OK!=(res=UdmSearchdConnect(query,cl))){
			return res;
		}
	}
	hdr.cmd=UDM_SEARCHD_CMD_WORDS;
	hdr.len=strlen(q);
	nsent=UdmSearchdSendPacket(cl,&hdr,q);
	
	return IND_OK;
}


static UDM_URL_CRD *UdmSearchdGetWordResponse(UDM_AGENT * query,UDM_SDCLIENT * cl){
	UDM_URL_CRD * wrd=NULL;
	UDM_SEARCHD_PACKET_HEADER hdr;
	ssize_t nrecv;
	char * msg;
	int done=0;

	query->total_found=0;

	if(cl->fd<=0){
		/* Add error: incorrect commands order */
		return NULL;
	}

	while(!done){
		nrecv=UdmSearchdRecvData(cl,&hdr,sizeof(hdr));
		if(nrecv!=sizeof(hdr)){
			UdmSearchdJustClose(query,cl);
			sprintf(UdmAgentErrorMsg(query),"Received incomplete header from searchd (%d bytes)",(int)nrecv);
			UdmSetDBErrorCode(query->db,1);
			return(NULL);
		}else{
#ifdef DEBUG_SDP
			printf("Received header cmd=%d len=%d\n",hdr.cmd,hdr.len);
#endif
		}

		switch(hdr.cmd){
			case UDM_SEARCHD_CMD_ERROR:
				msg=(char*)malloc(hdr.len+1); /* Add checking */
				nrecv=UdmSearchdRecvData(cl,msg,hdr.len);
				msg[nrecv]='\0';
				sprintf(UdmAgentErrorMsg(query),"Searchd error: '%s'",msg);
				UdmSetDBErrorCode(query->db,1);
				free(msg);
				done=1;
				break;
			case UDM_SEARCHD_CMD_MESSAGE:
				msg=(char*)malloc(hdr.len+1); /* Add checking */
				nrecv=UdmSearchdRecvData(cl,msg,hdr.len);
				msg[nrecv]='\0';
#ifdef DEBUG_SDP
				printf("Message from searchd: '%s'\n",msg);
#endif
				free(msg);
				break;
			case UDM_SEARCHD_CMD_WORDS:
				UDM_FREE(wrd);
				wrd=(UDM_URL_CRD*)malloc(hdr.len); /* FIXME: add checking */
				nrecv=UdmSearchdRecvData(cl,wrd,hdr.len);
				query->total_found=hdr.len/sizeof(*wrd);
#ifdef DEBUG_SDP
				printf("Received words size=%d nwrd=%d\n",hdr.len,query->total_found);
#endif
				done=1;
				break;
			default:
				sprintf(UdmAgentErrorMsg(query),"Unknown searchd response: cmd=%d len=%d",hdr.cmd,hdr.len);
				UdmSetDBErrorCode(query->db,1);
				done=1;
				break;
		}
	}
	return wrd;
}

UDM_RESULT * UdmSearchdFind(UDM_AGENT * query,const char * q){
	UDM_URL_CRD *wrdX[100];
	UDM_URL_CRD *wrd,*curwrd;
	UDM_RESULT * Res;
	unsigned long ticks;
	size_t i;
	size_t nwrd=0;
	size_t nwrdX[100];

	ticks=UdmStartTimer();

	for(i=0;i<query->sdcl.nclients;i++){
		int rc;
		if(IND_OK!=(rc=UdmSearchdSendWordRequest(query,&query->sdcl.Clients[i],q)))
			return(NULL);
	}
	for(i=0;i<query->sdcl.nclients;i++){
		wrdX[i]=UdmSearchdGetWordResponse(query,&query->sdcl.Clients[i]);
#ifdef DEBUG_SDP
		printf("WRD[%d] total_found=%d\n",i,query->total_found);
#endif
		nwrdX[i]=query->total_found;
		nwrd+=query->total_found;
	}
	
	curwrd=wrd=(UDM_URL_CRD*)malloc(sizeof(*wrd)*nwrd);
	for(i=0;i<query->sdcl.nclients;i++){
		if(wrdX[i]){
			size_t j;
			/* Set machine number */
			for(j=0;j<nwrdX[i];j++){
				wrdX[i][j].coord=(wrdX[i][j].coord&0xFFFFFF00)+i;
			}
			memcpy(curwrd,wrdX[i],sizeof(*curwrd)*nwrdX[i]);
			curwrd+=nwrdX[i];
			free(wrdX[i]);
		}
	}
	UdmSortSearchWordsByWeight(wrd,nwrd);

	query->total_found=nwrd;

	/* Allocate result */
	Res=(UDM_RESULT*)UdmXmalloc(sizeof(UDM_RESULT));
	Res->num_rows=query->loaded_nwords;
	Res->total_found=query->total_found;
	strncpy(Res->wordinfo,query->wordinfo,sizeof(Res->wordinfo)-1);

	/* Track query if required */
	/*
	if((query->track_mode)&UDM_TRACK_QUERIES){
		if(IND_OK!=UdmTrack(query,words))
			return NULL;
	}
	*/

	/* Return if no words found   */
	if(!query->total_found){
		Res->work_time=UdmStartTimer()-ticks;
		return(Res);
	}

	Res->first=query->page_number*query->page_size;	
	if(Res->first>=query->total_found)Res->first=query->total_found-1;
	if((Res->first+query->page_size)>query->total_found){
		Res->num_rows=query->total_found-Res->first;
	}else{
		Res->num_rows=query->page_size;
	}
	Res->last=Res->first+Res->num_rows-1;

#ifdef DEBUG_SEARCH
	fprintf(stderr,"First: %d Found: %d Display: %d\n",first,query->total_found,display_docs);
#endif

#ifdef DEBUG_SEARCH
	fprintf(stderr,"Start URL stuff\n");
	debug_search_ticks=UdmStartTimer();
#endif

	/* Allocate an array for documents information */
	Res->Doc=(UDM_DOCUMENT*)UdmXmalloc(sizeof(UDM_DOCUMENT)*(Res->num_rows));

	/* Copy url_id and coord to result */
	for(i=0;i<Res->num_rows;i++){
		UdmInitDoc(&Res->Doc[i]);
		Res->Doc[i].url_id=wrd[i+Res->first].url_id;
		Res->Doc[i].rating=wrd[i+Res->first].coord;
		Res->Doc[i].order=i+Res->first+1;
		Res->Doc[i].sdnum=(Res->Doc[i].rating&0xFF);
	}
	/* Free words array */
	UDM_FREE(wrd);

	for(i=0;i<query->sdcl.nclients;i++){
		if(IND_OK!=UdmResAddDocInfoSearchd(query,&query->sdcl.Clients[i],Res,i));
	}
	
	/* Load documents information */
/*
	if(IND_OK!=UdmResAddDocInfo(query,Res)){
		UdmFreeResult(Res);
		return NULL;
	}
*/

#ifdef DEBUG_SEARCH
	debug_search_ticks=UdmStartTimer()-debug_search_ticks;
	fprintf(stderr,"Stop URL stuff:\t\t%.2f\n",(float)debug_search_ticks/1000);
#endif

	/* first and last begins from 0, make it begin from 1 */
	Res->first++;
	Res->last++;
	Res->work_time=UdmStartTimer()-ticks;
	return(Res);
}

UDM_RESULT * UdmCloneListSearchd(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	return NULL;
}


/********************************************************/


UDM_SDCLLIST * UdmSDClientListInit(UDM_SDCLLIST * List){
	bzero(List,sizeof(*List));
	return(List);
}

size_t UdmSDCLientListAdd(UDM_SDCLLIST * List,const char * addr){
	List->Clients=(UDM_SDCLIENT*)realloc(List->Clients,(List->nclients+1)*sizeof(UDM_SDCLIENT));
	List->Clients[List->nclients].addr=strdup(addr);
	List->Clients[List->nclients].fd=0;
	List->nclients++;
	return(List->nclients);
}

void UdmSDClientListFree(UDM_SDCLLIST * List){
	size_t i;
	
	for(i=0;i<List->nclients;i++){
		UDM_FREE(List->Clients[i].addr);
	}
	UDM_FREE(List->Clients);
	UdmSDClientListInit(List);
}
