#include <udm_config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <netdb.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif


#include "udm_common.h"
#include "udm_utils.h"
#include "udm_db.h"
#include "udm_agent.h"
#include "udm_env.h"
#include "udm_conf.h"
#include "udm_services.h"
#include "udm_sdp.h"
#include "udm_log.h"
#include "udm_xmalloc.h"
#include "udm_doc.h"
#include "udm_db.h"
#include "udm_searchtool.h"
#include "udm_vars.h"


#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long)-1)
#endif

static int  mdone=0;
static char time_pid[100];

static char * time_pid_info(void){
	struct tm * tim;
	time_t t;
	t=time(NULL);
	tim=localtime(&t);
	strftime(time_pid,sizeof(time_pid),"%a %d %T",tim);
	sprintf(time_pid+strlen(time_pid)," [%d]",(int)getpid());
	return(time_pid);
}

/********************* SIG Handlers ****************/
static void sighandler(int sign);
static void init_signals(void){
	/* Set up signals handler*/
	signal(SIGPIPE,sighandler);
	signal(SIGHUP,sighandler);
	signal(SIGINT,sighandler);
	signal(SIGTERM,sighandler);
	signal(SIGCHLD,sighandler);
}

static void sighandler(int sign){
#ifdef UNIONWAIT                                                               
	union wait status;                                                         
#else                                                                          
	int status;                                                                
#endif   
	switch(sign){
		case SIGPIPE:
			fprintf(stderr,"%s Signal %d arrived\n",time_pid_info(),sign);
			fprintf(stderr,"%s FATAL: This should not ever happen! Broken pipe!\n",time_pid_info());
			break;
		case SIGCHLD:
			wait(&status);
			break;
		case SIGHUP:
			fprintf(stderr,"%s Signal %d arrived\n",time_pid_info(),sign);
			break;
		default:
			fprintf(stderr,"%s Signal %d arrived\n",time_pid_info(),sign);
			mdone=1;
	}
	init_signals();
}

/*************************************************************/

static int do_client(UDM_ENV * Conf, UDM_SDCLIENT * cl){
	UDM_SEARCHD_PACKET_HEADER hdr;
	UDM_AGENT * Agent;
	UDM_URL_CRD * wrd;
	UDM_RESULT  * Res;
	char buf[1024]="";
	char words[1024]="";
	const char * q;
	ssize_t nrecv,nsent;
	int verb=-1;
	int done=0;
	
	Agent=UdmAllocAgent(Conf,0,0);
	Agent->Conf->vars=UdmAllocVarList();

	while(!done){
		size_t dlen=0,ndocs,i;
		int * doc_id=NULL;
		char * dinfo=NULL;
		
		UdmLog(Agent,verb,"Waiting for command header");
		nrecv=UdmSearchdRecvData(cl,&hdr,sizeof(hdr));
		if(nrecv!=sizeof(hdr)){
			UdmLog(Agent,verb,"Received incomplete header nrecv=%d",(int)nrecv);
			break;
		}else{
			UdmLog(Agent,verb,"Received header cmd=%d len=%d",hdr.cmd,hdr.len);
		}
		switch(hdr.cmd){
			case UDM_SEARCHD_CMD_DOCINFO:
				ndocs=hdr.len/sizeof(*doc_id);
				UdmLog(Agent,verb,"Received DOCINFO command len=%d ndocs=%d",hdr.len,ndocs);
				doc_id=(int*)malloc((hdr.len+1)*sizeof(*doc_id));
				nrecv=UdmSearchdRecvData(cl,doc_id,hdr.len);
				if(nrecv!=hdr.len){
					UdmLog(Agent,verb,"Received incomplete data nbytes=%d nrecv=%d",hdr.len,(int)nrecv);
					UDM_FREE(doc_id);
					done=1;
					break;
				}
				
				Res=(UDM_RESULT*)UdmXmalloc(sizeof(UDM_RESULT));
				Res->num_rows=ndocs;
				Res->Doc=(UDM_DOCUMENT*)UdmXmalloc(sizeof(UDM_DOCUMENT)*(Res->num_rows));

				for(i=0;i<ndocs;i++){
					UdmInitDoc(&Res->Doc[i]);
					Res->Doc[i].url_id=doc_id[i];
				}
				UDM_FREE(doc_id);
				
				if(IND_OK!=UdmResAddDocInfo(Agent,Res)){
					UdmFreeResult(Res);
					snprintf(buf,sizeof(buf)-1,"%s",UdmDBErrorMsg(Agent->db));
					UdmLog(Agent,verb,"%s",UdmDBErrorMsg(Agent->db));
					hdr.cmd=UDM_SEARCHD_CMD_ERROR;
					hdr.len=strlen(buf);
					nsent=UdmSearchdSendPacket(cl,&hdr,buf);
					done=1;
					break;
				}
				
				dlen=0;
				
				for(i=0;i<Res->num_rows;i++){
					if(Res->Doc[i].url){
						size_t ulen;
						size_t olen;
						
						UdmDocToTextBuf(&Res->Doc[i]);
						if(Res->Doc[i].textbuf){
							ulen=strlen(Res->Doc[i].textbuf);
							olen=dlen;
							dlen=dlen+ulen;
							dinfo=(char*)realloc(dinfo,dlen+1);
							dinfo[olen]='\0';
							strcpy(dinfo+olen,Res->Doc[i].textbuf);
							UDM_FREE(Res->Doc[i].textbuf);
						}else{
							/*
							printf("notextbuf\n");
							*/
						}
					}
				}
				UdmFreeResult(Res);
				if(!dinfo)dinfo=strdup("nodocinfo");
				
				hdr.cmd=UDM_SEARCHD_CMD_DOCINFO;
				hdr.len=strlen(dinfo);
				nsent=UdmSearchdSendPacket(cl,&hdr,dinfo);
				UdmLog(Agent,verb,"Sent doc_info packet %d bytes",(int)nsent);
				UDM_FREE(dinfo);
				
				break;

			case UDM_SEARCHD_CMD_WORDS:
				nrecv=UdmSearchdRecvData(cl,words,hdr.len); /* FIXME: check */
				words[nrecv]='\0';
				UdmLog(Agent,verb,"Received words len=%d words='%s'",(int)nrecv,words);
				
				UdmParseQueryString(Agent,Agent->Conf->vars,words);
				q=UdmFindStrVar(Agent->Conf->vars,"q","");

				wrd=UdmFindWords(Agent,q);
				if(UdmDBErrorCode(Agent->db)){
					snprintf(buf,sizeof(buf)-1,"%s",UdmDBErrorMsg(Agent->db));
					UdmLog(Agent,verb,"%s",UdmDBErrorMsg(Agent->db));
					hdr.cmd=UDM_SEARCHD_CMD_ERROR;
					hdr.len=strlen(buf);
					UdmSearchdSendPacket(cl,&hdr,buf);
					done=1;
					break;
				}

				snprintf(buf,sizeof(buf)-1,"Total_found=%d",Agent->total_found);
				hdr.cmd=UDM_SEARCHD_CMD_MESSAGE;
				hdr.len=strlen(buf);
				nsent=UdmSearchdSendPacket(cl,&hdr,buf);
				UdmLog(Agent,verb,"Sent total_found packet %d bytes buf='%s'",(int)nsent,buf);

				hdr.cmd=UDM_SEARCHD_CMD_WORDS;
				hdr.len=Agent->total_found*sizeof(*wrd);
				nsent=UdmSearchdSendPacket(cl,&hdr,wrd);
				UdmLog(Agent,verb,"Sent words packet %d bytes cmd=%d len=%d nwords=%d",(int)nsent,hdr.cmd,hdr.len,Agent->total_found);
				/*
				for(i=0;i<Agent->total_found;i++){
					snprintf(buf,sizeof(buf)-1,"u=%08X c=%08X",wrd[i].url_id,wrd[i].coord);
					UdmLog(Agent,verb,"%s",buf);
				}
				*/
				break;
			case UDM_SEARCHD_CMD_GOODBYE:
				UdmLog(Agent,verb,"Received goodbye command",hdr.cmd);
				done=1;
				break;
			default:
				UdmLog(Agent,verb,"Unknown command %d",hdr.cmd);
				done=1;
				break;
		}
	}
	close(cl->fd);
	cl->fd=0;
	UdmLog(Agent,verb,"Quit");
	UdmFreeEnv(Agent->Conf);
	UdmFreeAgent(Agent);
	return 0;
}

/*************************************************************/

static void usage(void){

	fprintf(stderr,
"\n\
searchd from %s-%s-%s\n\
http://www.mnogosearch.org (C) 1998-2001, LavTech Corp.\n\
\n\
Usage: searchd [OPTIONS]\n\
\n\
Options are:\n\
  -p port       listen specified port (%d is default)\n\
  -l		do not log to stderr\n\
  -h,-?         print this help page and exit\n\
\n\
Please mail bug reports and suggestions to <general@mnogosearch.org>.\n",
	PACKAGE,VERSION,UDM_DBTYPE,UDM_SEARCHD_PORT);

	return;
}


int main(int argc, char **argv) {
	int nport=UDM_SEARCHD_PORT;
	struct sockaddr_in server_addr;
	int ctl_sock, on=1;
	const char * config_name = UDM_CONF_DIR "/searchd.conf";
	UDM_AGENT * Agent;
	fd_set mask;
	UDM_ENV * Conf=NULL;
	int verb=-1;
	int ch;
	int log2stderr=1;
	int flags=UDM_FLAG_SPELL;

	while ((ch = getopt(argc, argv, "p:l?")) != -1){
		switch (ch) {
			case 'p':
				nport=atoi(optarg);
				break;
			case 'l':
				log2stderr=0;
				break;
			case 'h':
			case '?':
			default:
				usage();
				return 1;
				break;
		}
	}
	argc -= optind;argv += optind;
	if(argc==1)config_name=argv[0];

	init_signals();

	UdmInit(); /* Initialize library */
	Conf = UdmAllocEnv();
	Agent=UdmAllocAgent(Conf,0,UDM_OPEN_MODE_READ);

	UdmLoadConfig(Conf,config_name,0,flags);

	UdmOpenLog("searchd",Conf,log2stderr);

	if(UdmEnvErrCode(Conf)){
		UdmLog(Agent,verb,"%s",UdmEnvErrMsg(Conf));
		UdmFreeAgent(Agent);
		UdmFreeEnv(Conf);
		exit(1);
	}

	UdmLog(Agent,verb,"searchd started with conf '%s'",config_name);
	UdmLog(Agent,verb,"VarDir '%s'",Agent->Conf->vardir);
	UdmLog(Agent,verb,"Affixes: %d, Spells: %d, Synonyms: %d\n",Conf->naffixes,Conf->nspell,Conf->SynList.nsynonyms);
	UdmLog(Agent,verb,"Ready");

	nport = htons((u_short) nport);

	if ((ctl_sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		UdmLog(Agent,verb,"%s socket() error %d",time_pid_info(),errno);
		exit(1);
	}

	if (setsockopt(ctl_sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) != 0){
		UdmLog(Agent,verb,"%s setsockopt() error %d",time_pid_info(),errno);
		exit(1);
	}

	bzero(&server_addr, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	server_addr.sin_port = nport;

	if (bind(ctl_sock, (struct sockaddr *)&server_addr, sizeof(server_addr)) == -1) {
		UdmLog(Agent,verb,"%s bind() error %d %s",time_pid_info(),errno,strerror(errno));
		exit(1);
	}


	/* Backlog 32 is enough */
	if (listen(ctl_sock, 32) == -1) {
		UdmLog(Agent,verb,"%s listen() error %d %s",time_pid_info(),errno,strerror(errno));
		exit(1);
	}

	FD_ZERO(&mask);
	FD_SET(ctl_sock,&mask);

	while(!mdone) {
		pid_t fres;
		int sel;
		struct timeval tval;
		fd_set msk;
		
		tval.tv_sec=300;
		tval.tv_usec=0;
		msk=mask;
		sel=select(16,&msk,0,0,&tval);
		
		if(mdone)break;
		
		if(sel==0)continue;
		if(sel==-1){
			switch(errno){
				case EINTR:	/* Child */
					break;
				default:
					UdmLog(Agent,verb,"FIXME select error %d %s",errno,strerror(errno));
			}
			continue;
		}
		
		if(FD_ISSET(ctl_sock,&msk)){
			int ns;
			struct sockaddr_in client_addr;
			socklen_t addrlen=sizeof(client_addr);
			
			if ((ns = accept(ctl_sock, (struct sockaddr *) &client_addr, &addrlen)) == -1) {
				UdmLog(Agent,verb,"%s accept() error %d %s",time_pid_info(),errno,strerror(errno));
				exit(1);
			}
			UdmLog(Agent,verb,"Connect %s", inet_ntoa(client_addr.sin_addr));
			
			if((fres=fork())==0){
				UDM_SDCLIENT cl;
				
				bzero(&cl,sizeof(cl));
				cl.fd=ns;
				closesocket(ctl_sock);
				do_client(Conf,&cl);
				exit(0);
			}else{
				close(ns);
				if(fres>0){
					
				}else{
					UdmLog(Agent,verb,"fork error %d",fres);
				}
			}
			
		}
	}
	UdmLog(Agent,verb,"Shutdown");
	
	return 0;
}
