#include "udm_config.h"

#ifdef HAVE_SQL

/*
#define DEBUG_SEARCH
*/

/*
#define DEBUG_SQL
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#if (WIN32|WINNT)
#include <time.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "udm_utils.h"
#include "udm_common.h"
#include "udm_spell.h"
#include "udm_robots.h"
#include "udm_mutex.h"
#include "udm_db.h"
#include "udm_unicode.h"
#include "udm_url.h"
#include "udm_log.h"
#include "udm_proto.h"
#include "udm_conf.h"
#include "udm_crc32.h"
#include "udm_xmalloc.h"
#include "udm_cache.h"
#include "udm_indexer.h"
#include "udm_boolean.h"
#include "udm_searchtool.h"
#include "udm_server.h"
#include "udm_stopwords.h"
#include "udm_doc.h"
#include "udm_worddb.h"
#include "udm_vars.h"
#include "udm_sdp.h"
#include "udm_agent.h"

#if (HAVE_MYSQL)
#include "mysql.h"
#define ER_DUP_ENTRY            1062
#define ER_DUP_KEY		1022
#define CR_SERVER_LOST          2013
#define CR_SERVER_GONE_ERROR    2006
#define ER_SERVER_SHUTDOWN      1053
#endif

#if (HAVE_PGSQL)
#include "libpq-fe.h"
#endif

#if HAVE_MSQL
#include "msql.h"
#endif

#if HAVE_IODBC
#include "sql.h"
#include "sqlext.h"
#endif

#if HAVE_EASYSOFT
#include "sql.h"
#include "sqlext.h"
#endif

#if HAVE_VIRT
#include "iodbc.h"
#include "isql.h"
#include "isqlext.h"
#endif

#if HAVE_UNIXODBC
#include "sql.h"
#include "sqlext.h"
#endif

#if HAVE_SAPDB
#include "WINDOWS.H"
#include "sql.h"
#include "sqlext.h"
#endif

#if HAVE_SOLID
#include "cli0cli.h"
#endif

#if HAVE_IBASE
#include "ibase.h"
#endif

#if HAVE_CTLIB
#include <ctpublic.h>
#endif

#if HAVE_ORACLE8
#include "oci.h"
#endif

#if HAVE_ORACLE7
 #include "ocidfn.h"
 #include "oratypes.h"
 #ifdef __STDC__
  #include <ociapr.h>
 #else
  #include <ocikpr.h>
 #endif
#endif



#ifdef HAVE_ORACLE8

#define	MAX_COLS_IN_TABLE	32
#define MAX_BIND_PARAM		3

struct param_struct{
	int out_rec;
	int out_pos[MAX_BIND_PARAM];
	int out_pos_val[MAX_BIND_PARAM][BUF_OUT_SIZE];
};
#endif



/* Multi-dict mode defines */
#define NDICTS	18
#define MINDICT	2
#define MAXDICT	NDICTS
static const int dictlen[NDICTS]={2,2,2,3,4,5,6,7,8,9,10,11,12,16,16,16,16,32};
#define	DICTNUM(l)	((l>16)?dictlen[17]:dictlen[l])

#define	URL_DELETE_CACHE 128
#define	URL_SELECT_CACHE 250
#define URL_LOCK_TIME    4*60*60

#define NDOCS_QUERY "SELECT count(*) FROM url"

typedef struct {
	int nRows;
	int nCols;
	int DBDriver;
	char ** items;

#ifdef HAVE_MYSQL
	MYSQL_RES *mysqlres;
	MYSQL_ROW  mysqlrow;
	size_t    lmysqlrow;
#endif

#ifdef HAVE_PGSQL
	PGresult  *pgsqlres;
#endif

#ifdef HAVE_MSQL
	m_result  *msqlres;
#endif

#ifdef HAVE_ORACLE8
	char	*defbuff[MAX_COLS_IN_TABLE]; /* Buffers for OCIStmtFetch */
	int	col_size[MAX_COLS_IN_TABLE]; /* Size of column */
	sb2	indbuff[MAX_COLS_IN_TABLE][BUF_OUT_SIZE];  /* Indicators for NULLs */
#endif

} UDB_RES;

#define sql_val(x,y,z)	((x)->items[(y)*((x)->nCols)+(z)])
#define sql_value(x,y,z) ((x)?(sql_val((x),y,z)?sql_val((x),y,z):""):"")


typedef struct struct_db {
	char * DBHost;
	char * DBName;
	char * DBUser;
	char * DBPass;
	int    DBPort;
	int    DBType;
	int    DBDriver;
	int    DBSQL_IN;
	int    DBSQL_LIMIT;
	int    DBSQL_GROUP;

#ifdef HAVE_MYSQL
	MYSQL	mysql;
#endif

#ifdef HAVE_PGSQL
	PGconn	*pgsql;
#endif

#ifdef HAVE_MSQL
	int	msql;
#endif

#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT || HAVE_EASYSOFT || HAVE_SAPDB)
	HDBC	hDbc;
	HENV	hEnv;
	HSTMT	hstmt;
#endif

#ifdef HAVE_IBASE
	isc_db_handle	DBH;		/* database handle    */
	ISC_STATUS	status[20];	/* status vector      */
	isc_tr_handle   tr_handle;	/* transaction handle */ 
#endif

#ifdef HAVE_ORACLE7
	Lda_Def     lda;
	ub1 hda[HDA_SIZE];
	Cda_Def     cursor;
#endif

#ifdef HAVE_CTLIB
	CS_CONTEXT	*ctx;
	CS_CONNECTION	*conn;
#endif

#ifdef HAVE_ORACLE8
	OCIEnv     *envhp;
	OCIError   *errhp;
	OCISvcCtx  *svchp;
	OCIStmt	   *stmthp;
	OCIParam   *param;
	OCIDefine  *defb[MAX_COLS_IN_TABLE];
	OCIBind    *bndhp[MAX_BIND_PARAM];
	struct param_struct *par;
#endif

	int connected;
	int res_limit;
	int commit_fl;
	int errcode;
	char errstr[2048];
} DB;


static int udb_free_result(UDB_RES *res){
	int i;
	if(res){
		if(res->items){
			for(i=0;i<res->nCols*res->nRows;i++)
				if(res->items[i])
					free(res->items[i]);
			free(res->items);
		}
	}
	return(0);
}

static void UdmSetDBType(DB * db,const char *str1) {
	db->DBType=UDM_DB_UNK;
	db->DBSQL_IN=0;
	db->DBSQL_LIMIT=0;
	db->DBSQL_GROUP=0;

	if(!UDM_STRNCASECMP(str1,"msql")){
		db->DBType=UDM_DB_MSQL;
	}else
	if(!UDM_STRNCASECMP(str1,"solid")){
		db->DBType=UDM_DB_SOLID;
		db->DBSQL_IN=1;
		db->DBSQL_GROUP=1;
	}else
	if(!UDM_STRNCASECMP(str1,"oracle")){
		db->DBType=UDM_DB_ORACLE;
		db->DBSQL_IN=1;
		db->DBSQL_GROUP=1;
	}else
	if(!UDM_STRNCASECMP(str1,"mssql")){
		db->DBType=UDM_DB_MSSQL;
		db->DBSQL_IN=1;
		db->DBSQL_GROUP=1;
	}else
	if(!UDM_STRNCASECMP(str1,"mysql")){
		db->DBType=UDM_DB_MYSQL;
		db->DBSQL_IN=1;
		db->DBSQL_LIMIT=1;
		db->DBSQL_GROUP=1;
	}else
	if(!UDM_STRNCASECMP(str1,"pgsql")){
		db->DBType=UDM_DB_PGSQL;
		db->DBSQL_IN=1;
		db->DBSQL_LIMIT=1;
		db->DBSQL_GROUP=1;
	}else
	if(!UDM_STRNCASECMP(str1,"ibase")){
		db->DBType=UDM_DB_IBASE;
		db->DBSQL_IN=1;
		db->DBSQL_GROUP=1;
	}else
	if(!UDM_STRNCASECMP(str1,"sapdb")){
		db->DBType=UDM_DB_SAPDB;
		db->DBSQL_IN=1;
		db->DBSQL_GROUP=1;
	}
	db->DBDriver=db->DBType;
}


static int UdmSetDBAddr(void * vdb,const char * addr){
	UDM_URL url;
	char buf[1024];
	char *s;
	int res;
	DB * db=(DB*)vdb;

	strncpy(buf,addr,sizeof(buf));
	buf[sizeof(buf)-1]=0;

	if((res=UdmParseURL(&url,buf)))
		return(res);

	db->DBHost=strdup(url.hostname);
	UdmSetDBType(db,url.schema);

	if(db->DBType==UDM_DB_IBASE){
		/* Ibase is a special case        */
		/* It's database name consists of */
		/* full path and file name        */ 
		db->DBName=(char*)malloc(strlen(url.path)+1+strlen(url.filename));
		sprintf(db->DBName,"%s%s",url.path,url.filename);
	}else{
		db->DBName=strdup(url.path);
		sscanf(url.path,"/%[^/]s",db->DBName);
	}
	db->DBPort=url.port;
	if((s=strchr(url.auth,':'))){
		*s=0;
		db->DBUser=strdup(url.auth);
		db->DBPass=strdup(s+1);
		*s=':';
	}else{
		db->DBUser=strdup(url.auth);
	}
	return(0);
}



/************************ MYSQL **************************************/
#if   (HAVE_MYSQL)

static int UdmMySQLInit(DB * db){
	mysql_init(&(db->mysql));
	if(!(mysql_real_connect(&(db->mysql),db->DBHost,db->DBUser,db->DBPass,db->DBName?db->DBName:"mnogosearch",(unsigned)db->DBPort,NULL,0))){
		db->errcode=1;
		return(1);
	}
	db->connected=1;
	return(0);
}
static void UdmMySQLDisplayError(DB * db){
	sprintf(db->errstr,"#%d: %s",
		mysql_errno(&((DB*)db)->mysql),
		mysql_error(&((DB*)db)->mysql));
}
static int safe_mysql_query(DB * db,const char *query){
	int err,i;

	if(!db->connected){
		UdmMySQLInit(db);
		if(db->errcode)
			return(db->errcode);
	}
	for(i=0;i<2;i++){
		if((err=mysql_query(&db->mysql,query))){
			if((mysql_errno(&db->mysql)==CR_SERVER_LOST)||
			(mysql_errno(&db->mysql)==CR_SERVER_GONE_ERROR)||
			(mysql_errno(&db->mysql)==ER_SERVER_SHUTDOWN)){
				UDMSLEEP(5);
			}else
				return(err);
		}else
			return(err);
	}
	return(err);
}
static MYSQL_RES * UdmMySQLQuery(DB * db,const char *query){
	int res;

	db->errcode=0;
	res=safe_mysql_query(db,query);
	if(res){
		UdmMySQLDisplayError(db);
		if((mysql_errno(&(db->mysql))!=ER_DUP_ENTRY) &&
			(mysql_errno(&(db->mysql))!=ER_DUP_KEY)){
			db->errcode=1;
		}
		return(0);
	}
	return(mysql_store_result(&db->mysql));
}
#endif


/*********************************** POSTGRESQL *********************/
#if (HAVE_PGSQL)

static int UdmPgSQLInitDB(DB * db){
	char port[8];

	sprintf(port,"%d",db->DBPort);
	db->pgsql = PQsetdbLogin(db->DBHost,db->DBPort?port:0, 0, 0, db->DBName, db->DBUser, db->DBPass);
	if (PQstatus(db->pgsql) == CONNECTION_BAD){
		db->errcode=1;
		return(1);
	}
	db->connected=1;
	return(0);
}
static void UdmPgSQLDisplayError(DB * db){
	sprintf(db->errstr, "%s", PQerrorMessage(db->pgsql));
}
static PGresult * safe_pgsql_query(DB * db, const char *q){

	if(!(db->connected)){
		UdmPgSQLInitDB(db);
		if(db->errcode)return(0);
	}
	return(PQexec(db->pgsql,q));
}
static PGresult * UdmPgSQLQuery(DB * db, const char *q){
	PGresult * res;

	db->errcode=0;
	res=safe_pgsql_query(db,q);
	if(!res){
		UdmPgSQLDisplayError(db);
		db->errcode=1;
		return(0);
	}
	if(PQresultStatus(res)==PGRES_COMMAND_OK){
		/* Free non-SELECT query */
		PQclear(res);
		return NULL;
	}
	if((PQresultStatus(res)!=PGRES_COMMAND_OK)&&(PQresultStatus(res)!=PGRES_TUPLES_OK)){
		UdmPgSQLDisplayError(db);
		if(!strstr(PQerrorMessage(db->pgsql),"duplicate"))
			db->errcode=1;
		PQclear(res);
		return(NULL);
	}
	return(res);
}

#endif


/*****************************  miniSQL *******************************/
#if (HAVE_MSQL)

static const char * msql_value(m_result *res,size_t i,size_t j){
	m_row row;

	if(i>=msqlNumRows(res))return(NULL);
	msqlDataSeek(res,(int)i);
	row=msqlFetchRow(res);
	if(row)return(row[j]?row[j]:"");
	else	return(0);
}
static int UdmMSQLInitDB(DB * db){
	char *host;
	
	/* To hide "never used" warning */
	if(msqlTypeNames[0][0]);
	
	if((host=db->DBHost))
		if(!strcmp(db->DBHost,"localhost"))
			host=NULL;
	if((db->msql=msqlConnect(host))<0){
		db->errcode=1;
		return(1);
	}
	if(msqlSelectDB(db->msql,db->DBName)==-1){
		db->errcode=1;
		return(1);
	}
	db->connected=1;
	return(0);
}
static void UdmMSQLDisplayError(DB *db){
	sprintf(db->errstr,"%s",msqlErrMsg);
}
static int safe_msql_query(DB * db,const char *query){
	int err;
	char * q;

	if(!(db->connected)){
		UdmMSQLInitDB(db);
		if(db->errcode)
			return(db->errcode);
	}
	q=strdup(query);
	if((err=msqlQuery(db->msql,q)==-1)){
		if(!strstr(msqlErrMsg,"Non unique")){
			free(q);
			db->errcode=1;
			return(err);
		}
		return(0);
	}
	free(q);
	return(0);
}
static m_result * UdmMSQLQuery(DB * db,const char *query){
	int err;

	db->errcode=0;
	if((err=safe_msql_query(db,query))){
		UdmMSQLDisplayError(db);
		return(NULL);
	}
	return(msqlStoreResult());
}
#endif


/************************** ODBC ***********************************/
#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT || HAVE_EASYSOFT || HAVE_SAPDB)


#define SQL_OK(rc)	((rc==SQL_SUCCESS)||(rc==SQL_SUCCESS_WITH_INFO))

static int UdmODBCDisplayError(DB * db){
	UCHAR szSqlState[6];
	UCHAR szErrMsg[256];
	SDWORD naterr;
	SWORD length;
	
	db->errstr[0]=0;
	while (SQL_SUCCESS == SQLError(db->hEnv,db->hDbc,db->hstmt,szSqlState,&naterr,szErrMsg,sizeof(szErrMsg),&length))
		strcat(db->errstr, szErrMsg);
	return(0);
}
static UDB_RES * execDB(DB*db,const char* sqlstr){
	RETCODE rc;
	SWORD iResColumns;
	SDWORD iRowCount;
	int i,res_count;
	char* p;
	UCHAR szColName[32];
	SWORD pcbColName;
	SWORD pfSQLType;
	UDWORD pcbColDef;
	SWORD pibScale;
	SWORD pfNullable;
	UDB_RES *result;
	SDWORD	pcbValue;
	static char	bindbuf[(int)(32L * 1024L - 16L)];

	/* -------- 
	p=sqlstr;
	while(*p){
		if(*p=='?')*p='.';
		p++;
	}
	*/

	rc=SQLAllocStmt(db->hDbc, &(db->hstmt));
	if (!SQL_OK(rc)){
		db->errcode=1;
		return(NULL);
	}
	rc=SQLExecDirect(db->hstmt,(const UCHAR*)sqlstr, SQL_NTS);
	if (!SQL_OK(rc)){
		db->errcode=1;
		return(NULL); 
	}
	rc=SQLNumResultCols(db->hstmt, &iResColumns);
	if(!SQL_OK(rc)){
		db->errcode=1;
		return(NULL);
	}
	if(!iResColumns) {
		rc=SQLRowCount(db->hstmt, &iRowCount);
		if (!SQL_OK(rc)){
			db->errcode=1;
			return(NULL);
		}
		result=NULL;
	}else{
		result=(UDB_RES*)UdmXmalloc(sizeof(UDB_RES));
		result->nRows = 0;
		result->nCols = iResColumns;
		result->items=NULL;

		rc = SQL_NO_DATA_FOUND;
		for (res_count=0;(db->res_limit?(res_count<db->res_limit):1);res_count++){
			rc=SQLFetch(db->hstmt);
			if (!SQL_OK(rc)) {
				if (rc!=SQL_NO_DATA_FOUND){
					db->errcode=1;
				}
				break;
			}
			if(!result->nRows){
				result->items=(char **)UdmXmalloc(1*iResColumns*sizeof(char*));
			}else{
				result->items=(char **)UdmXrealloc(result->items,((result->nRows+1)*iResColumns*sizeof(char*)));
			}
			for (i = 0; i < iResColumns; i++) {
				SQLDescribeCol(db->hstmt, i+1, szColName, sizeof(szColName),
					&pcbColName, &pfSQLType, &pcbColDef,&pibScale, &pfNullable);
				SQLGetData(db->hstmt,i+1,SQL_C_CHAR,bindbuf,sizeof(bindbuf),&pcbValue);
				if (pcbValue==SQL_NULL_DATA) p = NULL;
				else p = bindbuf;
				if(p)UdmRTrim(p," ");
				sql_val(result,result->nRows,i)=strdup(p?p:"");
			}
			result->nRows++;
		}
	}
	SQLFreeStmt(db->hstmt, SQL_DROP);
	
	db->res_limit=0;
	db->errcode=0;
	return(result);
}
static int UdmODBCInitDB(DB * db){
	char DSN[512]="";

#if (HAVE_SOLID)
	snprintf(DSN,sizeof(DSN)-1,"tcp %s %d",db->DBHost?db->DBHost:"localhost",db->DBPort?db->DBPort:1313);
#elif (HAVE_SAPDB)
	snprintf(DSN,sizeof(DSN)-1,"%s:%s",db->DBHost?db->DBHost:"localhost",db->DBName?db->DBName:"");
#else
	strncpy(DSN,db->DBName?db->DBName:"",sizeof(DSN)-1);
#endif

	db->errcode = SQLAllocEnv( &(db->hEnv) );
	if( SQL_SUCCESS != db->errcode )return -2;
	db->errcode = SQLAllocConnect( db->hEnv, &(db->hDbc) );
	if( SQL_SUCCESS != db->errcode )return -3;
	db->errcode = SQLSetConnectOption( db->hDbc, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_ON);
	if( SQL_SUCCESS != db->errcode )return -4;
	db->errcode = SQLConnect( db->hDbc, DSN, SQL_NTS, db->DBUser?(db->DBUser[0]?db->DBUser:NULL):NULL, SQL_NTS, db->DBPass?(db->DBPass[0]?db->DBPass:NULL):NULL, SQL_NTS);
	if( !SQL_OK(db->errcode)) return(-5);
	else db->errcode=0;
	db->connected=1;
	return 0;
}

static void UdmODBCCloseDB(DB * db){
	if(db->connected){
		db->connected=0;
		db->errcode = SQLTransact( db->hEnv, db->hDbc, SQL_COMMIT);
		if( SQL_SUCCESS != db->errcode )return;
		db->errcode = SQLDisconnect( db->hDbc );
		if( SQL_SUCCESS != db->errcode )return;
		db->errcode = SQLFreeConnect( db->hDbc );
		if( SQL_SUCCESS != db->errcode )return;
		else	db->hDbc = SQL_NULL_HDBC;
		db->errcode = SQLFreeEnv( db->hEnv );
		if( SQL_SUCCESS != db->errcode )return;
		else	db->hEnv = SQL_NULL_HENV;
	}
}
static UDB_RES * UdmODBCQuery(DB * db,const char *qbuf){
	UDB_RES * res;

	if(!db->connected){
		UdmODBCInitDB(db);
		if(db->errcode){
			UdmODBCDisplayError(db);
			return(0);
		}else{
			db->connected=1;
		}
	}
	res=execDB(db,qbuf);
	if((db->errcode)){
		UdmODBCDisplayError(db);
		if(strstr(db->errstr,"uplicat")){ /* PgSQL,MySQL*/
			db->errcode=0;
		}else
		if(strstr(db->errstr,"nique")){ /* Solid, Virtuoso */
			db->errcode=0;
		}else{
			db->errcode=1;
			res=NULL;
		}
		SQLFreeStmt(db->hstmt, SQL_DROP);
	}
	return(res);
}

#endif

/*******************************************************/

#if HAVE_IBASE

#define SQL_VARCHAR(len) struct {short vary_length; char vary_string[(len)+1];}

static void UdmIBaseDisplayError(DB * db){
	char * s = db->errstr;
	ISC_STATUS * ibstatus=db->status;
	
	while(isc_interprete(db->errstr ,&ibstatus)){
		strcat(s," ");
		s=db->errstr+strlen(db->errstr);
	}
}

static int UdmIBaseInitDB(DB * db){
	char dpb_buffer[256], *dpb, *p;
	int dpb_length, len;

	dpb = dpb_buffer;
	*dpb++ = isc_dpb_version1;

	if (db->DBUser != NULL && (len = strlen(db->DBUser))) {
		*dpb++ = isc_dpb_user_name;
		*dpb++ = len;
		for (p = db->DBUser; *p;) {
			*dpb++ = *p++;
		}
	}
	if (db->DBPass != NULL && (len = strlen(db->DBPass))) {
		*dpb++ = isc_dpb_password;
		*dpb++ = strlen(db->DBPass);
		for (p = db->DBPass; *p;) {
			*dpb++ = *p++;
		}
	}
	/*
	if (charset != NULL && (len = strlen(charset))) {
		*dpb++ = isc_dpb_lc_ctype;
		*dpb++ = strlen(charset);
		for (p = charset; *p;) {
			*dpb++ = *p++;
		}
	}
#ifdef isc_dpb_sql_role_name
	if (role != NULL && (len = strlen(role))) {
		*dpb++ = isc_dpb_sql_role_name;
		*dpb++ = strlen(role);
		for (p = role; *p;) {
			*dpb++ = *p++;
		}
	}
#endif
	*/

	dpb_length = dpb - dpb_buffer;

	if(isc_attach_database(db->status, strlen(db->DBName), db->DBName, &(db->DBH), dpb_length, dpb_buffer)){
		db->errcode=1;
		return(1);
	}
	return(0);
}
static void UdmIBaseCloseDB(DB*db){
	if(db->connected){
		if (isc_detach_database(db->status, &(db->DBH))){
			db->errcode=1;
		}
	}
}
static UDB_RES * sql_ibase_query(DB * db, const char *query){
	UDB_RES *result=NULL;
	char query_info[] = { isc_info_sql_stmt_type };
	char info_buffer[18];
	long query_type;
	isc_stmt_handle query_handle = NULL;
	int autocommit=1;
	ISC_STATUS status[20]; /* To not override db->status */

	if(!db->connected){
		UdmIBaseInitDB(db);
		if(db->errcode){
			UdmIBaseDisplayError(db);
			db->errcode=1;
			return NULL;
		}else{
			db->connected=1;
		}
	}

	if(!strcmp(query,"BEGIN")){
		if(!db->tr_handle){
			if (isc_start_transaction(db->status, &db->tr_handle, 1, &(db->DBH), 0, NULL))
				db->errcode=1;
			return(NULL);
		}else{
			db->errcode=1;
			snprintf(db->errstr,sizeof(db->errstr)-1,"Wrong calls order: begin");
		}
		return NULL;
	}
	
	if(!strcmp(query,"COMMIT")){
		if(db->tr_handle){
			if (isc_commit_transaction(db->status, &db->tr_handle))
				db->errcode=1;
			
			db->tr_handle=NULL;
			return(NULL);
		}else{
			db->errcode=1;
			snprintf(db->errstr,sizeof(db->errstr)-1,"Wrong calls order: commit");
		}
		return NULL;
	}

	if(!db->tr_handle){
		if (isc_start_transaction(db->status, &db->tr_handle, 1, &(db->DBH), 0, NULL)){
			db->errcode=1;
			return(0);
		}
		autocommit=1;
	}else{
		autocommit=0;
	}
	
	if (isc_dsql_allocate_statement(db->status, &(db->DBH), &query_handle)){
		db->errcode=1; 
		return(0);
	}
	if (isc_dsql_prepare(db->status, &db->tr_handle, &query_handle, 0, query, 1, NULL)){
		db->errcode=1; 
		return(0);
	}
	if (!isc_dsql_sql_info(db->status, &query_handle, sizeof(query_info), query_info, sizeof(info_buffer), info_buffer)) {
		short l;
		l = (short) isc_vax_integer((char ISC_FAR *) info_buffer + 1, 2);
		query_type = isc_vax_integer((char ISC_FAR *) info_buffer + 3, l);
	}

	/* Find out what kind of query is to be executed */
	if (query_type == isc_info_sql_stmt_select || query_type == isc_info_sql_stmt_select_for_upd) {
		XSQLDA *osqlda=NULL;
		long fetch_stat;
		int i;
		
		/*
		 * Select, need to allocate output sqlda and and prepare it for use.
		 */
		osqlda = (XSQLDA *) UdmXmalloc(XSQLDA_LENGTH(0));
		osqlda->sqln = 0;
		osqlda->version = SQLDA_VERSION1;

		if (isc_dsql_describe(db->status, &query_handle, 1, osqlda)) {
			free(osqlda);
			db->errcode=1;
			isc_rollback_transaction(status, &db->tr_handle);
			return NULL;
		}

		if (osqlda->sqld) {
			osqlda = (XSQLDA *) UdmXrealloc(osqlda, XSQLDA_LENGTH(osqlda->sqld));
			osqlda->sqln = osqlda->sqld;
			osqlda->version = SQLDA_VERSION1;
			if (isc_dsql_describe(db->status, &query_handle, 1, osqlda)) {
				free(osqlda);
				db->errcode=1;
				isc_rollback_transaction(status, &db->tr_handle);
				return NULL;
			}
		}
		for (i = 0; i < osqlda->sqld; i++) {
			int coltype;
			
			osqlda->sqlvar[i].sqlind = (short *) UdmXmalloc(sizeof(short));
			coltype = osqlda->sqlvar[i].sqltype & ~1;
			switch(coltype)
				{
				case SQL_TEXT:
					osqlda->sqlvar[i].sqldata = (char *) UdmXmalloc(sizeof(char)*(osqlda->sqlvar[i].sqllen));
					break;
				case SQL_VARYING:
					osqlda->sqlvar[i].sqldata = (char *) UdmXmalloc(sizeof(char)*(osqlda->sqlvar[i].sqllen+2));
					break;
				case SQL_SHORT:
					osqlda->sqlvar[i].sqldata = (char *) UdmXmalloc(sizeof(short));
					break;
				case SQL_LONG:
					osqlda->sqlvar[i].sqldata = (char *) UdmXmalloc(sizeof(long));
					break;
				case SQL_FLOAT:
					osqlda->sqlvar[i].sqldata = (char *) UdmXmalloc(sizeof(float));
					break;
				case SQL_DOUBLE:
					osqlda->sqlvar[i].sqldata = (char *) UdmXmalloc(sizeof(double));
					break;
				case SQL_DATE:
				case SQL_BLOB:
				case SQL_ARRAY:
					osqlda->sqlvar[i].sqldata = (char *) UdmXmalloc(sizeof(ISC_QUAD));
					break;
				}
		}
		if (isc_dsql_execute(db->status, &db->tr_handle, &query_handle, 1, NULL)) {
			free(osqlda);
			db->errcode=1;
			isc_rollback_transaction(status, &db->tr_handle);
			return NULL;
		}
		result=(UDB_RES*)UdmXmalloc(sizeof(UDB_RES));
		result->nRows = 0;
		result->nCols = osqlda->sqld;
		result->items=NULL;

		while ((fetch_stat = isc_dsql_fetch(db->status, &query_handle, 1, osqlda)) == 0){
			char buf[10*1024]=""; /* FIXME unsafe */
			XSQLVAR *var;
			
			var=osqlda->sqlvar;

			if(!result->nRows){
				result->items=(char **)UdmXmalloc(1*osqlda->sqld*sizeof(char*));
			}else{
				result->items=(char **)UdmXrealloc(result->items,((result->nRows+1)*osqlda->sqld*sizeof(char*)));
			}
			for(i=0;i<osqlda->sqld; i++, var++){
				if(*var->sqlind==-1)
					 buf[0]=0;  /* NULL data */
				else
				switch(var->sqltype & ~1){
					case SQL_TEXT:
					strncpy(buf,(char*)var->sqldata,var->sqllen);
					buf[var->sqllen]=0;
					break;
					case SQL_LONG:
					sprintf(buf,"%ld",*(long*)(var->sqldata));
					break;
					case SQL_SHORT:
					sprintf(buf,"%d",*(short*)(var->sqldata));
					break;
					default:
					sprintf(buf,"Unknown SQL type\n");
					buf[0]=0;
					break; 
				}
				UdmRTrim(buf," ");
				sql_val(result,result->nRows,i)=strdup(buf);
			}
			result->nRows++;
		}
		/* Free fetch buffer */
		for (i = 0; i < osqlda->sqld; i++) {
			free(osqlda->sqlvar[i].sqlind);
			free(osqlda->sqlvar[i].sqldata);
		}
		free(osqlda);

		if (fetch_stat != 100L){
			db->errcode=1; 
			isc_rollback_transaction(status, &db->tr_handle);
			return NULL;
		}
	} else {
		/* Not select */
		if (isc_dsql_execute(db->status, &db->tr_handle, &query_handle, 1, NULL)) {
			db->errcode=1;
			isc_rollback_transaction(status, &db->tr_handle);
			isc_dsql_free_statement(status, &query_handle, DSQL_drop);
			return NULL;
		}
		result=NULL;
	}
	if (isc_dsql_free_statement(db->status, &query_handle, DSQL_drop)){
		db->errcode=1; 
		isc_rollback_transaction(status, &db->tr_handle);
		return NULL;
	}
	if(autocommit){
		if (isc_commit_transaction(db->status, &db->tr_handle)){
			db->errcode=1; 
			db->tr_handle=NULL;
			return NULL;
		}
		db->tr_handle=NULL;
	}
	return result;
}
static UDB_RES * UdmIBaseQuery(DB * db, const char *query){
	UDB_RES * res;

	res=sql_ibase_query(db,query);
	if(db->errcode){
		UdmIBaseDisplayError(db);
		if(strstr(db->errstr,"uplicat")){
			db->errcode=0;
		}else{
			/*strcat(db->errstr," ");
			strcat(db->errstr,query);*/
		}
		return(NULL);
	}
	return(res);
}

#endif


/******************** Oracle8 OCI - native support driver ******************/

#if HAVE_ORACLE8

/* (C) copyleft 2000 Anton Zemlyanov, az@hotmail.ru */
/* TODO: efficient transactions, multi-row fetch, limits stuff */

#define BUF_OUT_SIZE		128


#define SQL_OK(rc)	((rc==OCI_SUCCESS)||(rc==OCI_SUCCESS_WITH_INFO))

static int oci_free_result(UDB_RES *res)
{
	int i;
	if(res){
		if(res->items){
			for(i=0;i<res->nCols*res->nRows;i++)
				if(res->items[i])
					free(res->items[i]);
			free(res->items);
		}
		if(res->defbuff){
			for(i=0;i<res->nCols;i++)
				if(res->defbuff[i])
					free(res->defbuff[i]);
		}
	}
	return(0);
}

static int UdmOracle8DisplayError(DB *db)
{
	sb4	errcode=0;
	text	errbuf[512];
	char	*ptr;

	db->errstr[0]='\0';

	switch (db->errcode)
	{
	case OCI_SUCCESS:
		sprintf(db->errstr,"Oracle - OCI_SUCCESS");
		break;
	case OCI_SUCCESS_WITH_INFO:
		sprintf(db->errstr,"Oracle - OCI_SUCCESS_WITH_INFO");
		break;
	case OCI_NEED_DATA:
		sprintf(db->errstr,"Oracle - OCI_NEED_DATA");
		break;
	case OCI_NO_DATA:
		sprintf(db->errstr,"Oracle - OCI_NODATA");
		break;
	case OCI_ERROR:
		OCIErrorGet((dvoid *)db->errhp, (ub4) 1, (text *) NULL, &errcode,
			errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
		ptr=errbuf;
		while(*ptr) {
			if(*ptr=='\n')
				*ptr='!';
			++ptr;
		}
		sprintf(db->errstr,"Oracle - %.*s", 512, errbuf);
		break;
	case OCI_INVALID_HANDLE:
		sprintf(db->errstr,"Oracle - OCI_INVALID_HANDLE");
		break;
	case OCI_STILL_EXECUTING:
		sprintf(db->errstr,"Oracle - OCI_STILL_EXECUTE");
		break;
	case OCI_CONTINUE:
		sprintf(db->errstr,"Oracle - OCI_CONTINUE");
		break;
	default:
		sprintf(db->errstr,"Oracle - unknown internal bug");
		break;
	}
	return 0;
}


static void UdmOracle8CloseDB(DB *db)
{
	if(!db->connected)
		return;
	db->errcode=OCILogoff(db->svchp,db->errhp);
	db->connected=0;
}

static int oci_bind(DB *db){
	int i,pos;
	sword   rc;

        for (i=0; i<MAX_BIND_PARAM; i++)
                db->bndhp[i] = (OCIBind *) 0;

        for( i=0; i<MAX_BIND_PARAM; i++){
                if ((pos=db->par->out_pos[i])>0){
                        if ( (rc = OCIBindByPos(db->stmthp, &db->bndhp[pos-1], db->errhp, (ub4) pos,
				(dvoid *) &db->par->out_pos_val[pos-1][0],
                                (sb4) sizeof(db->par->out_pos_val[pos-1][0]), SQLT_INT,                                
				(dvoid *) 0, (ub2 *)0, (ub2 *)0,
                                (ub4) 0, (ub4 *) 0, (ub4) OCI_DEFAULT))){

                                return rc;
                        }
                        /*  bind array  */
                        if ((rc = OCIBindArrayOfStruct(db->bndhp[pos-1], db->errhp,                                
				sizeof(db->par->out_pos_val[pos-1][0]),0, 0, 0))){

                                return rc;
                        }
                }
        }
        return 0;
}

void param_add(DB *db, int pos_out_1, int pos_out_2, int pos_out_3){
        db->par->out_pos_val[0][db->par->out_rec] = pos_out_1;
        db->par->out_pos_val[1][db->par->out_rec] = pos_out_2;
        db->par->out_pos_val[2][db->par->out_rec] = pos_out_3;
        db->par->out_rec++;
}

void param_init(DB *db, int pos1, int pos2, int pos3){
        db->par->out_pos[0]=pos1;
        db->par->out_pos[1]=pos2;
        db->par->out_pos[2]=pos3;
}

void param_free(DB *db){
        db->par->out_rec = 0;
}

static int UdmOracle8InitDB(DB * db){

	OCIInitialize( OCI_DEFAULT, NULL, NULL, NULL, NULL );
	OCIEnvInit( &(db->envhp), OCI_DEFAULT, 0, NULL );
	OCIHandleAlloc( db->envhp, (dvoid **)&(db->errhp), OCI_HTYPE_ERROR,0, NULL );
	db->errcode=OCILogon(db->envhp, db->errhp, &(db->svchp),
		db->DBUser, strlen(db->DBUser),
		db->DBPass, strlen(db->DBPass),
		db->DBName, strlen(db->DBName) );
	if(db->errcode!=OCI_SUCCESS)
		return -1;

	db->errcode=0;
	db->connected=1;
	db->commit_fl=0;
	param_free(db);

/*        sql_query(((DB*)(db)), "ALTER SESSION SET SQL_TRACE=TRUE"); */

	return 0;
}

static UDB_RES* sql_oracle_query(DB * db, char *qbuf)
{
	UDB_RES *result;
	sword	rc;
	ub2	stmt_type;
	int	cnt, buf_nRows, row;

	int	colcnt;
	ub2	coltype;
	ub2	colsize;
	int	oci_fl;
	sb4	errcode=0;
	text	errbuf[512];
	int	num_rec;

	rc=OCIHandleAlloc(db->envhp, (dvoid *)&(db->stmthp), OCI_HTYPE_STMT,0,NULL);
	if(!SQL_OK(rc)){
		db->errcode=rc;
		return NULL;
	}
	rc=OCIStmtPrepare(db->stmthp,db->errhp,qbuf,strlen(qbuf),
			OCI_NTV_SYNTAX,OCI_DEFAULT);
	if(!SQL_OK(rc)){
		db->errcode=rc;
		return NULL;
	}

        if (db->par->out_rec){
                oci_bind(db);
                num_rec = db->par->out_rec;
                param_free(db);
        }else{
                num_rec = 1;
        }
    
	rc=OCIAttrGet(db->stmthp,OCI_HTYPE_STMT,&stmt_type,0,
			OCI_ATTR_STMT_TYPE,db->errhp);
	if(!SQL_OK(rc)){
		db->errcode=rc;
		return NULL;
	}

	if(stmt_type!=OCI_STMT_SELECT) {
		/* non-select statements */
		/* COMMIT_ON_SUCCESS in inefficient */
		if (db->commit_fl)
			oci_fl=OCI_DEFAULT;
		else
			oci_fl=OCI_COMMIT_ON_SUCCESS;

		rc=OCIStmtExecute(db->svchp,db->stmthp,db->errhp, num_rec,0,
				NULL,NULL,oci_fl);

		if (num_rec>1)
			  (void) OCITransCommit(db->svchp, db->errhp, (ub4) 0);

		if(!SQL_OK(rc)){
			db->errcode=rc;
			OCIErrorGet((dvoid *)db->errhp, (ub4) 1, (text *) NULL, &errcode,
				errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
			if(strncmp(errbuf,"ORA-00001",9)) /* ignore ORA-00001 */
				return NULL;
			else 
				db->errcode=OCI_SUCCESS;
		}
		/* OCI_ATTR_ROW_COUNT of db->stmthp - Rows affected */
		rc=OCIHandleFree(db->stmthp,OCI_HTYPE_STMT);
		if(!SQL_OK(rc)){
			db->errcode=rc;
			return NULL;
		}
		return NULL;
	}

	/* select statements */
	/*Allocate result Set*/
	result=(UDB_RES*)UdmXmalloc(sizeof(UDB_RES));
	result->nRows=0;
	result->nCols=0;
	result->items=NULL;

	rc=OCIStmtExecute(db->svchp,db->stmthp,db->errhp,0,0,
			NULL,NULL,OCI_DEFAULT);
	if(!SQL_OK(rc)){
		db->errcode=rc;
		return NULL;
	}

	/*describe the select list and define buffers for the select list */
	colcnt=0;
	while(OCIParamGet(db->stmthp,OCI_HTYPE_STMT,db->errhp,(dvoid *)&(db->param),
			colcnt+1 ) == OCI_SUCCESS) {
		rc=OCIAttrGet(db->param,OCI_DTYPE_PARAM,&(coltype),0,
				OCI_ATTR_DATA_TYPE,db->errhp);
		if(!SQL_OK(rc)){
			db->errcode=rc;
			return NULL;
		}
		rc=OCIAttrGet(db->param,OCI_DTYPE_PARAM,&colsize,0,
			OCI_ATTR_DATA_SIZE,db->errhp);
		if(!SQL_OK(rc)){
			db->errcode=rc;
			return NULL;
		}

		/* OCIStmtFetch do not terminate data with \0 -
		add a byte for terminator to define buffers - insurance*/

		switch(coltype) {
		case SQLT_CHR: /* variable length string */
		case SQLT_AFC: /* fixed length string */
                        result->col_size[colcnt]=colsize;
                        result->defbuff[colcnt]=(char *)UdmXmalloc(result->col_size[colcnt]*BUF_OUT_SIZE+1);
			break;
		case SQLT_NUM: /* numbers up to 14 digits now */
                        result->col_size[colcnt]=14;
                        result->defbuff[colcnt]=(char *)UdmXmalloc(result->col_size[colcnt]*BUF_OUT_SIZE+1);
			break;
		default:
			printf("<P>Unknown datatype: %d\n",coltype);
			return (UDB_RES*)NULL;
		}
		rc=OCIDefineByPos(db->stmthp,&(db->defb[colcnt]),db->errhp,
			colcnt+1,result->defbuff[colcnt],result->col_size[colcnt],SQLT_CHR,
			&(result->indbuff[colcnt]),0,0,OCI_DEFAULT);
		if(!SQL_OK(rc)){
			db->errcode=rc;
	 		return NULL;
		}
                rc=OCIDefineArrayOfStruct(db->defb[colcnt], db->errhp,
                         result->col_size[colcnt], sizeof(result->indbuff[0][0]), 0, 0);
		if(!SQL_OK(rc)){
                        db->errcode=rc;
                        return NULL;
                }
		colcnt++;
	}
	result->nCols=colcnt;

	/* Now fetching the selected rows into the memory */
	while(1) {
		if (db->res_limit)
			if (db->res_limit == result->nRows)
				break;
                /*Fix me: Process of determine real fetched rows
                 Fill indicator buffer for value -6 to */
                for(row=0; row<BUF_OUT_SIZE; row++)
                        result->indbuff[result->nCols-1][row]=-6;

		rc=OCIStmtFetch(db->stmthp,db->errhp,BUF_OUT_SIZE,
			OCI_FETCH_NEXT,OCI_DEFAULT);

                if((rc!=OCI_NO_DATA) && !SQL_OK(rc)) {
                        db->errcode=rc;
                        return NULL;
                }

                /* Find number of fetched rows */
                for(buf_nRows=0; buf_nRows<BUF_OUT_SIZE; buf_nRows++){
                        if(result->indbuff[result->nCols-1][buf_nRows]==-6)
                                break;
                }

                if(!buf_nRows)
                        break;

		if(!result->nRows)  /* first row - allocate */
			result->items=(char **)UdmXmalloc(buf_nRows*result->nCols*sizeof(char*));
		else
			result->items=(char **)UdmXrealloc(result->items,
				((result->nRows+buf_nRows)*result->nCols*sizeof(char*)) );
		/* Limit rows */	
	        if (db->res_limit)
            		buf_nRows=result->nRows+buf_nRows<=db->res_limit?buf_nRows:(db->res_limit-result->nRows);
		for(row=0; row<buf_nRows; row++){
			for(cnt=0; cnt<result->nCols; ++cnt) {
                    		if(result->indbuff[cnt][row]==OCI_IND_NULL) {
                            		sql_val(result,(result->nRows+row),cnt) = strdup("");
                                }else{
                                        char *val=UdmXmalloc(result->col_size[cnt]+1);
                                        int offset=row*result->col_size[cnt];

                                        snprintf(val, result->col_size[cnt]+1, "%s",
						result->defbuff[cnt]+offset);
                                        sql_val(result,(result->nRows+row),cnt)=
                                    		strdup( UdmRTrim(val," "));
                                        UDM_FREE(val);
                                }
                        }
                }
		result->nRows+=row;
		if(rc==OCI_NO_DATA)
                        break;
                if(!SQL_OK(rc)) {
                        db->errcode=rc;
                        return NULL;
                }

	}
	rc=OCIHandleFree(db->stmthp,OCI_HTYPE_STMT);
	db->res_limit = 0;
	if(!SQL_OK(rc)){
		db->errcode=rc;
		return NULL;
	}

	return (UDB_RES*)result;
}

static UDB_RES* UdmOracle8Query(DB * db, char *qbuf) {
	UDB_RES *res;

	if(!db->connected) {
		UdmOracle8InitDB(db);
		if(db->errcode) {
			UdmOracle8DisplayError(db);
			return NULL;
		} else {
			db->connected=1;
		}
 	}
	res=sql_oracle_query(db,qbuf);

	if(db->errcode) {
		UdmOracle8DisplayError(db);
		return NULL;
	}
	return (UDB_RES*)res;
}

#endif



/******************** Oracle7 OCI - native support driver ******************/

#if HAVE_ORACLE7

 * (C) 2000 Kir Maximov, maxkir@email.com 
 * 
 * Without transactions ...
 */

#define MAX_COLS_IN_TABLE    32
#define HDA_SIZE 256
#define MAX_ITEM_BUFFER_SIZE 32

#define UDM_OCI_UNIQUE 1
#define ORA_VERSION_7 2

/*  some SQL and OCI function codes */
#define FT_INSERT                3
#define FT_SELECT                4
#define FT_UPDATE                5
#define FT_DELETE                9

/* Declare structures for query information. */
struct describe
{
    sb4             dbsize;
    sb2             dbtype;
    sb1             buf[MAX_ITEM_BUFFER_SIZE];
    sb4             buflen;
    sb4             dsize;
};

struct define 
{
    ub1             buf[1024*5];	/* FIXME unsafe */
    sb2             indp;
    ub2             col_retlen, col_retcode;
};


/*-------------------------------------------------------*/
static int oci_free_result(UDB_RES *res)
{
    int i;

    if(res)
    {
        if(res->items)
        {
            for(i=0;i<res->nCols*res->nRows;i++)
            {
                if(res->items[i])
                    free(res->items[i]);
            }
            free(res->items);
        }
    }
    return(0);
}
/*-------------------------------------------------------*/
static int UdmOracle7DisplayError(DB *db)
{
    sword n;
    text msg[512];

    db->errstr[0]='\0';
    if (!db->errcode) 
        return 0; 

    n = oerhms(&db->lda, db->cursor.rc, msg, (sword) sizeof msg);
    
    strcpy(db->errstr, "ORACLE - ");
    strcat(db->errstr, msg);

    printf("displayError: %p %s\n", db, msg);

    return 0;
}
/*-------------------------------------------------------*/
static int UdmOracle7InitDB(DB * db)
{
    Cda_Def *cda = &db->cursor;
    Lda_Def *lda = &db->lda;
    char buff[1024];

    snprintf(buff, sizeof(buff)-1, "%s/%s@%s",db->DBUser,db->DBPass,db->DBName);
    
    db->errcode = olog(lda, db->hda,buff, -1,(text *)0, -1,0, -1, OCI_LM_DEF);

    if ( db->errcode )
    {
        /* Error happened */
        return -1;
    }

    /* Now, open a cursor to be used in all queries*/
    oopen(cda, lda, (text*)0, -1, -1, (text*)0, -1);
    db->errcode = cda->rc;
    if (db->errcode)
    {
        UdmOracle7DisplayError(db);
        return 0;
    }
    
    db->errcode = 0;
    db->connected = 1;

    return 0;
}
/*-------------------------------------------------------*/
static void UdmOracle7CloseDB(DB *db)
{
    if(!db->connected)
        return;
    
    /* First, close a cursor */
    oclose(&db->cursor);

    /* Logout */
    db->errcode = ologof(&db->lda);
    db->connected = 0;
}
/*-------------------------------------------------------*/
static UDB_RES* fetch_data(DB *db)
{
    /* Real feching of data for SELECT statements */
    UDB_RES *res = NULL;
    int col, deflen;
    struct describe desc[MAX_COLS_IN_TABLE];
    struct define   def[MAX_COLS_IN_TABLE];
    Cda_Def *cda = &db->cursor;
    char *buf_ptr;
   

    /* Now, get column desriptions */
    for(col = 0; col < MAX_COLS_IN_TABLE; col ++)
    {
        desc[col].buflen = MAX_ITEM_BUFFER_SIZE;
        if ( odescr( cda, col+1, &desc[col].dbsize,
                     &desc[col].dbtype, desc[col].buf,
                     &desc[col].buflen, &desc[col].dsize,
                     0,0,0
                   )
           )
        {
            if ( cda->rc == 1007 ) /* No more variables */
                break;
            db->errcode = cda->rc;
            if (db->errcode)
            {
                UdmOracle7DisplayError(db);
                return NULL;
            }
            
        }
        
        switch(desc[col].dbtype)
        {
        case 2: /* NUMBER */
            deflen = 14;
            break;
        case 96: /* CHAR */
        case 1:  /* VARCHAR2 */
        default:
            deflen = (desc[col].dbsize > 1024 ? 1024 : desc[col].dbsize + 1);
        }
       
        if (odefin(cda, col + 1,
                    def[col].buf, deflen, 5, /* 5 - null terminated string */
                    -1, &def[col].indp, (text*)0, -1, -1,
                    &def[col].col_retlen,
                    &def[col].col_retcode ))
        {
            db->errcode = cda->rc;
            if (db->errcode)
            {
                UdmOracle7DisplayError(db);
                return NULL;
            }
        }
    }
    /* Now col contains a number of columns */

    /* Get memory for resulting data */
    res = (UDB_RES*)UdmXmalloc(sizeof(UDB_RES));
    if (!res)
        return NULL;

    /* Clear all structure data */
    memset(res, 0, sizeof(UDB_RES));

    res->nCols = col;

    /* Now, fetching the data */
    for (;;)
    {
        /* Fetch a row, break on end of fetch, */
        /* disregard null fetch "error" */
        
        if (ofetch(cda))
        {
            if ( cda->rc == 1403 ) /* No data found */
                break;
            if ( cda->rc != 1405 && cda->rc ) /* Null value returned */
            {
               SQL_FREE(res);
               db->errcode = cda->rc;
               UdmOracle7DisplayError(db);
               return NULL;
            }
        }
        /* Next row: */
        
        /* [Re]allocate memory for data */
        if (!res->items)
        {
            res->items = (char**) UdmXmalloc( res->nCols * sizeof(char*));
        }
        else
        {
            res->items = (char**) UdmXrealloc(res->items, 
                             (res->nRows + 1) * res->nCols * sizeof(char*));
        }
                
        for (col = 0; col < res->nCols; col++)
        {
            if (def[col].indp < 0) /* NULL value */
            {
                buf_ptr = strdup("");
            }
            else
            {
                buf_ptr = strdup((char*)def[col].buf);
                UdmRTrim(buf_ptr, " ");
            }
            /* sql_val(res, res->nRows, col)*/
            res->items[res->nRows*res->nCols + col] = buf_ptr;
            /*printf( "\n<P>Field %d Val '%s' Null '%d'",col,sql_value(res,res->nRows,col), def[col].indp );*/

        }
        res->nRows ++;
    }

    return res;    
}
/*-------------------------------------------------------*/
static UDB_RES* sql_oracle_query(DB * db, char *query)
{
    /* Make real query and store result */
    UDB_RES* res = NULL;
    Cda_Def *cda = &db->cursor;
    
    /* Parse SQL statement */
    oparse(cda, (text*) query, (sb4)-1, (sword) 0, (ub4)ORA_VERSION_7 );
    if ( cda->rc )
    {
        db->errcode = cda->rc;
        UdmOracle7DisplayError(db);
        return NULL;
    }
    /* Save sql function */

    /* Exec */
    oexec(cda);
    if (cda->rc)
    {
        if (cda->rc != UDM_OCI_UNIQUE)  /* ignore unique constraint violation */
        {
            UdmOracle7DisplayError(db);
            return NULL;
        }
        db->errcode = 0;
    }
    
    switch (cda->ft)
     {
        case FT_DELETE:
        case FT_INSERT:
        case FT_UPDATE:
            /* Empty result */
            res = NULL;

            /* commit */
            ocom(&db->lda);

            break;
        case FT_SELECT:
            /* Fetch data needed */
            res = fetch_data(db);
            break;
        default:
            res = NULL;
            break;
    }
    
    return res;
}

/*-------------------------------------------------------*/
static UDB_RES* UdmOracle7Query(DB * db, char *qbuf) 
{
    UDB_RES *res;

    if(!db->connected) 
    {
        UdmOracle7InitDB(db);
        if(db->errcode) 
        {
            UdmOracle7DisplayError(db);
            return NULL;
        }
    }
    
    /* Make real query */
    res = sql_oracle_query(db, qbuf);
    
    if(!res) 
    {
        if (db->errcode)
        {
            UdmOracle7DisplayError(db);
        }
        else
        {
            fprintf(stderr, "sql_query: this should never happen.\n");   
        }
    }
    return res;
}

#endif



/***********************************************************************/

#if HAVE_CTLIB

#ifndef MAX
#define MAX(X,Y)	(((X) > (Y)) ? (X) : (Y))
#endif

#ifndef MIN
#define MIN(X,Y)	(((X) < (Y)) ? (X) : (Y))
#endif

#define MAX_CHAR_BUF	1024

typedef struct _ex_column_data
{
	CS_INT		indicator;
	CS_CHAR		*value;
	CS_INT		valuelen;
} EX_COLUMN_DATA;

static char sybsrvmsg[2048]="";

static CS_RETCODE CS_PUBLIC
ex_clientmsg_cb(CS_CONTEXT *context,CS_CONNECTION *connection,CS_CLIENTMSG *errmsg)
{
	fprintf(stderr, "\nOpen Client Message:\n");
	fprintf(stderr, "Message number: LAYER = (%ld) ORIGIN = (%ld) ",(long)CS_LAYER(errmsg->msgnumber), (long)CS_ORIGIN(errmsg->msgnumber));
	fprintf(stderr, "SEVERITY = (%ld) NUMBER = (%ld)\n",(long)CS_SEVERITY(errmsg->msgnumber), (long)CS_NUMBER(errmsg->msgnumber));
	fprintf(stderr, "Message String: %s\n", errmsg->msgstring);
	if (errmsg->osstringlen > 0)
	{
		fprintf(stderr, "Operating System Error: %s\n",errmsg->osstring);
	}
	return CS_SUCCEED;
}

static CS_RETCODE CS_PUBLIC
ex_servermsg_cb(CS_CONTEXT *context,CS_CONNECTION *connection,CS_SERVERMSG *srvmsg)
{
#ifdef DEBUG_SQL
	char msg[1024];
	snprintf(msg, sizeof(msg)-1, "Server message: Message number: %ld, Severity %ld, State: %ld, Line: %ld, Msg: '%s'",(long)srvmsg->msgnumber, (long)srvmsg->severity, (long)srvmsg->state, (long)srvmsg->line,srvmsg->text);
	fprintf(stderr, "%s\n",msg);
#endif	

	/*
	if (srvmsg->svrnlen > 0){fprintf(stderr, "Server '%s'\n", srvmsg->svrname);}
	if (srvmsg->proclen > 0){fprintf(stderr, " Procedure '%s'\n", srvmsg->proc);}
	*/

	if(srvmsg->severity>1){
		snprintf(sybsrvmsg, sizeof(sybsrvmsg)-1, "Message number: %ld, Severity %ld, State: %ld, Line: %ld, Msg: '%s'",(long)srvmsg->msgnumber, (long)srvmsg->severity, (long)srvmsg->state, (long)srvmsg->line,srvmsg->text);
	}
	
	return CS_SUCCEED;
}

static int
ex_execute_cmd(DB * db,CS_CHAR *cmdbuf)
{
	CS_RETCODE      retcode;
	CS_INT          restype;
	CS_COMMAND      *cmd;
	CS_RETCODE      query_code=CS_SUCCEED;

	if ((retcode = ct_cmd_alloc(db->conn, &cmd)) != CS_SUCCEED)
	{
		snprintf(db->errstr,sizeof(db->errstr)-1,"ct_cmd_alloc() failed: %s",sybsrvmsg);
		query_code=retcode;
		goto unlock_ret;
	}
	if ((retcode = ct_command(cmd, CS_LANG_CMD, cmdbuf, CS_NULLTERM,CS_UNUSED)) != CS_SUCCEED)
	{
		snprintf(db->errstr,sizeof(db->errstr)-1,"ct_command() failed: %s",sybsrvmsg);
		ct_cmd_drop(cmd);
		query_code=retcode;
		goto unlock_ret;
	}
	if ((retcode = ct_send(cmd)) != CS_SUCCEED)
	{
		snprintf(db->errstr,sizeof(db->errstr)-1,"ct_send() failed: %s",sybsrvmsg);
		ct_cmd_drop(cmd);
		query_code=retcode;
		goto unlock_ret;
	}
	while ((retcode = ct_results(cmd, &restype)) == CS_SUCCEED)
	{
		switch((int)restype)
                {
                    case CS_CMD_SUCCEED:
                    case CS_CMD_DONE:
                        break;

                    case CS_CMD_FAIL:
                        query_code = CS_FAIL;
                        break;

                    case CS_STATUS_RESULT:
                        retcode = ct_cancel(NULL, cmd, CS_CANCEL_CURRENT);
                        if (retcode != CS_SUCCEED)
                        {
                                snprintf(db->errstr,sizeof(db->errstr)-1,"ct_cancel() failed: %s",sybsrvmsg);
                                query_code = CS_FAIL;
                        }
                        break;

                    default:
                        query_code = CS_FAIL;
                        break;
                }
                if (query_code == CS_FAIL)
                {
                        retcode = ct_cancel(NULL, cmd, CS_CANCEL_ALL);
                        if (retcode != CS_SUCCEED)
                        {
                                snprintf(db->errstr,sizeof(db->errstr)-1,"ct_cancel() failed: %s",sybsrvmsg);
                        }
                        break;
                }
        }
        if (retcode == CS_END_RESULTS)
        {
                retcode = ct_cmd_drop(cmd);
                if (retcode != CS_SUCCEED)
                {
                        query_code = CS_FAIL;
                }
        }else{
                (void)ct_cmd_drop(cmd);
                query_code = CS_FAIL;
        }
       
unlock_ret:
        return query_code;
}


static int UdmCTLIBInitDB(DB * db)
{
	CS_RETCODE	retcode;
	CS_INT		netio_type = CS_SYNC_IO;
	CS_INT		len;
	CS_CHAR         *cmdbuf;

	retcode = cs_ctx_alloc(CS_VERSION_100, &db->ctx);
	if (retcode != CS_SUCCEED)
	{
		snprintf(db->errstr,sizeof(db->errstr)-1,"cs_ctx_alloc() failed: %s",sybsrvmsg);
		db->errcode=1;
		goto unlock_ex;
	}
	retcode = ct_init(db->ctx, CS_VERSION_100);
	if (retcode != CS_SUCCEED)
	{
		snprintf(db->errstr,sizeof(db->errstr)-1,"ex_init: ct_init() failed: %s",sybsrvmsg);
		cs_ctx_drop(db->ctx);
		db->ctx = NULL;
		db->errcode=1;
		goto unlock_ex;
	}

#ifdef EX_API_DEBUG
	retcode = ct_debug(db->ctx, NULL, CS_SET_FLAG, CS_DBG_API_STATES,NULL, CS_UNUSED);
	if (retcode != CS_SUCCEED)
	{
		sprintf(db->errstr,"ex_init: ct_debug() failed");
	}
#endif

	if (retcode == CS_SUCCEED)
	{
		retcode = ct_callback(db->ctx, NULL, CS_SET, CS_CLIENTMSG_CB,(CS_VOID *)ex_clientmsg_cb);
		if (retcode != CS_SUCCEED)
		{
			snprintf(db->errstr,sizeof(db->errstr)-1,"ct_callback(clientmsg) failed: %s",sybsrvmsg);
			db->errcode=1;
		}
	}
	if (retcode == CS_SUCCEED)
	{
		retcode = ct_callback(db->ctx, NULL, CS_SET, CS_SERVERMSG_CB,(CS_VOID *)ex_servermsg_cb);
		if (retcode != CS_SUCCEED)
		{
			snprintf(db->errstr,sizeof(db->errstr)-1,"ct_callback(servermsg) failed: %s",sybsrvmsg);
			db->errcode=1;
		}
	}
	if (retcode == CS_SUCCEED)
	{
		retcode = ct_config(db->ctx, CS_SET, CS_NETIO, &netio_type, CS_UNUSED, NULL);
		if (retcode != CS_SUCCEED)
		{
			sprintf(db->errstr,"ct_config(netio) failed: %s",sybsrvmsg);
			db->errcode=1;
		}
	}
	if (retcode != CS_SUCCEED)
	{
		ct_exit(db->ctx, CS_FORCE_EXIT);
		cs_ctx_drop(db->ctx);
		db->ctx = NULL;
		db->errcode=1;
		goto unlock_ex;
	}
	if (retcode==CS_SUCCEED){
		retcode = ct_con_alloc(db->ctx, &db->conn);
	}
	if (retcode != CS_SUCCEED)
	{
		snprintf(db->errstr,sizeof(db->errstr)-1,"ct_con_alloc failed: %s",sybsrvmsg);
		db->errcode=1;
		goto unlock_ex;
	}
	if (retcode == CS_SUCCEED && db->DBUser != NULL)
	{
		if ((retcode = ct_con_props(db->conn, CS_SET, CS_USERNAME, db->DBUser, CS_NULLTERM, NULL)) != CS_SUCCEED)
		{
			snprintf(db->errstr,sizeof(db->errstr)-1,"ct_con_props(username) failed: %s",sybsrvmsg);
			db->errcode=1;
		}
	}
	if (retcode == CS_SUCCEED && db->DBPass != NULL)
	{
		if ((retcode = ct_con_props(db->conn, CS_SET, CS_PASSWORD, db->DBPass, CS_NULLTERM, NULL)) != CS_SUCCEED)
		{
			snprintf(db->errstr,sizeof(db->errstr)-1,"ct_con_props(password) failed: %s",sybsrvmsg);
			db->errcode=1;
		}
	}
	if (retcode == CS_SUCCEED)
	{
		if ((retcode = ct_con_props(db->conn, CS_SET, CS_APPNAME, "indexer", CS_NULLTERM, NULL)) != CS_SUCCEED)
		{
			snprintf(db->errstr,sizeof(db->errstr)-1,"ct_con_props(appname) failed: %s",sybsrvmsg);
			db->errcode=1;
		}
	}
	if (retcode == CS_SUCCEED)
	{
		len = (db->DBHost == NULL) ? 0 : CS_NULLTERM;
		retcode = ct_connect(db->conn,db->DBHost,len);
		if (retcode != CS_SUCCEED)
		{
			snprintf(db->errstr,sizeof(db->errstr)-1,"ct_connect failed: %s",sybsrvmsg);
			db->errcode=1;
		}
	}
	if (retcode != CS_SUCCEED)
	{
		ct_con_drop(db->conn);
		db->conn = NULL;
		db->errcode=1;
		goto unlock_ex;
	}
	
	db->connected=1;
        cmdbuf = (CS_CHAR *) malloc(1024);
	sprintf(cmdbuf, "use %s\n", db->DBName);
	
	if ((retcode = ex_execute_cmd(db, cmdbuf)) != CS_SUCCEED)
	{
		snprintf(db->errstr,sizeof(db->errstr)-1,"ex_execute_cmd(use db) failed: %s",sybsrvmsg);
		db->errcode=1;
	}
	free(cmdbuf);
	
unlock_ex:
	return retcode;
}


static void UdmCTLIBCloseDB(DB * db)
{
	CS_RETCODE	retcode=CS_SUCCEED;
	CS_INT		close_option;
	CS_INT		exit_option;
	CS_RETCODE	status=CS_SUCCEED; /* FIXME: previous retcode was here */

	if(db->conn){
		close_option = (status != CS_SUCCEED) ? CS_FORCE_CLOSE : CS_UNUSED;
		retcode = ct_close(db->conn, close_option);
		if (retcode != CS_SUCCEED)
		{
			sprintf(db->errstr,"ex_con_cleanup: ct_close() failed");
			db->errcode=1;
			goto unlock_ret;
		}
		retcode = ct_con_drop(db->conn);
		if (retcode != CS_SUCCEED)
		{
			sprintf(db->errstr,"ex_con_cleanup: ct_con_drop() failed");
			db->errcode=1;
			goto unlock_ret;
		}
	}	
	if(db->ctx){
		exit_option = (status != CS_SUCCEED) ? CS_FORCE_EXIT : CS_UNUSED;
		retcode = ct_exit(db->ctx, exit_option);
		if (retcode != CS_SUCCEED)
		{
			sprintf(db->errstr,"ex_ctx_cleanup: ct_exit() failed");
			db->errcode=1;
			goto unlock_ret;
		}
		retcode = cs_ctx_drop(db->ctx);
		if (retcode != CS_SUCCEED)
		{
			sprintf(db->errstr,"ex_ctx_cleanup: cs_ctx_drop() failed");
			db->errcode=1;
			goto unlock_ret;
		}
	}
unlock_ret:
	return;
}

static CS_INT CS_PUBLIC
ex_display_dlen(CS_DATAFMT * column)
{
	CS_INT		len;

	switch ((int) column->datatype)
	{
		case CS_CHAR_TYPE:
		case CS_VARCHAR_TYPE:
		case CS_TEXT_TYPE:
		case CS_IMAGE_TYPE:
			len = MIN(column->maxlength, MAX_CHAR_BUF);
			break;

		case CS_BINARY_TYPE:
		case CS_VARBINARY_TYPE:
			len = MIN((2 * column->maxlength) + 2, MAX_CHAR_BUF);
			break;

		case CS_BIT_TYPE:
		case CS_TINYINT_TYPE:
			len = 3;
			break;

		case CS_SMALLINT_TYPE:
			len = 6;
			break;

		case CS_INT_TYPE:
			len = 11;
			break;

		case CS_REAL_TYPE:
		case CS_FLOAT_TYPE:
			len = 20;
			break;

		case CS_MONEY_TYPE:
		case CS_MONEY4_TYPE:
			len = 24;
			break;

		case CS_DATETIME_TYPE:
		case CS_DATETIME4_TYPE:
			len = 30;
			break;

		case CS_NUMERIC_TYPE:
		case CS_DECIMAL_TYPE:
			len = (CS_MAX_PREC + 2);
			break;

		default:
			len = 12;
			break;
	}
	return MAX((CS_INT)(strlen(column->name) + 1), len);
}

static UDB_RES *
ex_fetch_data(DB * db,CS_COMMAND *cmd)
{
	CS_RETCODE		retcode;
	CS_INT			num_cols;
	CS_INT			i;
	CS_INT			j;
	CS_INT			row_count = 0;
	CS_INT			rows_read;
	CS_DATAFMT		*datafmt;
	EX_COLUMN_DATA		*coldata;
	UDB_RES			*result=NULL;

	retcode = ct_res_info(cmd, CS_NUMDATA, &num_cols, CS_UNUSED, NULL);
	if (retcode != CS_SUCCEED)
	{
		sprintf(db->errstr,"ex_fetch_data: ct_res_info() failed");
		db->errcode=1;
		return NULL;
	}
	if (num_cols <= 0)
	{
		sprintf(db->errstr,"ex_fetch_data: ct_res_info() returned zero columns");
		db->errcode=1;
		return NULL;
	}

	coldata = (EX_COLUMN_DATA *)malloc(num_cols * sizeof (EX_COLUMN_DATA));
	datafmt = (CS_DATAFMT *)malloc(num_cols * sizeof (CS_DATAFMT));

	for (i = 0; i < num_cols; i++)
	{
		retcode = ct_describe(cmd, (i + 1), &datafmt[i]);
		if (retcode != CS_SUCCEED)
		{
			sprintf(db->errstr,"ex_fetch_data: ct_describe() failed");
			db->errcode=1;
			break;
		}
		datafmt[i].maxlength = ex_display_dlen(&datafmt[i]) + 1;
		datafmt[i].datatype = CS_CHAR_TYPE;
		datafmt[i].format   = CS_FMT_NULLTERM;
		coldata[i].value = (CS_CHAR *)malloc((size_t)datafmt[i].maxlength);

		retcode = ct_bind(cmd, (i + 1), &datafmt[i],coldata[i].value, &coldata[i].valuelen,(CS_SMALLINT *)&coldata[i].indicator);

		if (retcode != CS_SUCCEED)
		{
			sprintf(db->errstr,"ex_fetch_data: ct_bind() failed");
			db->errcode=1;
			break;
		}
	}
	if (retcode != CS_SUCCEED)
	{
		for (j = 0; j < i; j++)
		{
			free(coldata[j].value);
		}
		free(coldata);
		free(datafmt);
		return NULL;
	}

	result=(UDB_RES*)malloc(sizeof(UDB_RES));
	result->nRows=0;
	result->nCols=num_cols;
	result->items=NULL;

	while (((retcode = ct_fetch(cmd, CS_UNUSED, CS_UNUSED, CS_UNUSED,&rows_read)) == CS_SUCCEED) || (retcode == CS_ROW_FAIL))
	{
		row_count = row_count + rows_read;
		if (retcode == CS_ROW_FAIL)
		{
			sprintf(db->errstr, "Error on row %d.\n", (int)row_count);
			db->errcode=1;
		}
		result->nRows++;
		result->items=(char**)UdmXrealloc(result->items,result->nRows*result->nCols*sizeof(char*));
		
		for (i = 0; i < num_cols; i++)
		{
			char * p;
			p=coldata[i].value;
			UdmTrim(p," ");
			sql_val(result,(result->nRows-1),(i))=strdup(p?p:"");
		} 
	}

	for (i = 0; i < num_cols; i++)
	{
		free(coldata[i].value);
	}
	free(coldata);
	free(datafmt);

	switch ((int)retcode)
	{
		case CS_END_DATA:
			retcode = CS_SUCCEED;
			break;

		case CS_FAIL:
			sprintf(db->errstr,"ex_fetch_data: ct_fetch() failed");
			db->errcode=1;
			return NULL;
			break;

		default:
			sprintf(db->errstr,"ex_fetch_data: ct_fetch() returned an expected retcode");
			db->errcode=1;
			return NULL;
			break;
	}
	return result;
}


static UDB_RES *
UdmCTLIBQuery(DB * db, const char * query)
{
	CS_RETCODE	retcode;
	CS_COMMAND	*cmd;
	CS_INT		res_type;
	UDB_RES		*res=NULL;

	if(!db->connected){
		UdmCTLIBInitDB(db);
		if(db->errcode)
			goto unlock_ret;
	}

        if ((retcode = ct_cmd_alloc(db->conn, &cmd)) != CS_SUCCEED)
        {
                sprintf(db->errstr,"ct_cmd_alloc() failed");
                db->errcode=1;
                goto unlock_ret;
        }

	retcode = ct_command(cmd, CS_LANG_CMD, query, CS_NULLTERM, CS_UNUSED);
	if (retcode != CS_SUCCEED)
	{
                sprintf(db->errstr,"ct_command() failed");
                db->errcode=1;
                goto unlock_ret;
	}

	if (ct_send(cmd) != CS_SUCCEED)
	{
                sprintf(db->errstr,"ct_send() failed");
                db->errcode=1;
                goto unlock_ret;
	}

	while ((retcode = ct_results(cmd, &res_type)) == CS_SUCCEED)
	{
		switch ((int)res_type)
		{
		case CS_CMD_SUCCEED:
		case CS_CMD_DONE:
   			break;

		case CS_CMD_FAIL:
			/* FIXME add mutex */
			if(!strstr(sybsrvmsg,"UNIQUE KEY")){
				snprintf(db->errstr,sizeof(db->errstr)-1,"%s",sybsrvmsg);
				db->errcode=1;
			}
			break;

		case CS_ROW_RESULT:
			res = ex_fetch_data(db,cmd);
			if (retcode != CS_SUCCEED)
			{
				sprintf(db->errstr,"ex_fetch_data() failed");
				db->errcode=1;
				goto unlock_ret;
			}
			break;

		case CS_COMPUTE_RESULT:
			/* FIXME: add checking */
			break; 

		default:
			sprintf(db->errstr,"DoCompute: ct_results() returned unexpected result type");
			db->errcode=1;
			goto unlock_ret;
		}
	}

	switch ((int)retcode)
	{
		case CS_END_RESULTS:
			break;

		case CS_FAIL:
			sprintf(db->errstr,"DoCompute: ct_results() failed");
			db->errcode=1;
			goto unlock_ret;

		default:
			sprintf(db->errstr,"DoCompute: ct_results() returned unexpected result code");
			db->errcode=1;
			goto unlock_ret;
	}

	if ((retcode = ct_cmd_drop(cmd)) != CS_SUCCEED)
	{
                sprintf(db->errstr,"DoCompute: ct_cmd_drop() failed");
                db->errcode=1;
                if(res){
                	udb_free_result(res);
                	res=NULL;
                }
                goto unlock_ret;
	}

unlock_ret:

	return res;
}

#endif


/***********************************************************************/

/*
 *   Wrappers for different databases
 *
 *   UdmDBEscStr()
 *   UdmAllocDB()
 *   UdmFreeDB()
 *   UdmDBErrorCode()
 *   UdmSetDBErrorCode();
 *   UdmDBErrorMsg()
 *   UdmSQLQuery()
 *   UdmSQLValue();
 *   UdmSQLFree();
 */


static char * UdmDBEscStr(int DBType,char *to,const char *from){
	char *s;
	if(!from)return(NULL);
	if(!to)to=(char*)malloc(strlen(from)*2+1);

#ifdef HAVE_MYSQL
	if(DBType==UDM_DB_MYSQL){
		mysql_escape_string(to,from,strlen(from));
		return(to);
	}
#endif
	s=to;
	if (DBType == UDM_DB_ORACLE || DBType == UDM_DB_ORACLE8  ||
	    DBType == UDM_DB_MSSQL || DBType == UDM_DB_ORACLE7 || 
	    DBType == UDM_DB_IBASE || DBType == UDM_DB_SAPDB)
	 {
	    while(*from){
		switch(*from){
			case '\'':
				/* Note that no break here!*/
				*to=*from;to++;
			default:
				*to=*from;
		}
		to++;from++;
	    }
	} else {
	    while(*from){
		switch(*from){
			case '\'':
			case '\\':
				*to='\\';to++;
			default:*to=*from;
		}
		to++;from++;
	    }
	}
	*to=0;return(s);
}


void * UdmAllocDB(UDM_AGENT * Indexer,const char * dbaddr,int mode){
	DB * db;

	db=(DB*)malloc(sizeof(DB));
	bzero(db,sizeof(*db));

#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT || HAVE_EASYSOFT || HAVE_SAPDB)
	db->hDbc=SQL_NULL_HDBC;
	db->hEnv=SQL_NULL_HENV;
	db->hstmt=SQL_NULL_HSTMT;
#endif
#if (HAVE_IBASE)
	db->DBH=NULL;
#endif
#if (HAVE_ORACLE8)
	db->par = UdmXmalloc(sizeof(struct param_struct));
#endif
	UdmSetDBAddr(db,dbaddr?dbaddr:"mysql://localhost/mnogosearch/");
	return((void*)(db));
}

void UdmFreeDB(void * vdb){
	DB * db=vdb;
	if(db->connected){
		db->connected=0;
#if HAVE_MYSQL
		if(db->DBDriver==UDM_DB_MYSQL){
			mysql_close(&db->mysql);
			free(db);
		}
		return;
#endif

#if HAVE_PGSQL
		if(db->DBDriver==UDM_DB_PGSQL){
			PQfinish(db->pgsql);
			free(db);
		}
		return;
#endif

#if HAVE_MSQL
		if(db->DBDriver==UDM_DB_MSQL){
			msqlClose(db->msql);
			free(db);
		}
		return;
#endif

#if HAVE_IBASE
		if(db->DBDriver==UDM_DB_IBASE){
			UdmIBaseCloseDB(db);
			free(db);
		}
		return;
#endif

#if HAVE_ORACLE8
		if(db->DBDriver==UDM_DB_ORACLE8){
			UdmOracle8CloseDB(db);
			free(db);
		}
		return;
#endif

#if HAVE_ORACLE7
		if(db->DBDriver==UDM_DB_ORACLE7){
			UdmOracle7CloseDB(db);
			free(db);
		}
		return;
#endif

#if HAVE_CTLIB
		if(db->DBDriver==UDM_DB_MSSQL){
			UdmCTLIBCloseDB(db);
			free(db);
		}
		return;
#endif

#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT || HAVE_EASYSOFT || HAVE_SAPDB)
		UdmODBCCloseDB(db);
		free(db);
		return;
#endif

	}
	free(db);
}
int UdmDBErrorCode(void *db){
	return(((DB*)(db))->errcode);
}
void UdmSetDBErrorCode(void * db,int code){
	((DB*)db)->errcode=code;
}
char * UdmDBErrorMsg(void *db){
	return(((DB*)(db))->errstr);
}

static UDB_RES * UdmSQLQuery(DB * db,const char * query){
	UDB_RES * res=NULL;

#ifdef DEBUG_SQL
	unsigned long ticks;
	ticks=UdmStartTimer();
#endif
	
#if HAVE_MYSQL
	if(db->DBDriver==UDM_DB_MYSQL){
		res=(UDB_RES*)malloc(sizeof(*res));
		bzero(res,sizeof(*res));
		res->mysqlres=UdmMySQLQuery(db,query);
		res->lmysqlrow=0;
		if(!res->mysqlres){
			free(res);
			res=NULL;
		}else{
			res->DBDriver=db->DBDriver;
		}
		goto ret;
	}
#endif

#if HAVE_PGSQL
	if(db->DBDriver==UDM_DB_PGSQL){
		res=(UDB_RES*)malloc(sizeof(*res));
		bzero(res,sizeof(*res));
		res->pgsqlres=UdmPgSQLQuery(db,query);
		if(!res->pgsqlres){
			free(res);
			res=NULL;
		}else{
			res->DBDriver=db->DBDriver;
		}
		goto ret;
	}
#endif

#if HAVE_MSQL
	if(db->DBDriver==UDM_DB_MSQL){
		res=(UDB_RES*)malloc(sizeof(*res));
		bzero(res,sizeof(*res));
		res->msqlres=UdmMSQLQuery(db,query);
		if(!res->msqlres){
			free(res);
			res=NULL;
		}else{
			res->DBDriver=db->DBDriver;
		}
		goto ret;
	}
#endif


#if HAVE_IBASE
	if(db->DBDriver==UDM_DB_IBASE){
		res=UdmIBaseQuery(db,query);
		if(res)res->DBDriver=db->DBDriver;
		goto ret;
	}
#endif

#if HAVE_ORACLE8
	if(db->DBDriver==UDM_DB_ORACLE8){
		res=UdmOracle8Query(db,query);
		if(res)res->DBDriver=db->DBDriver;
		goro ret;
	}
#endif

#if HAVE_ORACLE7
	if(db->DBDriver==UDM_DB_ORACLE7){
		res=UdmOracle7Query(db,query);
		if(res)res->DBDriver=db->DBDriver;
		return res;
	}
#endif

#if HAVE_CTLIB
	if(db->DBDriver==UDM_DB_MSSQL){
		res=UdmCTLIBQuery(db,query);
		if(res)res->DBDriver=db->DBDriver;
		goto ret;
	}
#endif

#if (HAVE_IODBC || HAVE_UNIXODBC || HAVE_SOLID || HAVE_VIRT || HAVE_EASYSOFT || HAVE_SAPDB)
	{	/* FIXME: add driver checking */
		res=UdmODBCQuery(db,query);
		if(res)res->DBDriver=db->DBDriver;
		goto ret;
	}
#endif
	db->errcode=1;
	snprintf(db->errstr,sizeof(db->errstr)-1,"Unsupported SQL database type");
ret:
#ifdef DEBUG_SQL
	ticks=UdmStartTimer()-ticks;
	fprintf(stderr,"[%d] SQL %.2fs: %s\n",getpid(),(float)ticks/1000,query);
#endif

	return res;
}

static size_t UdmSQLNumRows(UDB_RES * res){
#if HAVE_MYSQL
	if(res->DBDriver==UDM_DB_MYSQL)
		return(mysql_num_rows(res->mysqlres));
#endif

#if HAVE_PGSQL
	if(res->DBDriver==UDM_DB_PGSQL)
		return(PQntuples(res->pgsqlres));
#endif

#if HAVE_MSQL
	if(res->DBDriver==UDM_DB_MSQL)
		return(msqlNumRows(res->msqlres));
#endif
	return(res?res->nRows:0);
}

static const char * UdmSQLValue(UDB_RES * res,size_t i,size_t j){
	const char * v;
#if HAVE_MYSQL
	if(res->DBDriver==UDM_DB_MYSQL){
		if(i>=mysql_num_rows(res->mysqlres))return(NULL);
		
		if(res->lmysqlrow==i){
			if(res->lmysqlrow==0){
				mysql_data_seek(res->mysqlres,(my_ulonglong)0);
				res->mysqlrow=mysql_fetch_row(res->mysqlres);
			}else{
				/* Nothing */
			}
		}else
		if(res->lmysqlrow+1==i){
			res->mysqlrow=mysql_fetch_row(res->mysqlres);
		}else{
			mysql_data_seek(res->mysqlres,(my_ulonglong)i);
			res->mysqlrow=mysql_fetch_row(res->mysqlres);
		}
		res->lmysqlrow=i;
		if(res->mysqlrow)return(res->mysqlrow[j]?res->mysqlrow[j]:"");
		else	return(NULL);
	}
#endif

#if HAVE_PGSQL
	if(res->DBDriver==UDM_DB_PGSQL){
		return(PQgetvalue(res->pgsqlres,(int)(i),(int)(j)));
	}
#endif

#if HAVE_MSQL
	if(res->DBDriver==UDM_DB_MSQL){
		return(msql_value(res->msqlres,i,j));
	}
#endif
	v=sql_value(res,i,j);
	return v;

}

static void UdmSQLFree(UDB_RES * res){
#if HAVE_MYSQL
	if(res->DBDriver==UDM_DB_MYSQL){
		mysql_free_result(res->mysqlres);
		free(res);
		return;
	}
#endif

#if HAVE_PGSQL
	if(res->DBDriver==UDM_DB_PGSQL){
		PQclear(res->pgsqlres);
		free(res);
		return;
	}
#endif

#if HAVE_MSQL
	if(res->DBDriver==UDM_DB_MSQL){
		msqlFreeResult(res->msqlres);
		free(res);
		return;
	}
#endif

#if HAVE_ORACLE7
	if(res->DBDriver==UDM_DB_ORACLE7){
		oci_free_result(x)
		free(res);
		return;
	}
#endif

#if HAVE_ORACLE8
	if(res->DBDriver==UDM_DB_ORACLE8){
		oci_free_result(x);
		free(res);
		return;
	}
#endif
	udb_free_result(res);
	free(res);
}



/********************************************************************/
/* --------------------------------------------------------------- */
/* Almost Unified code for all supported SQL backends              */
/* --------------------------------------------------------------- */
/*******************************************************************/


static const int search_cache_size=1000;


static time_t now(void){ /* FIXME !!! is it reentrant? */
time_t tloc;
	return(time(&tloc));
}


/******************** Limits stuff *********************************/

__INDLIB__ int UdmAddTagLimit(UDM_ENV * Conf,char * tag){
	strcpy(Conf->taglimit,tag?tag:"");
	if(Conf->tagstr[0])strcpy(UDM_STREND(Conf->tagstr)-1," OR ");
	else	strcat(Conf->tagstr," AND (");
/*
	FIXME
	if(Conf->DBType==UDM_DB_PGSQL)
		sprintf(UDM_STREND(Conf->tagstr),"(url.tag || '') LIKE '%s')",tag);
	else
*/
		sprintf(UDM_STREND(Conf->tagstr),"url.tag LIKE '%s')",tag);
	return(0);
}

__INDLIB__ int UdmAddStatusLimit(UDM_ENV *c,int status){
/*
	FIXME
	if(c->DBSQL_IN){
		if(c->statusstr[0])sprintf(UDM_STREND(c->statusstr)-1,",%d)",status);
		else	sprintf(c->statusstr," AND url.status IN (%d)",status);
	}else
*/
	{
		if(c->statusstr[0])strcpy(UDM_STREND(c->statusstr)-1," OR ");
		else	strcat(c->statusstr," AND (");
		sprintf(UDM_STREND(c->statusstr),"url.status=%d)",status);
	}
	return(0);
}

__INDLIB__ int UdmAddURLLimit(UDM_ENV *c,char *URL){
	UDM_URL urlstruct;
	char site_id_str[512];

	UdmParseURL(&urlstruct,URL);
	if(urlstruct.schema[0]&&urlstruct.hostinfo[0]){
		snprintf(site_id_str,sizeof(site_id_str)-1,"%s://%s/",urlstruct.schema,urlstruct.hostinfo);
		c->sitelimit=UdmStrCRC32(site_id_str);
	}else{
		c->sitelimit=0;
	}
	if(c->urlstr[0])strcpy(UDM_STREND(c->urlstr)-1," OR ");
	else	strcat(c->urlstr," AND (");
/*
	FIXME
	if(c->DBType==UDM_DB_PGSQL)
		sprintf(UDM_STREND(c->urlstr),"(url.url || '') LIKE '%s')",URL);
	else
*/
		sprintf(UDM_STREND(c->urlstr),"url.url LIKE '%s')",URL);
	return(0);
}

__INDLIB__ int UdmAddLangLimit(UDM_ENV *c,char * lang){
	if(c->langstr[0])strcpy(UDM_STREND(c->langstr)-1," OR ");
	else	strcat(c->langstr," AND (");
	sprintf(UDM_STREND(c->langstr),"url.lang LIKE '%s')",lang);
	return(0);
}

__INDLIB__ int UdmAddCatLimit(UDM_ENV *c,char * cat){
	strcpy(c->catlimit,cat?cat:"");
	if(c->catstr[0])strcpy(UDM_STREND(c->catstr)-1," OR ");
	else    strcat(c->catstr," AND (");
	sprintf(UDM_STREND(c->catstr),"url.category LIKE '%s%%')",cat);
	return(0);
}



__INDLIB__ int UdmAddTimeLimit(UDM_ENV *c,struct udm_stl_info_t * stl){
	char lt_gt;
	
	switch(stl->type){
		case 1: case -1:
			if (stl->type==1)
				lt_gt='>';
			else
				lt_gt='<';
			sprintf(c->timestr, " AND (url.last_mod_time%c%li)", lt_gt, stl->t1);
			break;
		case 2: /* between */
			sprintf(c->timestr, " AND (url.last_mod_time BETWEEN %li AND %li)", stl->t1, stl->t2);
			break;
		default:
			return -1;
	}
	return(0);
}

__INDLIB__ int UdmClearLimits(UDM_ENV *c){
	c->urlstr[0]='\0';
	c->tagstr[0]='\0';
	c->statusstr[0]='\0';
	c->langstr[0]='\0';
	c->timestr[0]='\0';
	c->catstr[0]='\0';
	c->catlimit[0]='\0';
	c->taglimit[0]='\0';
	c->sitelimit=0;
	return(0);
}
__INDLIB__ int UdmClearURLLimit(UDM_ENV *c){
	c->urlstr[0]=0;
	return(0);
}
/************* Servers ******************************************/

int UdmLoadServerTable(UDM_AGENT * Indexer, char * name, int flags){
	size_t rows,i;
	UDM_SERVER Server;
	UDB_RES * res;
	char qbuf[1024];
	DB * db=(DB*)Indexer->db;

	snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT rec_id,url,period,tag,category,charset,lang,basic_auth,\
proxy,proxy_port,proxy_auth,\
maxhops,gindex,follow,deletebad,use_robots,delete_no_srv,use_clones,\
descweight,keywordweight,titleweight,bodyweight,\
urlweight,urlhostweight,urlpathweight,urlfileweight,\
correct_factor,incorrect_factor,number_factor,alnum_factor,\
max_net_errors,net_delay_time,read_timeout \
FROM %s WHERE active=1",name);

	res=UdmSQLQuery(db,qbuf);
	if(UdmDBErrorCode(Indexer->db)){
		return(IND_ERROR);
	}
	rows=UdmSQLNumRows(res);
	for(i=0;i<rows;i++){
		UdmInitServer(&Server);

		Server.url		=strdup(UDM_EMPTY2NULL(UdmSQLValue(res,i,1)));
		Server.period		=UDM_ATOI(UdmSQLValue(res,i,2));
		Server.tag		=strdup(UDM_EMPTY2NULL(UdmSQLValue(res,i,3)));
		Server.category		=strdup(UDM_EMPTY2NULL(UdmSQLValue(res,i,4)));
		Server.charset		=strdup(UDM_EMPTY2NULL(UdmSQLValue(res,i,5)));
		Server.lang		=strdup(UDM_NULL2EMPTY(UdmSQLValue(res,i,6)));
		Server.basic_auth	=strdup(UDM_NULL2EMPTY(UdmSQLValue(res,i,7)));
		
		if(Server.basic_auth){
			char * basic_auth;
			
			basic_auth=(char*) UdmXmalloc(BASE64_LEN(strlen(Server.basic_auth)));
			udm_base64_encode(Server.basic_auth,basic_auth,strlen(Server.basic_auth));
			Server.basic_auth=basic_auth;
		}

		Server.proxy		=strdup(UDM_EMPTY2NULL(UdmSQLValue(res,i,8)));
		Server.proxy_port	=UDM_ATOI(UdmSQLValue(res,i,9));
		Server.proxy_basic_auth	=strdup(UDM_EMPTY2NULL(UdmSQLValue(res,i,10)));
		Server.maxhops		=UDM_ATOI(UdmSQLValue(res,i,11));
		Server.index		=UDM_ATOI(UdmSQLValue(res,i,12));
		Server.follow		=UDM_ATOI(UdmSQLValue(res,i,13));
		Server.deletebad	=UDM_ATOI(UdmSQLValue(res,i,14));
		Server.use_robots	=UDM_ATOI(UdmSQLValue(res,i,15));
		Server.delete_no_server	=UDM_ATOI(UdmSQLValue(res,i,16));
		Server.use_clones	=UDM_ATOI(UdmSQLValue(res,i,17));
		Server.desc_section	=UDM_ATOI(UdmSQLValue(res,i,18));
		Server.kwrd_section	=UDM_ATOI(UdmSQLValue(res,i,19));
		Server.title_section	=UDM_ATOI(UdmSQLValue(res,i,20));
		Server.body_section	=UDM_ATOI(UdmSQLValue(res,i,21));
		/*Server.urlweight	=UDM_ATOI(UdmSQLValue(res,i,22));*/
		Server.host_section	=UDM_ATOI(UdmSQLValue(res,i,23));
		Server.path_section	=UDM_ATOI(UdmSQLValue(res,i,24));
		Server.file_section	=UDM_ATOI(UdmSQLValue(res,i,25));
		Server.correct_factor	=UDM_ATOI(UdmSQLValue(res,i,26));
		Server.incorrect_factor	=UDM_ATOI(UdmSQLValue(res,i,27));
		Server.number_factor	=UDM_ATOI(UdmSQLValue(res,i,28));
		Server.alnum_factor	=UDM_ATOI(UdmSQLValue(res,i,29));
		Server.max_net_errors	=UDM_ATOI(UdmSQLValue(res,i,30));
		Server.net_error_delay_time=UDM_ATOI(UdmSQLValue(res,i,31));
		Server.read_timeout	=UDM_ATOI(UdmSQLValue(res,i,32));
		        
		UdmAddServer(Indexer->Conf,&Server,flags,UDM_SERVER_SUBSTR|UDM_SERVER_MATCH);
		
		UDM_FREE(Server.url);
		UDM_FREE(Server.tag);
		UDM_FREE(Server.category);
		UDM_FREE(Server.charset);
		UDM_FREE(Server.lang);
		UDM_FREE(Server.basic_auth);
	}
	UdmSQLFree(res);
	return(IND_OK);
}

/************************* find url ********************************/

static int UdmFindURL(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc){
	UDB_RES * res;
	
	if(Indexer->Conf->use_crc32_url_id){
		/* Auto generation of rec_id */
		/* using CRC32 algorythm     */
		Doc->url_id=UdmStrCRC32(Doc->url);
	}else{
		size_t i;
		const char *o;
		char qbuf[1024];
		char e_url[UDM_URLSIZE*4];
		DB * db=(DB*)Indexer->db;

		/* Escape URL string */
		UdmDBEscStr(db->DBType,e_url,Doc->url);
		sprintf(qbuf,"SELECT rec_id FROM url WHERE url='%s'",e_url);
		res=UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
		for(i=0;i<UdmSQLNumRows(res);i++){
			if((o=UdmSQLValue(res,i,0))){
				Doc->url_id=atoi(o);
				break;
			}
		}
		UdmSQLFree(res);
	}
	return(IND_OK);
}


static int UdmFindMessage(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc){
	size_t i;
	char qbuf[1024];
	char eid[UDM_URLSIZE*4];
	UDB_RES * res;
	DB * db=(DB*)Indexer->db;

	/* Escape URL string */
	UdmDBEscStr(db->DBType,eid,Doc->message_id);
	
	sprintf(qbuf,"SELECT rec_id FROM url WHERE msg_id='%s'",eid);
	res=UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	for(i=0;i<UdmSQLNumRows(res);i++){
		const char * o;
		if((o=UdmSQLValue(res,i,0))){
			Doc->url_id=atoi(o);
			break;
		}
	}
	UdmSQLFree(res);
	return(IND_OK);
}


/********************** Words ***********************************/

static int UdmDeleteWordFromURL(UDM_AGENT *Indexer,UDM_DOCUMENT * Doc){
	char qbuf[512];
	int i,last=0;

	switch(Indexer->Conf->DBMode){
	
	case UDM_DBMODE_MULTI:
		for(i=MINDICT;i<MAXDICT;i++){
			if(last!=DICTNUM(i)){
				snprintf(qbuf,sizeof(qbuf),"DELETE FROM dict%d WHERE url_id=%d",DICTNUM(i),Doc->url_id);
				UdmSQLQuery(Indexer->db,qbuf);
				if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
				last=DICTNUM(i);
			}
		}
		break;
	case UDM_DBMODE_MULTI_CRC:
		for(i=MINDICT;i<MAXDICT;i++){
			if(last!=DICTNUM(i)){
#ifdef HAVE_ORACLE8
				snprintf(qbuf,sizeof(qbuf)-1,"DELETE FROM ndict%d WHERE url_id=:1", DICTNUM(i));
					param_init(Indexer->db, 1, 0, 0);
					param_add(Indexer->db, Doc->url_id, 0, 0);
#else
				snprintf(qbuf,sizeof(qbuf)-1,"DELETE FROM ndict%d WHERE url_id=%d",DICTNUM(i),Doc->url_id);
#endif
				UdmSQLQuery(Indexer->db,qbuf);
				if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
				last=DICTNUM(i);
			}
		}
		break;
	case UDM_DBMODE_SINGLE_CRC:
		snprintf(qbuf,sizeof(qbuf)-1,"DELETE FROM ndict WHERE url_id=%d",Doc->url_id);
		UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
		break;
	case UDM_DBMODE_BLOB:
		/* Let's lock it */
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_CACHE);
		i=UdmBlobWriteStamp(Indexer,Doc);
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_CACHE);
		return(i);
		break;
	case UDM_DBMODE_CACHE:
		/* Let's lock it */
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_CACHE);
		i=UdmDeleteURLFromCache(Indexer,Doc->url_id);
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_CACHE);
		return(i);
		break;
	default:  /* UDM_DBMODE_SINGLE */
		sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d",Doc->url_id);
		UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
		break;
	}
	return(IND_OK);
}

static int StoreWordsMulti(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	char qbuf[512];
	char tablename[64]="dict";
	char tbl_nm[64];
	int  n,prev_dictlen=0;
	DB * db=(DB*)Indexer->db;

	if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI){
		strcpy(tablename,"dict");
	}else{
		strcpy(tablename,"ndict");
	}

	for(n=0;n<NDICTS;n++){
		if(prev_dictlen==dictlen[n])continue;
		prev_dictlen=dictlen[n];

		sprintf(tbl_nm, "%s%d", tablename, dictlen[n]);
		if(Indexer->Conf->DBUseLock){
			switch(db->DBType){
				case UDM_DB_PGSQL:
					UdmSQLQuery(Indexer->db,"BEGIN WORK");
					break;
				case UDM_DB_ORACLE:
				case UDM_DB_SAPDB:
					UdmSQLQuery(Indexer->db,"COMMIT");
					((DB*)(Indexer->db))->commit_fl = 1;
					break;
				default:
					break;
			}
			if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
		}

		/* Delete old words */
		snprintf(qbuf,sizeof(qbuf),"DELETE FROM %s WHERE url_id=%d",tbl_nm,Doc->url_id);
		UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);

		/* Insert new word */
		if(db->DBType==UDM_DB_MYSQL){
			int have_words=0;
			char * qb,*qe;
			size_t step=1024,mlen=1024,len,i;

			qb=(char*)malloc(mlen);
			if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI){
				sprintf(qb,"INSERT INTO %s (url_id,word,intag) VALUES ",tbl_nm);
			}else{
				sprintf(qb,"INSERT INTO %s (url_id,word_id,intag) VALUES ",tbl_nm);
			}
			qe=qb+strlen(qb);

			for(i=0;i<Doc->WordList.nwords;i++){
				if (DICTNUM(strlen(Doc->WordList.Word[i].word)) == dictlen[n]){
					len=qe-qb;
					/* UDM_MAXWORDSIZE+100 should be enough */
					if((len+UDM_MAXWORDSIZE+100)>=mlen){
						mlen+=step;
						qb=(char*)realloc(qb,mlen);
						qe=qb+len;
					}
					if(have_words)strcpy(qe++,",");
					have_words++;
					if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI){
						sprintf(qe,"(%d,'%s',%d)",Doc->url_id,Doc->WordList.Word[i].word,Doc->WordList.Word[i].coord);
					}else{
						sprintf(qe,"(%d,%d,%d)",Doc->url_id,UdmStrCRC32(Doc->WordList.Word[i].word),Doc->WordList.Word[i].coord);
					}
					qe=qe+strlen(qe);
					
					/* Insert 64k packets max to stay */
					/* under MySQL's default limit. */
					if((qe-qb)>=64*1024){
						UdmSQLQuery(Indexer->db,qb);
						if(UdmDBErrorCode(Indexer->db)){
							free(qb);
							return(IND_ERROR);
						}
						/* Init query again */
						if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI){
							sprintf(qb,"INSERT INTO %s (url_id,word,intag) VALUES ",tbl_nm);
						}else{
							sprintf(qb,"INSERT INTO %s (url_id,word_id,intag) VALUES ",tbl_nm);
						}
						qe=qb+strlen(qb);
						have_words = 0;
					}
				}
			}
			if(have_words){
				UdmSQLQuery(Indexer->db,qb);
				if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
			}
			free(qb);
		}else{
			size_t i;
			for(i=0;i<Doc->WordList.nwords;i++){
				int len;
				len=strlen(Doc->WordList.Word[i].word);
				if (DICTNUM(len) == dictlen[n]){
					if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI){
						snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO %s (url_id,word,intag) VALUES(%d,'%s',%d)",tbl_nm,Doc->url_id,Doc->WordList.Word[i].word,Doc->WordList.Word[i].coord);
					}else{
						snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO %s (url_id,word_id,intag) VALUES(%d,%d,%d)",tbl_nm,Doc->url_id,UdmStrCRC32(Doc->WordList.Word[i].word),Doc->WordList.Word[i].coord);
					}
					UdmSQLQuery(Indexer->db,qbuf);
					if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
				}
			}
		}

		if(Indexer->Conf->DBUseLock){
			switch(db->DBType){
				case UDM_DB_PGSQL:
					UdmSQLQuery(Indexer->db,"END WORK");
					break;
				case UDM_DB_ORACLE:
				case UDM_DB_SAPDB:
					UdmSQLQuery(Indexer->db,"COMMIT");
					((DB*)(Indexer->db))->commit_fl = 0;
					break;
				default:
					break;
			}
			if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
		}
	}
	return(IND_OK);
}


static int StoreWordsSingle(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	size_t i;
	char qbuf[256]="";
	time_t stmp;
	DB * db=(DB*)Indexer->db;

	stmp=time(NULL);

	/* Start transaction if supported */
	/* This is to make stuff faster   */

	if(Indexer->Conf->DBUseLock){
		switch(db->DBType){
			case UDM_DB_PGSQL:
				UdmSQLQuery(Indexer->db,"BEGIN WORK");
				break;
			case UDM_DB_MSSQL:
				UdmSQLQuery(Indexer->db,"BEGIN TRANSACTION");
				break;
			case UDM_DB_ORACLE:
			case UDM_DB_SAPDB:
				UdmSQLQuery(Indexer->db,"COMMIT");
				((DB*)(Indexer->db))->commit_fl = 1;
				break;
			case UDM_DB_IBASE:
				UdmSQLQuery(Indexer->db,"BEGIN");
				((DB*)(Indexer->db))->commit_fl = 1;
				break;
			default:
				break;
		}
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	}

	/* Delete old words */
	if(Indexer->Conf->DBMode==UDM_DBMODE_SINGLE){
		sprintf(qbuf,"DELETE FROM dict WHERE url_id=%d",Doc->url_id);
	}else
	if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI){
		sprintf(qbuf,"DELETE FROM ndict WHERE url_id=%d",Doc->url_id);
	}

	UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);

	/* Insert new words */
	if(db->DBType==UDM_DB_MYSQL){
		if(Doc->WordList.nwords){
			char * qb,*qe;
			size_t step=1024,mlen=1024,len;

			qb=(char*)malloc(mlen);
			if(Indexer->Conf->DBMode==UDM_DBMODE_SINGLE){
				strcpy(qb,"INSERT INTO dict (url_id,word,intag) VALUES ");
			}else
			if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI){
				strcpy(qb,"INSERT INTO ndict (url_id,word_id,intag) VALUES ");
			}
			qe=qb+strlen(qb);

			for(i=0;i<Doc->WordList.nwords;i++){
				len=qe-qb;
				/* UDM_MAXWORDSIZE+100 should be enough */
				if((len+UDM_MAXWORDSIZE+100)>=mlen){
					mlen+=step;
					qb=(char*)realloc(qb,mlen);
					qe=qb+len;
				}
				if(i>0)strcpy(qe++,",");
				if(Indexer->Conf->DBMode==UDM_DBMODE_SINGLE){
					sprintf(qe,"(%d,'%s',%d)",Doc->url_id,Doc->WordList.Word[i].word,Doc->WordList.Word[i].coord);
				}else
				if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI){
					sprintf(qe,"(%d,%d,%d)",Doc->url_id,UdmStrCRC32(Doc->WordList.Word[i].word),Doc->WordList.Word[i].coord);
				}else
				if(Indexer->Conf->DBMode==UDM_DBMODE_BLOB){
					sprintf(qe,"('%s',%d,%d,%li)",Doc->WordList.Word[i].word,Doc->url_id,Doc->WordList.Word[i].coord,stmp);
				}
				qe=qe+strlen(qe);
			}
			UdmSQLQuery(Indexer->db,qb);
			if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
			free(qb);
		}
	}else{
		for(i=0;i<Doc->WordList.nwords;i++){
			if(Indexer->Conf->DBMode==UDM_DBMODE_SINGLE){
				sprintf(qbuf,"INSERT INTO dict (url_id,word,intag) VALUES(%d,'%s',%d)",Doc->url_id,Doc->WordList.Word[i].word,Doc->WordList.Word[i].coord);
			}else{
				sprintf(qbuf,"INSERT INTO ndict (url_id,word_id,intag) VALUES(%d,%d,%d)",Doc->url_id,UdmStrCRC32(Doc->WordList.Word[i].word),Doc->WordList.Word[i].coord);
			}
			UdmSQLQuery(Indexer->db,qbuf);
			if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
		}
	}

	/* Commit */
	if(Indexer->Conf->DBUseLock){
		switch(db->DBType){
			case UDM_DB_PGSQL:
				UdmSQLQuery(Indexer->db,"END WORK");
				break;
			case UDM_DB_MSSQL:
				UdmSQLQuery(Indexer->db,"COMMIT");
				break;
			case UDM_DB_ORACLE:
			case UDM_DB_SAPDB:
				((DB*)(Indexer->db))->commit_fl = 0;
				UdmSQLQuery(Indexer->db,"COMMIT");
				break;
			case UDM_DB_IBASE:
				UdmSQLQuery(Indexer->db,"COMMIT");
				((DB*)(Indexer->db))->commit_fl = 1;
				break;
			default:
				break;
		}
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	}
	return(IND_OK);
}


static int UdmStoreWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	int res;
	switch(Indexer->Conf->DBMode){
		case UDM_DBMODE_CACHE:
			/* It's not reentrant, let's lock it */
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_CACHE);
			res=UdmStoreWordsCache(Indexer,Doc);
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_CACHE);
			break;
		case UDM_DBMODE_MULTI:
		case UDM_DBMODE_MULTI_CRC:
			res=StoreWordsMulti(Indexer,Doc);
			break;
		case UDM_DBMODE_BLOB:
			res=UdmStoreWordsBlob(Indexer,Doc);
			break;
		case UDM_DBMODE_SINGLE:
		case UDM_DBMODE_SINGLE_CRC:
		default:
			res=StoreWordsSingle(Indexer,Doc);
			break;
	}
	return(res);
}

static int UdmDeleteAllFromDict(UDM_AGENT *Indexer){
	char qbuf[512];
	int i,last=0;
	DB * db=(DB*)Indexer->db;

	switch(Indexer->Conf->DBMode){
	case UDM_DBMODE_MULTI:
		for(i=MINDICT;i<MAXDICT;i++){
			if(last!=DICTNUM(i)){
				if (db->DBType == UDM_DB_ORACLE||db->DBType == UDM_DB_MSSQL)
					sprintf(qbuf,"TRUNCATE TABLE dict%d",DICTNUM(i));
				else
					sprintf(qbuf,"DELETE FROM dict%d",DICTNUM(i));

				UdmSQLQuery(Indexer->db,qbuf);
				if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
				last=DICTNUM(i);
			}
		}
		break;
	case UDM_DBMODE_MULTI_CRC:
		for(i=MINDICT;i<MAXDICT;i++){
			if(last!=DICTNUM(i)){
				if (db->DBType == UDM_DB_ORACLE||db->DBType == UDM_DB_MSSQL)
					sprintf(qbuf,"TRUNCATE TABLE ndict%d",DICTNUM(i));
				else
					sprintf(qbuf,"DELETE FROM ndict%d",DICTNUM(i));
	
				UdmSQLQuery(Indexer->db,qbuf);
				if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
				last=DICTNUM(i);
			}
		}
		break;
	case UDM_DBMODE_SINGLE_CRC:
		if (db->DBType == UDM_DB_ORACLE||db->DBType == UDM_DB_MSSQL)
			sprintf(qbuf,"TRUNCATE TABLE ndict");
		else
			sprintf(qbuf,"DELETE FROM ndict");
    	
		UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
		break;
	case UDM_DBMODE_BLOB:
		break;
	default:
		if (db->DBType == UDM_DB_ORACLE||db->DBType == UDM_DB_MSSQL)
			sprintf(qbuf,"TRUNCATE TABLE dict");
		else
			sprintf(qbuf,"DELETE FROM dict");

		UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
		break;
	}
	return(IND_OK);
}


/***************** CrossWords *******************************/

static int UdmDeleteAllFromCrossDict(UDM_AGENT * Indexer){
	char qbuf[1024];
	char table[64]="ncrossdict";
	int crcmode=1;

	if((Indexer->Conf->DBMode==UDM_DBMODE_SINGLE)||(Indexer->Conf->DBMode==UDM_DBMODE_MULTI)){
		strcpy(table,"crossdict");
		crcmode=0;
	}
	sprintf(qbuf,"DELETE FROM %s",table);
	UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}


static int UdmDeleteCrossWordFromURL(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	char qbuf[1024];
	char table[64]="ncrossdict";
	int crcmode=1;

	if((Indexer->Conf->DBMode==UDM_DBMODE_SINGLE)||(Indexer->Conf->DBMode==UDM_DBMODE_MULTI)){
		strcpy(table,"crossdict");
		crcmode=0;
	}
	if(Doc->url_id){
		sprintf(qbuf,"DELETE FROM %s WHERE url_id=%d",table,Doc->url_id);
		UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	}
	if(Doc->referrer){
		sprintf(qbuf,"DELETE FROM %s WHERE ref_id=%d",table,Doc->referrer);
		UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	}
	return(IND_OK);
}


static int UdmStoreCrossWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	UDM_DOCUMENT U;
	size_t i;
	char qbuf[1024];
	char table[64]="ncrossdict";
	int crcmode=1;
	const char * lasturl="scrap";
	int result=IND_OK;
	DB * db=(DB*)Indexer->db;

	UdmInitDoc(&U);
	U.referrer=Doc->url_id;
	result=UdmDeleteCrossWordFromURL(Indexer,&U);
	UdmInitDoc(&U);
	
	if(result!=IND_OK)return(result);
	
	if(Doc->CrossList.ncrosswords==0)return(IND_OK);

	if((Indexer->Conf->DBMode==UDM_DBMODE_SINGLE)||(Indexer->Conf->DBMode==UDM_DBMODE_MULTI)){
		strcpy(table,"crossdict");
		crcmode=0;
	}
	
	for(i=0;i<Doc->CrossList.ncrosswords;i++){
		if(strcmp(lasturl,Doc->CrossList.CrossWord[i].url)){
			U.url=Doc->CrossList.CrossWord[i].url;
			result=UdmFindURL(Indexer,Doc);
			if(result!=IND_OK)return(result);
			Doc->CrossList.CrossWord[i].referree_id=U.url_id;
			lasturl=Doc->CrossList.CrossWord[i].url;
		}
	}

	/* Begin transacttion/lock */
	if(Indexer->Conf->DBUseLock){
		switch(db->DBType){
			case UDM_DB_MYSQL:
				sprintf(qbuf,"LOCK TABLES %s WRITE",table);
				UdmSQLQuery(Indexer->db,qbuf);
				break;
			case UDM_DB_PGSQL:
				UdmSQLQuery(Indexer->db,"BEGIN WORK");
				break;
			case UDM_DB_ORACLE:
			case UDM_DB_SAPDB:
				((DB*)(Indexer->db))->commit_fl = 1;
				UdmSQLQuery(Indexer->db,"COMMIT");
				break;
			default:
				break;
		}
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	}

	/* Insert new words */
	for(i=0;i<Doc->CrossList.ncrosswords;i++){
		if(Doc->CrossList.CrossWord[i].weight){
			if(Doc->CrossList.CrossWord[i].referree_id){
				if(crcmode){
					sprintf(qbuf,"INSERT INTO %s (ref_id,url_id,word_id,intag) VALUES(%d,%d,%d,%d)",table,Doc->url_id,Doc->CrossList.CrossWord[i].referree_id,UdmStrCRC32(Doc->CrossList.CrossWord[i].word),Doc->WordList.Word[i].coord);
				}else{
					sprintf(qbuf,"INSERT INTO %s (ref_id,url_id,word,intag) VALUES(%d,%d,'%s',%d)",table,Doc->url_id,Doc->CrossList.CrossWord[i].referree_id,Doc->CrossList.CrossWord[i].word,Doc->WordList.Word[i].coord);
				}
				UdmSQLQuery(Indexer->db,qbuf);
				if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
			}		
		}
	}

	/* COMMIT/UNLOCK */
	if(Indexer->Conf->DBUseLock){
		switch(db->DBType){
			case UDM_DB_MYSQL:
				UdmSQLQuery(Indexer->db,"UNLOCK TABLES");
				break;
			case UDM_DB_PGSQL:
				UdmSQLQuery(Indexer->db,"END WORK");
				break;
			case UDM_DB_ORACLE:
			case UDM_DB_SAPDB:
				((DB*)(Indexer->db))->commit_fl = 0;
				UdmSQLQuery(Indexer->db,"COMMIT");
				break;
			default:
				break;
		}
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	}
	return(IND_OK);
}




/************************ URLs ***********************************/


static int UdmAddURL(UDM_AGENT *Indexer,UDM_DOCUMENT * Doc){
	char e_url[UDM_URLSIZE*4]="";
	int next_url_id=0;
	char qbuf[UDM_URLSIZE*4]="AddURL";
	const char * tag="", * category="";
	UDB_RES * res;
	DB * db=(DB*)Indexer->db;

	if(Doc->tag)tag=Doc->tag;
	if(Doc->category)category=Doc->category;

	/* Escape URL string */
	UdmDBEscStr(db->DBType,e_url,Doc->url);
	
	if(Indexer->Conf->use_crc32_url_id){
		/* Auto generation of rec_id */
		/* using CRC32 algorythm     */
		int rec_id=UdmStrCRC32(Doc->url);

		snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO url (rec_id,url,referrer,hops,crc32,next_index_time,status,tag,category) VALUES (%d,'%s',%d,%d,0,%d,0,'%s','%s')",rec_id,e_url,Doc->referrer,Doc->hops,(int)now(),tag,category);
	}else{
		/* Use dabatase generated rec_id */
		/* It depends on used DBType     */
		switch(db->DBType){
		case UDM_DB_MSQL:
			/* miniSQL has _seq as autoincrement value */
			res=UdmSQLQuery(Indexer->db,"SELECT _seq FROM url");
			if(UdmDBErrorCode(Indexer->db)){
				return(IND_ERROR);
			}
			next_url_id=atoi(UdmSQLValue(res,0,0));
			UdmSQLFree(res);
			snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO url (url,referrer,hops,rec_id,crc32,next_index_time,status,tag,category) VALUES ('%s',%d,%d,%d,0,%d,0,'%s','%s')",e_url,Doc->referrer,Doc->hops,next_url_id,(int)now(),tag,category);
			break;

		case UDM_DB_SOLID:
		case UDM_DB_ORACLE:
		case UDM_DB_ORACLE7:
		case UDM_DB_ORACLE8:
		case UDM_DB_SAPDB:
			/* FIXME: Dirty hack for stupid too smart databases */
			if (strlen(e_url)>UDM_URLSIZE)e_url[UDM_URLSIZE]=0;
			/* Use sequence next_url_id.nextval */
			snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO url (url,referrer,hops,rec_id,crc32,next_index_time,status,tag,category) VALUES ('%s',%d,%d,next_url_id.nextval,0,%d,0,'%s','%s')",e_url,Doc->referrer,Doc->hops,(int)now(),tag,category);
			break;

		case UDM_DB_MYSQL:
			/* MySQL generates itself */
		default:	
			snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO url (url,referrer,hops,crc32,next_index_time,status,tag,category) VALUES ('%s',%d,%d,0,%d,0,'%s','%s')",e_url,Doc->referrer,Doc->hops,(int)now(),tag,category);
		}
	}

	/* Exec INSERT now */
	UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))
		return(IND_ERROR);
	else
		return(IND_OK);
}
static int UdmDeleteURL(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc){
	char qbuf[128];
	int res;

	if(Indexer->Conf->use_crossword&&Indexer->Conf->DBMode!=UDM_DBMODE_CACHE){
		if((res=UdmDeleteCrossWordFromURL(Indexer,Doc))!=IND_OK)return(res);
	}
	if((res=UdmDeleteWordFromURL(Indexer,Doc))!=IND_OK)return(res);
	sprintf(qbuf,"DELETE FROM url WHERE rec_id=%d",Doc->url_id);
	UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}

__INDLIB__ int UdmMarkForReindex(UDM_AGENT *Indexer){
	char qbuf[1024];
	
	snprintf(qbuf,sizeof(qbuf),"UPDATE url SET next_index_time=%d WHERE rec_id<>0 %s%s%s%s%s",(int)now(),
		Indexer->Conf->tagstr,
		Indexer->Conf->urlstr,
		Indexer->Conf->statusstr,
		Indexer->Conf->langstr,
		Indexer->Conf->catstr);
	UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}

static int UdmUpdateUrl(UDM_AGENT *Indexer,UDM_DOCUMENT * Doc){
	char qbuf[256];
	
#ifdef HAVE_ORACLE8
	sprintf(qbuf,"UPDATE url SET status=:1,next_index_time=:2 WHERE rec_id=:3");
	param_init(Indexer->db, 1, 2, 3);
	param_add(Indexer->db, Doc->status, Doc->next_index_time, Doc->url_id);
#else
	sprintf(qbuf,"UPDATE url SET status=%d,next_index_time=%d WHERE rec_id=%d",Doc->status,(int)Doc->next_index_time,Doc->url_id);
#endif
	UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}

static int UdmRegisterChild(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc)
{
	char qbuf[1024];
	snprintf(qbuf,sizeof(qbuf),"insert into thread values(%d,%d)",Doc->parent_id,Doc->url_id);
	UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}

static int UdmDeleteAllFromThread(UDM_AGENT *Indexer){
	UdmSQLQuery(Indexer->db,"DELETE FROM thread");
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}


#define ECPY(to,from)	to=UdmDBEscStr(db->DBType,NULL,UDM_NULL2EMPTY(from))
static UDM_DOCUMENT * esc_copy_doc(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc,size_t *len){
	UDM_DOCUMENT * EDoc;
	DB * db=(DB*)Indexer->db;

	*len=0;
	EDoc=UdmAllocDoc();
	ECPY(EDoc->text,Doc->text);			(*len)+=strlen(EDoc->text);
	ECPY(EDoc->title,Doc->title);			(*len)+=strlen(EDoc->title);
	ECPY(EDoc->content_type,Doc->content_type);	(*len)+=strlen(EDoc->content_type);
	ECPY(EDoc->keywords,Doc->keywords);		(*len)+=strlen(EDoc->keywords);
	ECPY(EDoc->description,Doc->description);	(*len)+=strlen(EDoc->description);
	ECPY(EDoc->lang,Doc->lang);			(*len)+=strlen(EDoc->lang);
	ECPY(EDoc->tag,Doc->tag);			(*len)+=strlen(EDoc->tag);
	ECPY(EDoc->category,Doc->category);		(*len)+=strlen(EDoc->category);
	ECPY(EDoc->subject,Doc->subject);		(*len)+=strlen(EDoc->subject);
	ECPY(EDoc->from,Doc->from);			(*len)+=strlen(EDoc->from);
	ECPY(EDoc->references,Doc->references);		(*len)+=strlen(EDoc->references);
	ECPY(EDoc->newsgroups,Doc->newsgroups);		(*len)+=strlen(EDoc->newsgroups);
	ECPY(EDoc->message_id,Doc->message_id);		(*len)+=strlen(EDoc->message_id);

	return(EDoc);
}


static int UdmNewsUpdateURL(UDM_AGENT *Indexer,UDM_DOCUMENT * Doc){
	UDM_DOCUMENT * EDoc;
	char * qbuf;
	size_t len;

	EDoc=esc_copy_doc(Indexer,Doc,&len);
	qbuf=(char*)malloc(len+1024);

sprintf(qbuf,"\
UPDATE url SET \
status=%d,last_mod_time=%li,\
next_index_time=%li,\
tag='%s',txt='%s',title='%s',content_type='%s',docsize=%d,\
keywords='%s',description='%s',crc32=%d,lang='%s',\
header_subj='%s',header_from='%s',header_group='%s',\
header_refs='%s', msg_id='%s', charset='%s' \
WHERE rec_id=%d",
	Doc->status,
	Doc->last_mod_time,
	Doc->next_index_time,
	EDoc->tag,
	EDoc->text,
	EDoc->title,
	EDoc->content_type,
	Doc->size,
	EDoc->keywords,
	EDoc->description,
	Doc->crc32,
	EDoc->lang,
	EDoc->subject,
	EDoc->from,
	EDoc->newsgroups,
	EDoc->references,
	EDoc->message_id,
	(Doc->charset == NULL)?"":Doc->charset,
	Doc->url_id);
	
	UdmSQLQuery(Indexer->db,qbuf);
	UdmFreeDoc(EDoc);
	free(qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}


static int UdmLongUpdateURL(UDM_AGENT *Indexer,UDM_DOCUMENT * Doc){
	char * qbuf;
	size_t len;
	UDM_DOCUMENT * EDoc;

	if(Indexer->Conf->use_newsext)
		return(UdmNewsUpdateURL(Indexer,Doc));

	EDoc=esc_copy_doc(Indexer,Doc,&len);
	qbuf=(char*)malloc(len+1024);

sprintf(qbuf,"\
UPDATE url SET \
status=%d,last_mod_time=%li,\
next_index_time=%li,\
tag='%s',txt='%s',title='%s',content_type='%s',docsize=%d,\
keywords='%s',description='%s',crc32=%d,lang='%s',category='%s', charset='%s' \
WHERE rec_id=%d",
	Doc->status,Doc->last_mod_time,Doc->next_index_time,
	EDoc->tag,
	EDoc->text,EDoc->title,EDoc->content_type,
	Doc->size,
	EDoc->keywords,
	EDoc->description,
	Doc->crc32,
	EDoc->lang,
	EDoc->category,
	(Doc->charset == NULL)?"":Doc->charset,
	Doc->url_id);
	
	UdmSQLQuery(Indexer->db,qbuf);

	UdmFreeDoc(EDoc);
	free(qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}


static int UdmDeleteAllFromUrl(UDM_AGENT *Indexer){
	char qbuf[64];
	DB * db=(DB*)Indexer->db;

	switch(db->DBType){
		case UDM_DB_MSSQL:
		case UDM_DB_ORACLE:
			sprintf(qbuf,"TRUNCATE TABLE url");
			break;
		default:
			sprintf(qbuf,"DELETE FROM url");
	}	
	UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	return(IND_OK);
}



/************************ Clones stuff ***************************/
static int UdmFindOrigin(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc){
	size_t i=0;
	char qbuf[256];
	UDB_RES * res;

	Doc->origin_id=0;
	if (Doc->crc32==0)return IND_OK;
#ifdef HAVE_ORACLE8
	/* FIXME: remove Doc->size from here */
	sprintf(qbuf,"SELECT rec_id FROM url WHERE crc32=:1 AND (status=200 OR status=304 OR status=206) AND docsize=:2");
	param_init(Indexer->db, 1, 2, 0);
	param_add(Indexer->db, Doc->crc32, Doc->size, 0);
#else
	sprintf(qbuf,"SELECT rec_id FROM url WHERE crc32=%d AND (status=200 OR status=304 OR status=206)",Doc->crc32);
#endif
	Doc->origin_id=0;
	res=UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	for(i=0;i<UdmSQLNumRows(res);i++){
		const char *o;
		if((o=UdmSQLValue(res,i,0)))
			if((!Doc->origin_id)||(Doc->origin_id>atoi(o)))
				Doc->origin_id=atoi(o);
	}
	UdmSQLFree(res);
	return(IND_OK);
}

static int  UdmUpdateClone(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc){
	char * qbuf;

	/* This should be enough */
	qbuf=(char*)malloc(strlen(Doc->content_type)+128);
	sprintf(qbuf,"UPDATE url SET crc32=%d,status=%d,content_type='%s',last_mod_time=%li,next_index_time=%li WHERE rec_id=%d",
		Doc->crc32,Doc->status,Doc->content_type,Doc->last_mod_time,Doc->next_index_time,Doc->url_id);

	UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	free(qbuf);
	return(IND_OK);
}


/************** Get Target to be indexed ***********************/

UDM_DOCUMENT * UdmGetDocInfo(UDM_AGENT *Indexer,int index_flags){
	char sortstr[64]="";
	char updstr[64]="";
	char lmtstr[64]="";
	size_t i=0,url_num;
	UDM_DOCUMENT * Result=NULL;
	UDM_DOCUMENT * Doc=NULL;
	char urlin[1024*50]="";
	char qbuf[1024*50];
	DB* db=(DB*)(Indexer->db);
	UDB_RES * res;

	if(Indexer->Conf->Targets.num_indexed>=Indexer->Conf->Targets.num_rows){
		UdmFreeTargets(&Indexer->Conf->Targets);
		bzero(&Indexer->Conf->Targets,sizeof(Indexer->Conf->Targets));

		url_num=URL_SELECT_CACHE;

		if(Indexer->Conf->DBUseLock){
			switch(db->DBType){
				case UDM_DB_MYSQL:
					UdmSQLQuery(Indexer->db,"LOCK TABLES url WRITE");
					break;
				case UDM_DB_PGSQL:
					UdmSQLQuery(Indexer->db,"BEGIN WORK");
					UdmSQLQuery(Indexer->db,"LOCK url");
					break;
				case UDM_DB_ORACLE:
					sprintf(updstr, " FOR UPDATE ");
#if HAVE_ORACLE8
					sprintf(lmtstr, " AND ROWNUM <=:2"); 
#else
					sprintf(lmtstr, " AND ROWNUM <=%d", url_num); 
#endif
					break;
				case UDM_DB_SAPDB:
					sprintf(updstr, " WITH LOCK ");
					strcpy(lmtstr, "");
					break;
				default:
					break;
			}
			if(UdmDBErrorCode(Indexer->db))return(NULL);
		}

		sprintf(sortstr,(index_flags&UDM_FLAG_EXP_FIRST)?" ORDER BY next_index_time":"");
		if (index_flags & UDM_FLAG_SORT_HOPS) {
			if (strlen(sortstr) == 0) {
				sprintf(sortstr," ORDER BY hops ASC ");
			} else {
				strcat(sortstr,", hops ASC ");
			}
		}

		if(db->DBSQL_LIMIT){
			sprintf(qbuf,"SELECT url,rec_id,docsize,status,hops,crc32,last_mod_time FROM url WHERE next_index_time<=%d %s%s%s%s%s%s LIMIT %d",
				(int)now(),
				Indexer->Conf->tagstr,
				Indexer->Conf->urlstr,
				Indexer->Conf->statusstr,
				Indexer->Conf->langstr,
				Indexer->Conf->catstr,
				sortstr,url_num);
		}else{
			db->res_limit=url_num;
#if HAVE_ORACLE8
			param_init(Indexer->db, 1, 2, 0);
			param_add(Indexer->db, (int)now(), url_num, 0);
			sprintf(qbuf,"SELECT url,rec_id,docsize,status,hops,crc32,last_mod_time FROM url WHERE next_index_time<=:1 %s%s%s%s%s%s%s%s",
				Indexer->Conf->tagstr,
				Indexer->Conf->urlstr,
				Indexer->Conf->statusstr,
				Indexer->Conf->langstr,
				Indexer->Conf->catstr,
				lmtstr,sortstr,updstr);
#else
			sprintf(qbuf,"SELECT url,rec_id,docsize,status,hops,crc32,last_mod_time FROM url WHERE next_index_time<=%d %s%s%s%s%s%s%s%s",
				(int)now(),
				Indexer->Conf->tagstr,
				Indexer->Conf->urlstr,
				Indexer->Conf->statusstr,
				Indexer->Conf->langstr,
				lmtstr,sortstr,
				Indexer->Conf->catstr,
				updstr);
#endif
		}
		res=UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db)){
			return(NULL);
		}
		Indexer->Conf->Targets.num_rows=UdmSQLNumRows(res);
		Indexer->Conf->Targets.num_indexed=0;
		Indexer->Conf->Targets.Doc=(UDM_DOCUMENT*)(UDM_DOCUMENT*)UdmXmalloc(sizeof(UDM_DOCUMENT)*(Indexer->Conf->Targets.num_rows));

		if(!Indexer->Conf->Targets.num_rows){
			UdmSQLFree(res);
			if(Indexer->Conf->DBUseLock){
				switch(db->DBType){
					case UDM_DB_MYSQL:
						UdmSQLQuery(Indexer->db,"UNLOCK TABLES");
						break;
					case UDM_DB_PGSQL:
						UdmSQLQuery(Indexer->db,"END WORK");
						break;
					default:
						break;
				}
			}
			return(NULL);
		}
		
		for(i=0;i<Indexer->Conf->Targets.num_rows;i++){
			Doc=&Indexer->Conf->Targets.Doc[i];
			Doc->url=strdup(UdmSQLValue(res,i,0));
			Doc->url_id=atoi(UdmSQLValue(res,i,1));
			Doc->size=atoi(UdmSQLValue(res,i,2));
			Doc->status=atoi(UdmSQLValue(res,i,3));
			Doc->hops=atoi(UdmSQLValue(res,i,4));
			Doc->crc32=strtol(UdmSQLValue(res,i,5), NULL, 10);
			Doc->last_mod_time=atol(UdmSQLValue(res,i,6));
		}
		UdmSQLFree(res);

		urlin[0]=0;
		for(i=0;i<Indexer->Conf->Targets.num_rows;i++){
			Doc=&Indexer->Conf->Targets.Doc[i];
			if(db->DBSQL_IN){
				if(urlin[0])strcat(urlin,",");
				sprintf(urlin+strlen(urlin),"%d",Doc->url_id);
			}else{
				sprintf(qbuf,"UPDATE url SET next_index_time=%d WHERE rec_id=%d",(int)(now()+URL_LOCK_TIME),Doc->url_id);
				UdmSQLQuery(Indexer->db,qbuf);
				if(UdmDBErrorCode(Indexer->db)){
					return(NULL);
				}
			}
		}

		if(db->DBSQL_IN){
			sprintf(qbuf,"UPDATE url SET next_index_time=%d WHERE rec_id in (%s)",(int)(now()+URL_LOCK_TIME),urlin);
			UdmSQLQuery(Indexer->db,qbuf);
			if(UdmDBErrorCode(Indexer->db)){
				return(NULL);
			}
		}
		if(Indexer->Conf->DBUseLock){
			switch(db->DBType){
				case UDM_DB_MYSQL:
					UdmSQLQuery(Indexer->db,"UNLOCK TABLES");
					break;
				case UDM_DB_PGSQL:
					UdmSQLQuery(Indexer->db,"END WORK");
					break;
				default:
					break;
			}
			if(UdmDBErrorCode(Indexer->db))return(NULL);
		}
	}
	Result=UdmAllocDoc();
	Doc=&Indexer->Conf->Targets.Doc[Indexer->Conf->Targets.num_indexed];
	Result->url=strdup(Doc->url?Doc->url:"");
	Result->url_id=Doc->url_id;
	Result->size=Doc->size;
	Result->status=Doc->status;
	Result->hops=Doc->hops;
	Result->crc32=Doc->crc32;
	Result->last_mod_time=Doc->last_mod_time;
	Indexer->Conf->Targets.num_indexed++;
	return(Result);
}


/************************************************************/
/* Misc functions                                           */
/************************************************************/

typedef struct stat_struct {
	int status;
	int expired;
	int total;
} UDM_STAT;



__INDLIB__ int UdmGetDocCount(UDM_AGENT * Indexer){
	char qbuf[200]="";
	int docc=0;
	UDB_RES * res;

	sprintf(qbuf,NDOCS_QUERY);
	res=UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(0);
	if(UdmSQLNumRows(res)){
		const char * s;
		s=UdmSQLValue(res,0,0);
		if(s)docc=atoi(s);
	}
	UdmSQLFree(res);
	return(docc);
}


__INDLIB__ int UdmGetStatistics(UDM_AGENT *Indexer){
	size_t i,j;
	char lims[2048];
	char qbuf[2048];
	int expired_total=0;
	int total_total=0;
	UDB_RES * res;
	DB * db=(DB*)Indexer->db;
	int have_group=db->DBSQL_GROUP;
	
	if(!Indexer->Conf->StatInfo)return(IND_OK);
	
	if(db->DBType==UDM_DB_IBASE)
		have_group=0;

	snprintf(lims,sizeof(lims),"%s%s%s%s%s",
		Indexer->Conf->tagstr,
		Indexer->Conf->urlstr,
		Indexer->Conf->statusstr,
		Indexer->Conf->langstr,
		Indexer->Conf->catstr);

	if(have_group){
		
		switch(db->DBType){
		
			case UDM_DB_MYSQL:
				snprintf(qbuf,sizeof(qbuf)-1,"SELECT status,sum(next_index_time<=%d),count(*) FROM url WHERE rec_id<>0 %s GROUP BY status",(int)now(),lims);
				break;
		
			case UDM_DB_PGSQL:
			case UDM_DB_MSSQL:
				snprintf(qbuf,sizeof(qbuf)-1,"SELECT status,sum(case when next_index_time<=%d then 1 else 0 end),count(*) FROM url WHERE rec_id<>0 %s GROUP BY status",(int)now(),lims);
				break;

			case UDM_DB_ORACLE:
			case UDM_DB_SAPDB:
				snprintf(qbuf,sizeof(qbuf)-1,"SELECT status, SUM(DECODE(SIGN(%d-next_index_time),-1,0,1,1)), count(*) FROM url WHERE rec_id<>0 %s GROUP BY status",(int)now(),lims);
				break;
		}

		res=UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
		if((j=UdmSQLNumRows(res))){
			for(i=0;i<j;i++){
				int status,expired,total;
				status=atoi(UdmSQLValue(res,i,0));
				expired=atoi(UdmSQLValue(res,i,1));
				total=atoi(UdmSQLValue(res,i,2));
				total_total+=total;
				expired_total+=expired;
				Indexer->Conf->StatInfo(Indexer->handle,status,expired,total,UdmHTTPErrMsg(status));
			}
		}
		UdmSQLFree(res);
	
	}else{
/*	
	FIXME: learn how to get it from SOLID and IBASE
	(HAVE_IBASE||HAVE_MSQL || HAVE_SOLID || HAVE_VIRT )
	
*/
		snprintf(qbuf,sizeof(qbuf)-1,"SELECT status,next_index_time FROM url WHERE rec_id>0 %s",lims);

		res=UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
		if(UdmSQLNumRows(res)){
			UDM_STAT *udmstat=NULL;
			int n,num=0;
		
			n=now();
			for(i=0;i<UdmSQLNumRows(res);i++){
				for(j=0;j<num;j++){
					if(udmstat[j].status==atoi(UdmSQLValue(res,i,0))){
						if(atoi(UdmSQLValue(res,i,1))<=n)
							udmstat[j].expired++;
						udmstat[j].total++;
						break;
					}
				}
				if(j==num){
				
					if(!num)
						udmstat=(UDM_STAT *)UdmXmalloc(sizeof(UDM_STAT)*(num+1));
					else
						udmstat=(UDM_STAT *)UdmXrealloc(udmstat,sizeof(UDM_STAT)*(num+1));
					udmstat[j].status=atoi(UdmSQLValue(res,i,0));
					udmstat[j].expired=0;
					if(atoi(UdmSQLValue(res,i,1))<=n)
						udmstat[j].expired++;
					udmstat[j].total=1;
					num++;
				}
			}
			for(i=0; i<num; i++){
				Indexer->Conf->StatInfo(Indexer->handle,
					udmstat->status,udmstat->expired,udmstat->total,
					UdmHTTPErrMsg(udmstat->status));
				total_total+=udmstat->total;
				expired_total+=udmstat->expired;
				udmstat++;
			}
		}
		UdmSQLFree(res);
	}
	Indexer->Conf->StatInfo(Indexer->handle,-1,expired_total,total_total,"");
	return(IND_OK);
}



__INDLIB__ int UdmGetReferers(UDM_AGENT *Indexer){
	size_t i,j;
	char qbuf[2048];
	UDB_RES * res;

	snprintf(qbuf,sizeof(qbuf),"SELECT url.status,url2.url,url.url FROM url,url url2 WHERE url.referrer=url2.rec_id %s%s%s%s%s",
		Indexer->Conf->tagstr,
		Indexer->Conf->urlstr,
		Indexer->Conf->statusstr,
		Indexer->Conf->langstr,
		Indexer->Conf->catstr);

	res=UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
	j=UdmSQLNumRows(res);
	for(i=0;i<j;i++){
		if(Indexer->Conf->RefInfo)Indexer->Conf->RefInfo(
			atoi(UdmSQLValue(res,i,0)),
			UdmSQLValue(res,i,2),
			UdmSQLValue(res,i,1)
		);
	}
	UdmSQLFree(res);
	return(IND_OK);
}

int UdmClearDB(UDM_AGENT *Indexer){
	size_t i,j;
	int err;
	char qbuf[1024*4];
	DB * db=(DB*)Indexer->db;

	if(!Indexer->Conf->tagstr[0]&&!Indexer->Conf->urlstr[0]&&!Indexer->Conf->statusstr[0]&&!Indexer->Conf->langstr[0]&&!Indexer->Conf->catstr[0]){

		if(Indexer->Conf->use_newsext){
			if((IND_OK!=(err=UdmDeleteAllFromThread(Indexer))))return(err);
		}
		
		if(Indexer->Conf->DBMode==UDM_DBMODE_CACHE){
			UdmClearCacheTree(Indexer->Conf);
		}

		if(Indexer->Conf->use_crossword&&Indexer->Conf->DBMode!=UDM_DBMODE_CACHE){
			if((IND_OK!=(err=UdmDeleteAllFromCrossDict(Indexer))))return(err);
		}
		if((IND_OK!=(err=UdmDeleteAllFromDict(Indexer))))return(err);
		if((IND_OK!=(err=UdmDeleteAllFromUrl(Indexer))))return(err);
	}else{
		j=0;
		while(1){
			int last=0;
			char urlin[1024*3]="";
			char limit[100]="";
			UDB_RES * res;
			
			if(db->DBSQL_LIMIT){
				sprintf(limit," LIMIT %d",URL_DELETE_CACHE);
			}
			sprintf(qbuf,"SELECT rec_id FROM url WHERE rec_id<>0 %s%s%s%s%s%s",
				Indexer->Conf->tagstr,
				Indexer->Conf->urlstr,
				Indexer->Conf->statusstr,
				Indexer->Conf->langstr,
				Indexer->Conf->catstr,
				limit);

			res=UdmSQLQuery(Indexer->db,qbuf);
			if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
			if(UdmSQLNumRows(res)){
				UDM_DOCUMENT Doc;
				
				bzero(&Doc,sizeof(Doc));
				if(Indexer->Conf->DBMode==UDM_DBMODE_CACHE){
					
					if(IND_OK!=UdmOpenCache(Indexer->Conf,UdmDBErrorMsg(Indexer->db))){
						return(IND_ERROR);
					}
					for(i=0;i<UdmSQLNumRows(res);i++){
						Doc.url_id=atoi(UdmSQLValue(res,i,0));
						if(IND_OK!=UdmDeleteURL(Indexer,&Doc))
							return(IND_ERROR);
					}
					continue;
				}
				if(db->DBSQL_IN){
					urlin[0]=0;
					for(i=0;i<UdmSQLNumRows(res);i++){
						if(i)strcat(urlin,",");
						strcat(urlin,UdmSQLValue(res,i,0));
					}
					j+=i;
					switch(Indexer->Conf->DBMode){
					case UDM_DBMODE_MULTI:
						for(i=MINDICT;i<MAXDICT;i++){
							if(last!=DICTNUM(i)){
								sprintf(qbuf,"DELETE FROM dict%d WHERE url_id in (%s)",DICTNUM(i),urlin);
								UdmSQLQuery(Indexer->db,qbuf);
								if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
								last=DICTNUM(i);
							}
						}
						break;
					case UDM_DBMODE_MULTI_CRC:
						for(i=MINDICT;i<MAXDICT;i++){
							if(last!=DICTNUM(i)){
								sprintf(qbuf,"DELETE FROM ndict%d WHERE url_id in (%s)",
									DICTNUM(i),urlin);
								UdmSQLQuery(Indexer->db,qbuf);
								if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
								last=DICTNUM(i);
							}
						}
						break;
					case UDM_DBMODE_CACHE:
						break;
					default:
						sprintf(qbuf,"DELETE FROM dict WHERE url_id in (%s)",urlin);
						UdmSQLQuery(Indexer->db,qbuf);
						if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);
						break;
					}
					sprintf(qbuf,"DELETE FROM url WHERE rec_id in (%s)",urlin);
					UdmSQLQuery(Indexer->db,qbuf);
					if(UdmDBErrorCode(Indexer->db))return(IND_ERROR);

					UdmSQLFree(res);
				}else{
					for(i=0;i<UdmSQLNumRows(res);i++)
						Doc.url_id=atoi(UdmSQLValue(res,i,0));
						if(IND_OK!=UdmDeleteURL(Indexer,&Doc))
							return(IND_ERROR);
					j+=i;
				}
			}else{
				UdmSQLFree(res);
				break;
			}
		}
	}
	return(IND_OK);
}

/********************* Categories ************************************/
UDM_CATEGORY * UdmCatList(UDM_AGENT * Indexer,const char * addr){
	UDM_CATEGORY * r=NULL;
	size_t i;
	char qbuf[1024];
	DB * db;
	UDB_RES * res;

	db=(DB*)Indexer->db;
	if(!addr)addr="";
	
	if(db->DBType==UDM_DB_SAPDB){
		snprintf(qbuf,sizeof(qbuf)-1,"SELECT rec_id,path,lnk,name FROM categories WHERE path LIKE '%s__'",addr);
	}else{
		snprintf(qbuf,sizeof(qbuf)-1,"SELECT rec_id,path,link,name FROM categories WHERE path LIKE '%s__'",addr);
	}
	res=UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(db))return(NULL);
	if(UdmSQLNumRows(res)){
		size_t nbytes;
		
		nbytes=sizeof(UDM_CATEGORY)*(UdmSQLNumRows(res)+1);
		r=(UDM_CATEGORY*)UdmXmalloc(nbytes);
		for(i=0;i<UdmSQLNumRows(res);i++){
			r[i].rec_id=atoi(UdmSQLValue(res,i,0));
			strcpy(r[i].path,UdmSQLValue(res,i,1));
			strcpy(r[i].link,UdmSQLValue(res,i,2));
			strcpy(r[i].name,UdmSQLValue(res,i,3));
			r[i+1].rec_id=0;
		}
	}
	UdmSQLFree(res);
	return(r);
}

UDM_CATEGORY * UdmCatPath(UDM_AGENT * Indexer,const char * addr){
	UDM_CATEGORY * r=NULL;
	size_t i,l;
	char qbuf[1024];
	DB * db=(DB*)Indexer->db;

	db=Indexer->db;
	if(!addr)addr="";
	l=(strlen(addr)/2)+1;

	r=(UDM_CATEGORY*)UdmXmalloc(sizeof(UDM_CATEGORY)*(l+1));
	r[0].rec_id=0;

	for(i=0;i<l;i++){
		char head[128];
		UDB_RES * res;
		
		r[i+1].rec_id=0;
		strncpy(head,addr,i*2);head[i*2]=0;

		if(db->DBType==UDM_DB_SAPDB){
			snprintf(qbuf,sizeof(qbuf)-1,"SELECT rec_id,path,lnk,name FROM categories WHERE path='%s'",head);
		}else{
			snprintf(qbuf,sizeof(qbuf)-1,"SELECT rec_id,path,link,name FROM categories WHERE path='%s'",head);
		}

		res=UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(db))return(NULL);
		if(UdmSQLNumRows(res)){
			r[i].rec_id=atoi(UdmSQLValue(res,0,0));
			strcpy(r[i].path,UdmSQLValue(res,0,1));
			strcpy(r[i].link,UdmSQLValue(res,0,2));
			strcpy(r[i].name,UdmSQLValue(res,0,3));
		}
		UdmSQLFree(res);
	}
	return(r);
}


/******************* Search stuff ************************************/

UDM_RESULT * UdmCloneList(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	UDM_RESULT * Res;
	size_t i;
	char qbuf[256];
	void * db;
	UDB_RES * res;

	if(Indexer->sdcl.nclients){
		return(UdmCloneListSearchd(Indexer,Doc));
	}

	db=Indexer->db;
	sprintf(qbuf,"SELECT rec_id,url,content_type,last_mod_time,docsize FROM url WHERE crc32=%d AND (status=200 OR status=304 OR status=206)",Doc->crc32);
	res=UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(db))return(NULL);
	
	Res=(UDM_RESULT*)UdmXmalloc(sizeof(UDM_RESULT));
	Res->Doc=NULL;
	Res->num_rows=UdmSQLNumRows(res);

	if(!(Res->num_rows))return(Res);

	Res->Doc=(UDM_DOCUMENT*)UdmXmalloc(sizeof(UDM_DOCUMENT)*(Res->num_rows));

	for(i=0;i<UdmSQLNumRows(res);i++){
		UdmInitDoc(&Res->Doc[i]);
		Res->Doc[i].url_id=atoi(UdmSQLValue(res,i,0));
		Res->Doc[i].url=strdup(UdmSQLValue(res,i,1));
		Res->Doc[i].content_type=strdup(UdmSQLValue(res,i,2));
		Res->Doc[i].last_mod_time=atol(UdmSQLValue(res,i,3));
		Res->Doc[i].size=atol(UdmSQLValue(res,i,4));
	}
	UdmSQLFree(res);
	return(Res);
}
/********************************************************/


static UDM_URL_CRD * UdmFindSQL(UDM_AGENT * query,const char * text){
	char qbuf[1024*4];
	UDM_URL_CRD * wrd=NULL;	
	size_t wordnum;
	int has_crosswrd=0;
	UDB_RES * sqlres;
	DB * db=(DB*)query->db;
	
	
#ifdef DEBUG_SEARCH
	unsigned long debug_search_ticks;
	debug_search_ticks=UdmStartTimer();
#endif

	UdmPrepare(query,text);

	/* Now find each word */
	for(wordnum=0;wordnum<query->WWList.nwords;wordnum++){
		int numrows,firstnum,curnum,tnum,tmin,tmax,tlst=-1;
		size_t i;
		char tablename[32]="dict";
		char escwrd[1000];
		
		UdmDBEscStr(db->DBType,escwrd,query->WWList.Words[wordnum].word);

		if((query->Conf->DBMode==UDM_DBMODE_MULTI)&&(query->word_match!=UDM_MATCH_FULL)){
			/* This is for substring search!  */
			/* In Multi mode: we have to scan */
			/* almost all tables except those */
			/* with to short words            */
			
			tmin=DICTNUM(strlen(query->WWList.Words[wordnum].word));
			tmax=MAXDICT;
		}else{
			tmin=tmax=DICTNUM(strlen(query->WWList.Words[wordnum].word));
		}

		for(tnum=tmin;tnum<=tmax;tnum++){

			if(tlst!=DICTNUM(tnum)){
				tlst=DICTNUM(tnum);		
#ifdef DEBUG_SEARCH
				debug_search_ticks=UdmStartTimer();
				fprintf(stderr,"Start search for '%s'\n",query->WWList.Words[wordnum].word);
#endif
				switch(query->Conf->DBMode){
				case UDM_DBMODE_MULTI:
					sprintf(tablename,"dict%d",DICTNUM(tnum));
					break;
				case UDM_DBMODE_MULTI_CRC:
					sprintf(tablename,"ndict%d",DICTNUM(tnum));
					break;
				case UDM_DBMODE_SINGLE_CRC:
					strcpy(tablename,"ndict");
					break;
				default:
					break;
				}
				if((query->Conf->DBMode==UDM_DBMODE_SINGLE_CRC)||(query->Conf->DBMode==UDM_DBMODE_MULTI_CRC)){
					int crc;
					crc=query->WWList.Words[wordnum].crcword;
					if(query->Conf->tagstr[0] || query->Conf->statusstr[0] || query->Conf->urlstr[0] || query->Conf->langstr[0] || query->Conf->timestr[0] || query->Conf->catstr[0]){
						snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT %s.url_id,%s.intag \
FROM %s,url \
WHERE %s.word_id=%d \
AND url.rec_id=%s.url_id %s%s%s%s%s%s",
						tablename,tablename,
						tablename,tablename,
						crc,tablename,
						query->Conf->tagstr,
						query->Conf->statusstr,
						query->Conf->urlstr,
						query->Conf->langstr,
						query->Conf->timestr,
						query->Conf->catstr);
					}else{
						snprintf(qbuf,sizeof(qbuf)-1,"SELECT url_id,intag FROM %s WHERE word_id=%d",tablename,crc);
					}
				}else{
					char cmparg[256];
					switch(query->word_match){	
						case UDM_MATCH_BEGIN:
							sprintf(cmparg," LIKE '%s%%'",escwrd);
							break;
						case UDM_MATCH_END:
							sprintf(cmparg," LIKE '%%%s'",escwrd);
							break;
						case UDM_MATCH_SUBSTR:
							sprintf(cmparg," LIKE '%%%s%%'",escwrd);
							break;
						case UDM_MATCH_FULL:
						default:
							sprintf(cmparg,"='%s'",escwrd);
							break;
					}
					if(query->Conf->tagstr[0]||query->Conf->statusstr[0]||query->Conf->urlstr[0]||query->Conf->langstr[0]||query->Conf->catstr[0]||query->Conf->timestr[0]){
						snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT %s.url_id,%s.intag \
FROM %s,url \
WHERE %s.word%s \
AND url.rec_id=%s.url_id %s%s%s%s%s%s",
						tablename,tablename,
						tablename,tablename,
						cmparg,tablename,
						query->Conf->tagstr,
						query->Conf->statusstr,
						query->Conf->urlstr,
						query->Conf->langstr,
						query->Conf->timestr,
						query->Conf->catstr);
					}else{
						snprintf(qbuf,sizeof(qbuf)-1,"SELECT url_id,intag FROM %s WHERE word%s",tablename,cmparg);
					}
				}
				sqlres=UdmSQLQuery(query->db,qbuf);
				if(UdmDBErrorCode(query->db))return(NULL);
				numrows=UdmSQLNumRows(sqlres);

#ifdef DEBUG_SEARCH
				debug_search_ticks=UdmStartTimer()-debug_search_ticks;
				fprintf(stderr,"Stop search for '%s'\t%.2f  %d found\n",query->WWList.Words[wordnum].word,(float)debug_search_ticks/1000,numrows);
#endif

				/* Add new found word to the list */
				if(!query->total_found){
					wrd=(UDM_URL_CRD*)UdmXmalloc((query->total_found+numrows)*sizeof(UDM_URL_CRD));
				}else{
					wrd=(UDM_URL_CRD*)UdmXrealloc(wrd,(query->total_found+numrows)*sizeof(UDM_URL_CRD));
				}

				firstnum=curnum=query->total_found;
				for(i=0;i<numrows;i++){
					int url_id=0;
					uint4 coord=0;
					uint4 weight;
					uint4 section;
					int done=0;
#ifdef HAVE_MYSQL
					if(db->DBDriver==UDM_DB_MYSQL){
						MYSQL_ROW row;
						/* mysql_data_seek is slow */
						/* We will use sequential fetch instead*/
						row=mysql_fetch_row(sqlres->mysqlres);
						url_id=atoi(row[0]);
						coord=atoi(row[1]);
						done=1;
					}
#endif
					if(!done){
						url_id=atoi(UdmSQLValue(sqlres,i,0));
						coord=atoi(UdmSQLValue(sqlres,i,1));
					}

					section=UDM_WRDSEC(coord);
					weight=query->wf[section];
					if(weight){
						coord=coord&0xFFFF0000;
						wrd[curnum].url_id=url_id;
						wrd[curnum].coord=coord+(weight<<8)+query->WWList.Words[wordnum].order;
						/* printf("%d %08X\n",wrd[curnum].url_id,wrd[curnum].coord);*/
						curnum++;
					}
				}
				UdmSQLFree(sqlres);
				query->WWList.Words[wordnum].count+=curnum-firstnum;
				query->total_found=curnum;
				wrd=(UDM_URL_CRD*)UdmXrealloc(wrd,query->total_found*sizeof(UDM_URL_CRD));
			}
		}	
	}

	/* Now find each word in crosstable */
	has_crosswrd=((query->Conf->use_crossword)&&(query->Conf->DBMode!=UDM_DBMODE_CACHE));
	for(wordnum=0;((has_crosswrd)&&(wordnum<query->WWList.nwords));wordnum++){
		int numrows,firstnum,curnum;
		char tablename[32]="";
		char escwrd[1000];
		size_t i;

		UdmDBEscStr(db->DBType,escwrd,query->WWList.Words[wordnum].word);
		
#ifdef DEBUG_SEARCH
		debug_search_ticks=UdmStartTimer();
		fprintf(stderr,"Start search for '%s'\n",query->WWList.Words[wordnum].word);
#endif
		switch(query->Conf->DBMode){
		case UDM_DBMODE_SINGLE_CRC:
		case UDM_DBMODE_MULTI_CRC:
			strcpy(tablename,"ncrossdict");
			break;
		case UDM_DBMODE_MULTI:
		default:
			strcpy(tablename,"crossdict");
			break;
		}
		if((query->Conf->DBMode==UDM_DBMODE_SINGLE_CRC)||
			(query->Conf->DBMode==UDM_DBMODE_MULTI_CRC)){
			int crc;
			crc=query->WWList.Words[wordnum].crcword;
			if(query->Conf->tagstr[0] || query->Conf->statusstr[0] || query->Conf->urlstr[0] || query->Conf->langstr[0] || query->Conf->timestr[0] || query->Conf->catstr[0]){
				snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT %s.url_id,%s.intag \
FROM %s,url \
WHERE %s.word_id=%d \
AND url.rec_id=%s.url_id %s%s%s%s%s%s",
				tablename,tablename,
				tablename,tablename,
				crc,tablename,
				query->Conf->tagstr,
				query->Conf->statusstr,
				query->Conf->urlstr,
				query->Conf->langstr,
				query->Conf->timestr,
				query->Conf->catstr);
			}else{
				snprintf(qbuf,sizeof(qbuf)-1,"SELECT url_id,intag FROM %s WHERE word_id=%d",tablename,crc);
			}
		}else{
			char cmparg[256];
			switch(query->word_match){	
				case UDM_MATCH_BEGIN:
					sprintf(cmparg," LIKE '%s%%'",escwrd);
					break;
				case UDM_MATCH_END:
					sprintf(cmparg," LIKE '%%%s'",escwrd);
					break;
				case UDM_MATCH_SUBSTR:
					sprintf(cmparg," LIKE '%%%s%%'",escwrd);
					break;
				case UDM_MATCH_FULL:
				default:
					sprintf(cmparg,"='%s'",escwrd);
					break;
			}
			if(query->Conf->tagstr[0]||query->Conf->statusstr[0]||query->Conf->urlstr[0]||query->Conf->langstr[0]||query->Conf->catstr[0]||query->Conf->timestr[0]){
		
				snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT %s.url_id,%s.intag \
FROM %s,url \
WHERE %s.word%s \
AND url.rec_id=%s.url_id %s%s%s%s%s%s",
				tablename,tablename,
				tablename,tablename,
				cmparg,tablename,
				query->Conf->tagstr,
				query->Conf->statusstr,
				query->Conf->urlstr,
				query->Conf->langstr,
				query->Conf->timestr,
				query->Conf->catstr);
			}else{
				snprintf(qbuf,sizeof(qbuf)-1,"SELECT url_id,intag FROM %s WHERE word%s",tablename,cmparg);
			}
		}
		sqlres=UdmSQLQuery(query->db,qbuf);
		if(UdmDBErrorCode(query->db))return(NULL);
		numrows=UdmSQLNumRows(sqlres);

#ifdef DEBUG_SEARCH
		debug_search_ticks=UdmStartTimer()-debug_search_ticks;
		fprintf(stderr,"Stop search for '%s'\t%.2f  %d found\n",query->WWList.Words[wordnum].word,(float)debug_search_ticks/1000,numrows);
#endif

		/* Add new found word to the list */
		if(!query->total_found){
			wrd=(UDM_URL_CRD*)UdmXmalloc((query->total_found+numrows)*sizeof(UDM_URL_CRD));
		}else{
			wrd=(UDM_URL_CRD*)UdmXrealloc(wrd,(query->total_found+numrows)*sizeof(UDM_URL_CRD));
		}

		firstnum=curnum=query->total_found;
		for(i=0;i<numrows;i++){
			int url_id=0;
			int coord=0;
			int done=0;
			int weight;
			int section;
#ifdef HAVE_MYSQL
			if(db->DBDriver==UDM_DB_MYSQL){
				MYSQL_ROW row;
				
				/* mysql_data_seek is slow */
				/* We will use sequential fetch instead*/
				
				row=mysql_fetch_row(sqlres->mysqlres);
				url_id=atoi(row[0]);
				coord=atoi(row[1]);
				done=1;
			}
#endif

			if(!done){
				url_id=atoi(UdmSQLValue(sqlres,i,0));
				coord=atoi(UdmSQLValue(sqlres,i,1));
			}

			section=UDM_WRDSEC(coord);
			weight=query->wf[section];
			if(weight){
				coord=coord&0xFFFF0000;
				wrd[curnum].url_id=url_id;
				wrd[curnum].coord=coord+(weight<<8)+query->WWList.Words[wordnum].order;
				curnum++;
			}
		}
		UdmSQLFree(sqlres);
		query->total_found=curnum;
		query->WWList.Words[wordnum].count+=curnum-firstnum;
		wrd=(UDM_URL_CRD*)UdmXrealloc(wrd,query->total_found*sizeof(UDM_URL_CRD));
	}

	for(wordnum=0;wordnum<query->WWList.nwords;wordnum++){
		if(query->wordinfo[0])strcat(query->wordinfo,", ");
		sprintf(UDM_STREND(query->wordinfo)," %s : %d",query->WWList.Words[wordnum].word,query->WWList.Words[wordnum].count);
	}

	if(query->total_found){
		/* Sort in URL order */
#ifdef DEBUG_SEARCH
		fprintf(stderr,"Start sort by url_id %d words\n",query->total_found);
		debug_search_ticks=UdmStartTimer();
#endif
		UdmSortSearchWordsByURL(wrd,query->total_found);
#ifdef DEBUG_SEARCH
		debug_search_ticks=UdmStartTimer()-debug_search_ticks;
		fprintf(stderr,"Stop sort by url_id:\t%.2f\n",(float)debug_search_ticks/1000);
#endif
		/* Group words by URL */
#ifdef DEBUG_SEARCH
		fprintf(stderr,"Start group by url_id %d docs\n",query->total_found);
		debug_search_ticks=UdmStartTimer();
#endif

		UdmGroupByURL(query,wrd);
#ifdef DEBUG_SEARCH
		debug_search_ticks=UdmStartTimer()-debug_search_ticks;
		fprintf(stderr,"Stop group by url_id:\t%.2f\n",(float)debug_search_ticks/1000);
#endif

	}
	return(wrd);
}


static char * UdmHlConvert(UDM_AGENT * query,const char * src, UDM_CHARSET * lcs, UDM_CHARSET * bcs)
{
	int * tok, * lt, ures;
	char * hpart, * htxt;
	size_t len;
	int * uni;

	if(!src)return NULL;
	
	len=strlen(src);
	hpart=(char*)malloc(len+1);
	htxt=(char*)malloc(len*10);
	htxt[0]='\0';

	/* Convert to unicode */
	uni=(int *)malloc((len+10)*sizeof(int));
	ures=lcs->to_uni(uni,src,len+1);

	/* Parse unicode string */
	tok=UdmUniGetSepToken(uni,&lt);
	while(tok){
		int found=0;
		size_t slen,flen;
		int euchar;
		size_t uw;

		flen=lt-tok;

		/* Convert token to BrowserCharset */
		euchar=tok[flen];
		tok[flen]=0;
		hpart[0]='\0';
		ures=bcs->from_uni(hpart,tok,len+1);
		tok[flen]=euchar;

		/* Check that it is word to be marked */
		for(uw=0;uw<query->WWList.nwords;uw++){
			slen=UdmUniLen(query->WWList.Words[uw].uword);
			if((flen==slen)&&(!UdmUniStrNCaseCmp(tok,query->WWList.Words[uw].uword,slen))){
				found=1;
				break;
			}
		}
		if(found)strcat(htxt,"\2");
		strcat(htxt,hpart);
		if(found)strcat(htxt,"\3");
		
		tok=UdmUniGetSepToken(NULL,&lt);
	}
	free(hpart);
	free(uni);
	/*
	fprintf(stderr,"otxt='%s'\n",src);
	fprintf(stderr,"htxt='%s'\n",htxt);
	*/
	return(htxt);
}

static int UdmTrack(UDM_AGENT * query,const char * words){
	char qbuf[1024*4]="";
	char text_escaped[1024*4]="";
	DB * db=(DB*)query->db;

	/* Escape text to track it  */
	UdmDBEscStr(db->DBType,text_escaped,words);
	
	snprintf(qbuf,sizeof(qbuf)-1,"INSERT INTO qtrack (qwords,qtime,found,ps,np,m,wm,o,t,cat,ul,wf,g,tmplt) VALUES ('%s',%d,%d,'%s','%s','%s','%s','%s','%s','%s','%s','%s','%s','%s')",
		text_escaped,(int)now(),query->total_found,
		UdmFindStrVar(query->Conf->vars, "ps", ""),
		UdmFindStrVar(query->Conf->vars, "np", ""),
		UdmFindStrVar(query->Conf->vars, "m", ""),
		UdmFindStrVar(query->Conf->vars, "wm", ""),
		UdmFindStrVar(query->Conf->vars, "o", ""),
		UdmFindStrVar(query->Conf->vars, "t", ""),
		UdmFindStrVar(query->Conf->vars, "cat", ""),
		UdmFindStrVar(query->Conf->vars, "ul", ""),
		UdmFindStrVar(query->Conf->vars, "wf", ""),
		UdmFindStrVar(query->Conf->vars, "g", ""),
		UdmFindStrVar(query->Conf->vars, "tmplt", "")
	);
	UdmSQLQuery(query->db,qbuf);
	if(UdmDBErrorCode(query->db))return(IND_ERROR);
	return(IND_OK);
	
}

int UdmResAddDocInfo(UDM_AGENT * query, UDM_RESULT * Res){
	size_t i,j;
	char instr[1024*4]="";
	char qbuf[1024*4];
	UDM_CHARSET * lcs, * bcs;
	UDB_RES * sqlres;
	DB * db=(DB*)query->db;

	if(db->DBSQL_IN){

		/* Compose IN string and set to zero url_id field */
		for(i=0;i<Res->num_rows;i++){
			sprintf(UDM_STREND(instr),"%s%d",i?",":"",Res->Doc[i].url_id);
		}

		snprintf(qbuf,sizeof(qbuf),"SELECT rec_id,url,content_type,last_mod_time,title,txt,docsize,next_index_time,referrer,keywords,description,crc32,tag,category,lang FROM url WHERE rec_id IN (%s) ORDER BY rec_id",instr);
		sqlres=UdmSQLQuery(query->db,qbuf);
		if(UdmDBErrorCode(query->db)){
			
			return IND_ERROR;
		}

		for(j=0;j<Res->num_rows;j++) {
			for(i=0;i<UdmSQLNumRows(sqlres);i++) {
				if(Res->Doc[j].url_id == atoi(UdmSQLValue(sqlres,i,0))) {
					Res->Doc[j].order=j+Res->first+1;
					Res->Doc[j].url=strdup(UdmSQLValue(sqlres,i,1));
					Res->Doc[j].content_type=strdup(UdmSQLValue(sqlres,i,2));
					Res->Doc[j].last_mod_time=atol(UdmSQLValue(sqlres,i,3));
					Res->Doc[j].title=strdup(UdmSQLValue(sqlres,i,4));
					Res->Doc[j].text=strdup(UdmSQLValue(sqlres,i,5));
					Res->Doc[j].size=atoi(UdmSQLValue(sqlres,i,6));
					Res->Doc[j].next_index_time=atoi(UdmSQLValue(sqlres,i,7));
					Res->Doc[j].referrer=atoi(UdmSQLValue(sqlres,i,8));
					Res->Doc[j].keywords=strdup(UdmSQLValue(sqlres,i,9));
					Res->Doc[j].description=strdup(UdmSQLValue(sqlres,i,10));
					Res->Doc[j].crc32=atoi(UdmSQLValue(sqlres,i,11));
					Res->Doc[j].tag=strdup(UdmSQLValue(sqlres,i,12));
					Res->Doc[j].category=strdup(UdmSQLValue(sqlres,i,13));
					Res->Doc[j].lang=strdup(UdmSQLValue(sqlres,i,14));
					break;
				}
			}
		}
		UdmSQLFree(sqlres);
	}else{
		for(i=0;i<Res->num_rows;i++){

			sprintf(qbuf,"SELECT rec_id,url,content_type,last_mod_time,title,txt,docsize,next_index_time,referrer,keywords,description,crc32,tag,category,lang FROM url WHERE rec_id=%d",Res->Doc[i].url_id);
			sqlres=UdmSQLQuery(query->db,qbuf);
			if(UdmDBErrorCode((query->db))){
				return(IND_ERROR);
			}
			if(UdmSQLNumRows(sqlres)){
				Res->Doc[i].order=i+Res->first+1;
				Res->Doc[i].url=strdup(UdmSQLValue(sqlres,0,1));
				Res->Doc[i].content_type=strdup(UdmSQLValue(sqlres,0,2));
				Res->Doc[i].last_mod_time=atol(UdmSQLValue(sqlres,0,3));
				Res->Doc[i].title=strdup(UdmSQLValue(sqlres,0,4));
				Res->Doc[i].text=strdup(UdmSQLValue(sqlres,0,5));
				Res->Doc[i].size=atoi(UdmSQLValue(sqlres,0,6));
				Res->Doc[i].next_index_time=atoi(UdmSQLValue(sqlres,0,7));
				Res->Doc[i].referrer=atoi(UdmSQLValue(sqlres,0,8));
				Res->Doc[i].keywords=strdup(UdmSQLValue(sqlres,0,9));
				Res->Doc[i].description=strdup(UdmSQLValue(sqlres,0,10));
				Res->Doc[i].crc32=atoi(UdmSQLValue(sqlres,0,11));
				Res->Doc[i].tag=strdup(UdmSQLValue(sqlres,0,12));
				Res->Doc[i].category=strdup(UdmSQLValue(sqlres,0,13));
				Res->Doc[i].lang=strdup(UdmSQLValue(sqlres,0,14));	
			}
			UdmSQLFree(sqlres);
		}
	}
	
	/* Recode & Mark WORDINFO, Title, Description, Keywords, Body */
	lcs=UdmGetCharSet(query->Conf->local_charset?query->Conf->local_charset:"utf-8");
	bcs=UdmGetCharSet(query->Conf->browser_charset?query->Conf->browser_charset:"us-ascii");
	
	if(lcs)lcs=UdmGetCharSet("utf-8");
	if(bcs)bcs=UdmGetCharSet("utf-8");
	
	{
		size_t len;
		int * uni, ures;
		len=strlen(Res->wordinfo);
		uni=(int *)malloc((len+10)*sizeof(int));
		ures=lcs->to_uni(uni,Res->wordinfo,len+1);
		ures=bcs->from_uni(Res->wordinfo,uni,len+1);
		UDM_FREE(uni);
	}

	for(i=0;i<Res->num_rows;i++){
		char * htxt;

		htxt=UdmHlConvert(query,Res->Doc[i].title,lcs,bcs);
		UDM_FREE(Res->Doc[i].title);
		Res->Doc[i].title=strdup(htxt?htxt:"");
		free(htxt);

		htxt=UdmHlConvert(query,Res->Doc[i].description,lcs,bcs);
		UDM_FREE(Res->Doc[i].description);
		Res->Doc[i].description=strdup(htxt?htxt:"");
		free(htxt);
			
		htxt=UdmHlConvert(query,Res->Doc[i].keywords,lcs,bcs);
		UDM_FREE(Res->Doc[i].keywords);
		Res->Doc[i].keywords=strdup(htxt?htxt:"");
		free(htxt);
			
		htxt=UdmHlConvert(query,Res->Doc[i].text,lcs,bcs);
		UDM_FREE(Res->Doc[i].text);
		Res->Doc[i].text=strdup(htxt?htxt:"");
		free(htxt);
	}
	return(IND_OK);
}


UDM_URL_CRD * UdmFindWords(UDM_AGENT * query, const char * words){
	UDM_URL_CRD * wrd=NULL;
	char param[1024*4]="";
	
	query->wordinfo[0]='\0';
	query->total_found=0;
	query->loaded_nwords=0;
	query->withoffset=0;


	/* Compose search cache file name   */
	/* taking in account all parameters */
	/* and query words themself         */
	snprintf(param,sizeof(param)-1,"%s %s %s %s %s %s %s %s %d %d",
		query->Conf->catstr,
		query->Conf->langstr,
		query->Conf->tagstr,
		query->Conf->statusstr,
		query->Conf->urlstr,
		query->Conf->timestr,
		query->Conf->catlimit,
		query->Conf->taglimit,
		query->Conf->sitelimit,
		query->search_mode);

	snprintf(param,sizeof(param),"%s%s%s%08X.%08X",query->Conf->vardir,"cache",UDMSLASHSTR,UdmStrCRC32(param),UdmStrCRC32(words));

	if((query->cache_mode==UDM_CACHE_ENABLED)&&(wrd=UdmFindInCache(query,param,&query->loaded_nwords))){
		query->withoffset=0;
	}else{
		/* If not found in search cache       */
		/* Let's do actual search now:        */
		/* Get groupped by url_id words array */
		
		if(query->Conf->DBMode==UDM_DBMODE_CACHE){
			wrd=UdmFindCache(query,words);
		}else
		if(query->Conf->DBMode==UDM_DBMODE_BLOB){
			wrd=UdmFindBlob(query,words);
		}else{
			wrd=UdmFindSQL(query,words);
			if(UdmDBErrorCode(query->db))return(NULL);
		}
#ifdef DEBUG_SEARCH
		fprintf(stderr,"Start sort by weight %d docs\n",query->total_found);
		debug_search_ticks=UdmStartTimer();
#endif
		/* Now sort results by the weight */
		if(query->page_number<FAST_PRESORT_PAGES){
			size_t topcount;
			topcount=query->page_size*(query->page_number+1)-1;
			if(topcount>=query->total_found)topcount=query->total_found-1;
			UdmWrdTopSort(wrd,query->total_found,topcount);
		}else{
			UdmSortSearchWordsByWeight(wrd,query->total_found);
		}
#ifdef DEBUG_SEARCH
		debug_search_ticks=UdmStartTimer()-debug_search_ticks;
		fprintf(stderr,"Stop sort by weight:\t%.2f\n",(float)debug_search_ticks/1000);
#endif
		/* Store into search result cache */
#if (WIN32|WINNT)
#else
		if((query->cache_mode==UDM_CACHE_ENABLED)&&(search_cache_size>-1)){
			fflush(stdout);
			fflush(stderr);
			if(fork()==0){
				/* Store to cache and exit program */
				UdmStoreToCache(query,param,wrd,query->total_found);
				exit(0);
			}
		}
#endif
		query->withoffset=1;
	}
	return wrd;
}


UDM_RESULT * UdmFind(UDM_AGENT * query,const char * words){
	UDM_RESULT * Res=NULL;
	UDM_URL_CRD *wrd=NULL;
	size_t i;
	unsigned long ticks;
#ifdef DEBUG_SEARCH
	 unsigned long debug_search_ticks;
#endif

	ticks=UdmStartTimer();
	
	if(query->sdcl.nclients){
		const char * request;
		request=UdmFindStrVar(query->Conf->vars,"QUERY_STRING",words);
		return(UdmSearchdFind(query,request));
	}

	/* Load words array */
	wrd=UdmFindWords(query,words);
	if(UdmDBErrorCode(query->db))return(NULL);

	/* Allocate result */
	Res=(UDM_RESULT*)UdmXmalloc(sizeof(UDM_RESULT));
	Res->num_rows=query->loaded_nwords;
	Res->total_found=query->total_found;
	strncpy(Res->wordinfo,query->wordinfo,sizeof(Res->wordinfo)-1);

	/* Track query if required */
	if((query->track_mode)&UDM_TRACK_QUERIES){
		if(IND_OK!=UdmTrack(query,words))
			return NULL;
		
	}

	/* Return if no words found   */
	if(!query->total_found){
		Res->work_time=UdmStartTimer()-ticks;
		return(Res);
	}

	/* Find curent page offsets */
	/* And If there was a request for */
	/* too large page number we will  */
	/* display the only one last doc  */
	Res->first=query->page_number*query->page_size;	
	if(Res->first>=query->total_found)Res->first=query->total_found-1;

	/* If results more than 1 page */
	/* we must cut the tail        */
	if((Res->first+query->page_size)>query->total_found){
		Res->num_rows=query->total_found-Res->first;
	}else{
		Res->num_rows=query->page_size;
	}
	Res->last=Res->first+Res->num_rows-1;

#ifdef DEBUG_SEARCH
	fprintf(stderr,"First: %d Found: %d Display: %d\n",first,query->total_found,display_docs);
#endif

#ifdef DEBUG_SEARCH
	fprintf(stderr,"Start URL stuff\n");
	debug_search_ticks=UdmStartTimer();
#endif

	/* Allocate an array for documents information */
	Res->Doc=(UDM_DOCUMENT*)UdmXmalloc(sizeof(UDM_DOCUMENT)*(Res->num_rows));

	/* Copy url_id and coord to result */
	for(i=0;i<Res->num_rows;i++){
		UdmInitDoc(&Res->Doc[i]);
		Res->Doc[i].url_id=wrd[i+Res->first*query->withoffset].url_id;
		Res->Doc[i].rating=wrd[i+Res->first*query->withoffset].coord;
	}

	/* Free words array */
	UDM_FREE(wrd);

	/* Load documents information */
	

	if(IND_OK!=UdmResAddDocInfo(query,Res)){
		UdmFreeResult(Res);
		return NULL;
	}


#ifdef DEBUG_SEARCH
	debug_search_ticks=UdmStartTimer()-debug_search_ticks;
	fprintf(stderr,"Stop URL stuff:\t\t%.2f\n",(float)debug_search_ticks/1000);
#endif

	/* first and last begins from 0, make it begin from 1 */
	Res->first++;
	Res->last++;
	Res->work_time=UdmStartTimer()-ticks;
	return(Res);
}



/***********************************************************/
/*  HTDB stuff:  Indexing of database content              */
/***********************************************************/


#ifdef USE_HTDB
static char * get_path_part(char *path,char *dst,int part){
	const char *s;
	char *e;
	int i=0;

	s=path;
	while(*s){
		if(i==part){
			if((e=strchr(s,'/')))strncpy(dst,s,(unsigned)(e-s));
			else	strcpy(dst,s);
			return(dst);
		}
		if(*s=='/')i++;
		s++;
	}
	*dst=0;
	return(dst);
}
static char * include_params(const char *src,char *path,char *dst){
	const char *s;
	char *e;
	int i;
	
	s=src;e=dst;*e=0;
	while(*s){
		if((*s=='\\')){
			*e=*(s+1);e++;*e=0;s+=2;
			continue;
		}
		if(*s!='$'){
			*e=*s;
			s++;e++;*e=0;
			continue;
		}
		s++;i=atoi(s);
		while((*s>='0')&&(*s<='9'))s++;
		get_path_part(path,e,i);
		while(*e)e++;
	}
	return(dst);
}


int UdmHTDBGet(UDM_AGENT * Indexer,char *path,char *filename,char * htdb_list, char * htdb_doc){
	char qbuf[1024*4]="";
	DB * db;
	UDB_RES * res;

	db=(DB*)Indexer->db;
	Indexer->buf[0]=0;
	if(*filename){
		char real_path[1024]="";
		
		snprintf(real_path,sizeof(real_path)-1,"%s%s",path,filename);
		include_params(htdb_doc?htdb_doc:"",real_path,qbuf);
		res=UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(db))return(0);
		if(UdmSQLNumRows(res)==1){
			sprintf(Indexer->buf,"%s",UdmSQLValue(res,0,0));
		}else{
			sprintf(Indexer->buf,"HTTP/1.0 404 Not Found\r\n\r\n");
		}
		UdmSQLFree(res);
	}else{
		char *s;
		size_t len,i;

		include_params(htdb_list,path,qbuf);
		res=UdmSQLQuery(Indexer->db,qbuf);
		if(UdmDBErrorCode(db))return(0);
		sprintf(Indexer->buf,"HTTP/1.0 200 OK\r\nContent-type: text/html\r\n\r\n<HTML><BODY>\n");
		
		if(res){
			s=UDM_STREND(Indexer->buf);
			len=UdmSQLNumRows(res);
			for(i=0;i<UdmSQLNumRows(res);i++){
				sprintf(UDM_STREND(s),"<a href=\"%s\">%s</a><br>\n",
					UdmSQLValue(res,i,0),UdmSQLValue(res,i,0));
			}
			UdmSQLFree(res);
		}
		strcat(Indexer->buf,"</BODY></HTML>\n");
	}
	return(strlen(Indexer->buf));
}
#endif

/************************** make index stuff *******************************/

UDM_UINT8_URLID* GetPair8FromSQL(UDM_AGENT *Agent,const char *field,int *count,int type){
	char	qbuf[128];
	DB	*db;
	size_t	i;
	UDB_RES * res;

	db=(DB*)Agent->db;
	snprintf(qbuf,sizeof(qbuf)-1,"SELECT %s,rec_id FROM url",field);
	
	res=UdmSQLQuery(Agent->db,qbuf);
	if(UdmDBErrorCode(db)){
		fprintf(stderr,"Error: %s\n",UdmDBErrorMsg(db));
		return(NULL);
	}
	if(res){
		size_t nrows=UdmSQLNumRows(res);
		UDM_UINT8_URLID	*data=NULL;
	
		data=(UDM_UINT8_URLID*)malloc(nrows*sizeof(UDM_UINT8_URLID));
		if(!data){
			fprintf(stderr,"Error: %s\n",strerror(errno));
			return(NULL);
		}
		for(i=0;i<nrows;i++){
			const char *val0=UdmSQLValue(res,i,0);
			const char *val1=UdmSQLValue(res,i,1);

			switch(type){
				case 0: UdmDecodeHex8Str(val0,&data[i].hi,&data[i].lo); break;
				case 1: data[i].hi=atoi(val0); data[i].lo=0; break;
			}
			data[i].url_id=atoi(val1);
		}
		*count=nrows;
		UdmSQLFree(res);
		return(data);
	}	
	return(NULL);
}

UDM_UINT4_URLID* GetPair4FromSQL(UDM_AGENT *Agent,const char *field,int *count,int type){
	char	qbuf[128];
	DB	*db;
	size_t	i;
	UDB_RES * res;

	db=(DB*)Agent->db;
	snprintf(qbuf,sizeof(qbuf)-1,"SELECT %s,rec_id FROM url",field);
	
	res=UdmSQLQuery(Agent->db,qbuf);
	if(UdmDBErrorCode(db)){
		fprintf(stderr,"Error: %s\n",UdmDBErrorMsg(db));
		return(NULL);
	}
	if(res){
		size_t nrows=UdmSQLNumRows(res);
		UDM_UINT4_URLID	*data=NULL;
	
		data=(UDM_UINT4_URLID*)malloc(nrows*sizeof(UDM_UINT4_URLID));
		if(!data){
			fprintf(stderr,"Error: %s\n",strerror(errno));
			return(NULL);
		}
		for(i=0;i<nrows;i++){
			const char *val0=UdmSQLValue(res,i,0);
			const char *val1=UdmSQLValue(res,i,1);

			switch(type){
				case 0: data[i].val=atoi(val0)/3600; break;
				case 1: data[i].val=atoi(val0)/60; break;
			}
			data[i].url_id=atoi(val1);
		}
		*count=nrows;
		UdmSQLFree(res);
		return(data);
	}	
	return(NULL);
}

/************************** /make index stuff *******************************/

__INDLIB__ UDM_URLSTATE * UdmGetURLState(UDM_AGENT * Indexer, const char * hostinfo,time_t lmt){
	char qbuf[1024]="";
	size_t rows,i;
	UDM_URLSTATE * Res=NULL;
	DB * db;
	UDB_RES * res;

	db=(DB*)Indexer->db;
	
	sprintf(qbuf,"SELECT status,url FROM url WHERE last_mod_time>=%li AND url LIKE '%s%%'",lmt,hostinfo);
	res=UdmSQLQuery(Indexer->db,qbuf);
	if(UdmDBErrorCode(db))return(NULL);
	if(res){
		rows=UdmSQLNumRows(res);
		Res=(UDM_URLSTATE *)malloc((rows+1)*sizeof(UDM_URLSTATE));
		for(i=0;i<UdmSQLNumRows(res);i++){
			Res[i].status=atoi(UdmSQLValue(res,i,0));
			Res[i].url=strdup(UdmSQLValue(res,i,0));
		}
		UdmSQLFree(res);
		Res[i].status=0;
		Res[i].url=NULL;
	}
	return(Res);
}

/***************************************************************************/


int UdmURLAction(UDM_AGENT * A, UDM_DOCUMENT * D, int cmd){
	int res;

	switch(cmd){
		case UDM_URL_ACTION_DELETE:
			res=UdmDeleteURL(A,D);
			break;
			
		case UDM_URL_ACTION_ADD:
			res=UdmAddURL(A,D);
			break;
			
		case UDM_URL_ACTION_SUPDATE:
			res=UdmUpdateUrl(A,D);
			break;
			
		case UDM_URL_ACTION_LUPDATE:
			res=UdmLongUpdateURL(A,D);
			break;
			
		case UDM_URL_ACTION_INSWORDS:
			res=UdmStoreWords(A,D);
			break;
			
		case UDM_URL_ACTION_INSCWORDS:
			res=UdmStoreCrossWords(A,D);
			break;
			
		case UDM_URL_ACTION_DELWORDS:
			res=UdmDeleteWordFromURL(A,D);
			break;
			
		case UDM_URL_ACTION_DELCWORDS:
			res=UdmDeleteCrossWordFromURL(A,D);
			break;
			
		case UDM_URL_ACTION_UPDCLONE:
			res=UdmUpdateClone(A,D);
			break;
			
		case UDM_URL_ACTION_REGCHILD:
			res=UdmRegisterChild(A,D);
			break;
			
		case UDM_URL_ACTION_FINDBYURL:
			res=UdmFindURL(A,D);
			break;
			
		case UDM_URL_ACTION_FINDBYMSG:
			res=UdmFindMessage(A,D);
			break;
			
		case UDM_URL_ACTION_FINDORIG:
			res=UdmFindOrigin(A,D);
			break;
			
		default:
			res=IND_ERROR;
	}
	return res;
}

#endif /* HAVE_SQL */
