#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <dirent.h>

#include "udm_common.h"
#include "udm_services.h"
#include "udm_utils.h"
#include "udm_urldcl.h"
#include "udm_doc.h"

#include "udm_proto.h"


#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long) -1)
#endif


static int open_host(char *hostname,int port, int timeout)
{
	int net;
	struct hostent *host;
	struct sockaddr_in sa_in;

	bzero((char*)&sa_in,sizeof(sa_in));

	if (port){
		sa_in.sin_port= htons((u_short)port);
	}else{
		return(UDM_NET_ERROR);
	}

	if ((sa_in.sin_addr.s_addr=inet_addr(hostname)) != INADDR_NONE){
		sa_in.sin_family=AF_INET;
	}else{
		host=gethostbyname(hostname);
		if (host){
			sa_in.sin_family=host->h_addrtype;
			memcpy(&sa_in.sin_addr, host->h_addr, (size_t)host->h_length);
		}else{
			return(UDM_NET_CANT_RESOLVE);
		}
	}
	net=socket(AF_INET, SOCK_STREAM, 0);

	if(connect(net, (struct sockaddr *)&sa_in, sizeof (sa_in)))
		return(UDM_NET_CANT_CONNECT);

	return(net);
}


int UdmOpenURLD(UDM_ENV * Conf){
	if(Conf->urld_addr){
		char * host,*prt;
		int port=UDM_URLD_PORT;
		
		host=(Conf->urld_addr)?strdup(Conf->urld_addr):strdup("localhost");
		if((prt=strchr(host,':'))){
			*prt='\0';
			if(!(port=atoi(prt+1)))
				port=UDM_URLD_PORT;
		}
		if((Conf->urld_fd=open_host(host,port,60)) < 0){
			sprintf(Conf->errstr,"Can't connect to urld at %s:%d",host,port);
			UDM_FREE(host);
			Conf->urld_fd=0;
			return(IND_ERROR);
		}
		UDM_FREE(host);
	}
	return(IND_OK);
}

void UdmCloseURLD(UDM_ENV * Conf){
	if(Conf->urld_fd>0)closesocket(Conf->urld_fd);
}


UDM_DOCUMENT * UdmURLDSelect(UDM_ENV * Env){
	char buf[1024*2]="select\r\n";
	int size,recieved=0,ntok=0;
	UDM_DOCUMENT * Res;
	char * tok, *lt;

	size=send(Env->urld_fd,(void*)&buf,strlen(buf),0);
	printf("%s\n",buf);

	Res=UdmAllocDoc();
	
	while((size=recv(Env->urld_fd,(void*)(&buf+recieved),sizeof(buf),0))>0){
		printf("size=%d\n",size);
		recieved+=size;
		buf[recieved]='\0';
		if(strchr(buf,'\n'))
			break;
	}
	printf("'%s' %d\n",buf,size);
	tok=UdmGetToken(buf,"\t\r\n",&lt);
	while(tok){
		switch(ntok){
			case 0: 
				Res->url=strdup(tok);
				if(!strcmp(Res->url,"-")){
					free(Res);
					return NULL;
				}
				break;
			case 1: Res->last_mod_time=atoi(tok);break;
			case 2: Res->url_id=atoi(tok);break;
			case 3: Res->size=atoi(tok);break;
			case 4: Res->status=atoi(tok);break;
			case 5: Res->hops=atoi(tok);break;
			case 6: Res->crc32=atoi(tok);break;
		}	
		ntok++;
		tok=UdmGetToken(NULL,"\t\r\n",&lt);
	}
	return Res;
}

int UdmURLDSUpdate(UDM_ENV * Env,UDM_DOCUMENT * Doc){
	char buf[1024];
	int size;
	
	sprintf(buf,"supdate\t%d\t%d\t%li\n",Doc->url_id,Doc->status,Doc->next_index_time);
	size=send(Env->urld_fd,(void*)&buf,strlen(buf),0);
	printf("%s\n",buf);
	return size;
}

int UdmURLDInsert(UDM_ENV * Env,UDM_DOCUMENT * Doc){
	char buf[1024];
	int size;
	
	sprintf(buf,"insert\t%s\t%d\t%d\t%s\t%s\n",Doc->url,Doc->referrer,Doc->hops,Doc->tag?Doc->tag:"",Doc->category?Doc->category:"");
	size=send(Env->urld_fd,(void*)&buf,strlen(buf),0);
	printf("%s\n",buf);
	return size;
}



int UdmURLDLUpdate(UDM_ENV * Env,UDM_DOCUMENT * Doc){
	char buf[10*1024];
	int size;
	
	sprintf(buf,"lupdate\t%d\t%d\t%d\t%d\t%d\t%d\t%li\t%li\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n",
		Doc->url_id,
		Doc->status,
		Doc->size,
		Doc->referrer,
		Doc->hops,
		Doc->crc32,
		Doc->last_mod_time,
		Doc->next_index_time,
		Doc->tag?Doc->tag:"",
		Doc->category?Doc->category:"",
		Doc->url,
		Doc->content_type?Doc->content_type:"",
		Doc->title?Doc->title:"",
		Doc->keywords?Doc->keywords:"",
		Doc->description?Doc->description:"",
		Doc->text?Doc->text:""
		);
	size=send(Env->urld_fd,(void*)&buf,strlen(buf),0);
	printf("%s\n",buf);
	return size;
}

